/*
   This function performs Cyclic Reduction on a Block Tridiagonal Matrix.
   It is used as a direct Poisson Solver for the Pressure Corrector.  It is
   taken from the algorithms 4.5.14 and 4.5.15 given in "Matrix Computations",
   by Golub, page 179.


   Because I use pointer notation rather than [][], which I call bracket notation,
   it is probably difficult to keep track of the indexing of the variables.  So
   I provide the following table:

   pointer notation              bracket notation

   *(A + n2*ii + jj)             A[ii][jj]
   *(D + dp + q*ii + jj)         D[p][ii][jj]
   *(F + dp + q*ii + jj)         F[p][ii][jj]
   *(D + dpm1 + q*ii + jj)       D[p-1][ii][jj]
   *(F + dpm1 + q*ii + jj)       F[p-1][ii][jj]

   *(b + bp + jr + ii)           b[p][j*r][ii]
   *(b + bpm1 + jr + ii)         b[p-1][j*r][ii]
   *(b + bpm1 + jrmr2 + ii)      b[p-1][j*r - r/2 - 1][ii]
   *(b + bpm1 + jrpr2 + ii)      b[p-1][j*r + r/2 - 1][ii]

   *(x + xkm1 + ii)              x[pow(2, k-1) - 1][ii]
   *(x + jm1r + ii)              x[(2*j-1)*r - 1][ii]
   *(x + jm2r + ii)              x[(2*j-2)*r - 1][ii]

   You should compare this with the algorithms on page 179.


    SLCFD source file
    Version:  1.0
    Copyright (C) 2006  San Le 

    The source code contained in this file is released under the
    terms of the GNU Library General Public License.
 
             Updated 5/9/06
*/

#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#define CFD_FLAG        1

int ludcmp(double *, int , int *, double *, double *);
int lubksb(double *, int, int *, double *);

int decomp(double *, int );

int backsub( double *, int , double *);

int solve2X2( double *, double *, double * );

int solve3X3( double *, double *, double * );

int dotX( double *,double *,double *, int );

int matX( double *,double *,double *, int ,int ,int );

int matSquare( double *, double *, int );

int Cyclic_Reduction1(double *D0, double *F0, double *B, double *x, int n2)
{
/* For when the Tridiagonal block in A(n,n) is > 1X1 or greater

             Updated 5/3/06
*/

	int ii, jj, kk, i ,j, k, p, check, n;
	int bp, bpm1, bkm1, dp, dpm1, dkm1, xkm1,  r, r2;
	int jr, jrmr2, jrpr2, jkmp, jm1r, jm2r, jkmpm1;
	double *b, c[1], *D, *F, fdum, fdum2;

	n = n2;
	fdum = (double)(n+1);
	fdum = log(fdum)/log(2.0);
	k = (int)fdum;

	printf( "k, n %d %d\n", k, n);

	D = (double *)calloc(k,sizeof(double));
	F = (double *)calloc(k,sizeof(double));
	b = (double *)calloc(k*n2,sizeof(double));

/* To maintain the same natation as the reference, I am using ii, jj, and kk
   as the dummy variables reserving j and k for the algorithm variables.

	printf( "\n\n");
	for( ii = 0; ii < n2; ++ii )
	{
	    for( jj = 0; jj < n2; ++jj )
	    {
		    printf( "%16.6e ", *(A + n2*ii + jj));
	    }
	    printf( "\n");
	}
*/

/* Calculate b(0), D(0), and F(0) by setting them to the tridiagonal block components
   of A.
*/
	for( ii = 0; ii < n2; ++ii )
	{
		*(b + ii) = *(B + ii);
		/*printf( "%16.6e\n", *(b + ii));*/
	}

	printf( "\n\n");
	*(D) = *(D0);
	*(F) = *(F0);
	/*printf( "%16.6e %16.6e\n", *(D), *(F));*/

/* This is algorithm 4.5.14
*/
	for( p = 1; p <= k - 1; ++p )
	{
		r = pow(2, p);
		r2 = pow(2, p-1);
		dp = p;
		dpm1 = p-1;
		bp = n2*p;
		bpm1 = n2*(p-1);

		*(F+dp) = *(F+dpm1)*( *(F+dpm1) );
		*(D+dp) = 2.0*( *(F+dp) ) - *(D+dpm1)*( *(D+dpm1) );

		jkmp = pow(2, k-p) - 1;
		for( j = 1; j <= jkmp; ++j )
		{
		   jr = j*r - 1;
		   jrmr2 = j*r - r2 - 1;
		   jrpr2 = j*r + r2 - 1;
		   *(b+bp+jr) =
			*(F+dpm1)*( *(b+bpm1+jrmr2) + *(b+bpm1+jrpr2) ) -
			*(D+dpm1)*( *(b+bpm1+jr));
		}
	}

/* This is algorithm 4.5.15
*/
	dkm1 = k-1;
	bkm1 = n2*(k-1) + pow(2, k-1) - 1;
	xkm1 = pow(2, k-1) - 1;

	*(x+xkm1) = *(b+bkm1)/( *(D+dkm1) );

	for( p = k-2; p >= 0; --p )
	{
		dp = p;
		bp = n2*p;

		r = pow(2, p);

/* I break the next loop of j in the algorithm into 3 parts:

      1) j = 1
      2) 2 = j < 2^(k-p-1)
      3) j = 2^(k-p-1)

*/
		j = 1;
		jr = 2*j*r - 1;
		jm1r = (2*j-1)*r - 1;
		jm2r = (2*j-2)*r - 1;
		*(c) = *(b+bp+jm1r) - *(F+dp)*( *(x+jr) );

/* Solve D(q,q)*x(q) = c(q) */

		*(x+jm1r) = *(c)/( *(D+dp) );

		jkmpm1 = pow(2, k-p-1);
		for( j = 2; j <= jkmpm1; ++j )
		{
		    jr = 2*j*r - 1;
		    jm1r = (2*j-1)*r - 1;
		    jm2r = (2*j-2)*r - 1;
		    *(c) = *(b+bp+jm1r) - *(F+dp)*( *(x+jr) + *(x+jm2r) );

/* Solve D(q,q)*x(q) = c(q) */

		    *(x+jm1r) = *(c)/( *(D+dp) );

		}

		if(jkmpm1 >= 2)
		{
		    *(c) = *(b+bp+jm1r) - *(F+dp)*( *(x+jm2r) );

/* Solve D(q,q)*x(q) = c(q) */

		    *(x+jm1r) = *(c)/( *(D+dp) );
		}
	}

	return 1;
}

int Cyclic_Reduction (double *D0, double *F0, double *B, double *x, int n2, int q)
{

/* For when the Tridiagonal block in A(n,n) is of any size

             Updated 5/3/06
*/

	int ii, jj, kk, i ,j, k, p, check, n, qsq;
	int bp, bpm1, bkm1, dp, dpm1, dkm1, xkm1,  r, r2;
	int jr, jrmr2, jrpr2, jkmp, jm1r, jm2r, jkmpm1;
	double *b, *D, *F, *mat_dum, *vec_dum, *c, fdum, fdum2;

	double d, *vv;
	int *indx;

	n = n2/q;
	qsq = q*q;
	fdum = (double)(n+1);
	fdum = log(fdum)/log(2.0);
	k = (int)fdum;

	printf( "k, n, q, sq %d %d %d %d\n\n", k, n, q, qsq);

	D = (double *)calloc(k*qsq,sizeof(double));
	F = (double *)calloc(k*qsq,sizeof(double));
	mat_dum = (double *)calloc(qsq,sizeof(double));
	b = (double *)calloc(k*n2,sizeof(double));
	vec_dum = (double *)calloc(q,sizeof(double));
	c = (double *)calloc(q,sizeof(double));

#if 1
	vv=(double *)malloc(q*sizeof(double));
	if(!vv) printf( "error allocating memory for vv\n");
	indx=(int *)malloc(q*sizeof(int));
	if(!indx) printf( "error allocating memory for index\n");
#endif

/* To maintain the same natation as the reference, I am using ii, jj, and kk
   as the dummy variables reserving j and k for the algorithm variables.

	printf( "\n\n");
	for( ii = 0; ii < n2; ++ii )
	{
	    for( jj = 0; jj < n2; ++jj )
	    {
		    printf( "%16.6e ", *(A + n2*ii + jj));
	    }
	    printf( "\n");
	}
*/

/* Calculate b(0), D(0), and F(0) by setting them to the tridiagonal block components
   of A.
*/
	for( ii = 0; ii < n2; ++ii )
	{
		*(b + ii) = *(B + ii);
		/*printf( "%16.6e\n", *(b + ii));*/
	}

	printf( "\n\n");
	for( ii = 0; ii < q; ++ii )
	{
		for( jj = 0; jj < q; ++jj )
		{
		    *(D + q*ii + jj) = *(D0 + q*ii + jj);
		    *(F + q*ii + jj) = *(F0 + q*ii + jj);
		    /*printf( "%4d %4d %16.6e %16.6e\n", ii, jj, *(D + q*ii + jj), *(F + q*ii + jj));*/
		}
	}

/* This is algorithm 4.5.14
*/
	for( p = 1; p <= k - 1; ++p )
	{
		r = pow(2, p);
		r2 = pow(2, p-1);
		dp = qsq*p;
		dpm1 = qsq*(p-1);
		bp = n2*p;
		bpm1 = n2*(p-1);

		check = matSquare( mat_dum, (F+dpm1), q);
		if(!check) printf( "Problems with matSquare\n");
		memcpy( (F+dp), mat_dum, qsq*sizeof(double));

		check = matSquare( mat_dum, (D+dpm1), q);
		if(!check) printf( "Problems with matSquare\n");
		for( ii = 0; ii < qsq; ++ii )
		{
		    *(D+dp+ii) = 2.0*( *(F+dp+ii) ) - *(mat_dum+ii);
		}

	    /*printf( "y1y1y1y1 %4d %4d\n", p, dp);
	    for( ii = 0; ii < q; ++ii )
	    {
		for( jj = 0; jj < q; ++jj )
		{
			printf( "yyyyyy %4d %4d %16.6e\n", ii, jj, *(D+dp + q*ii + jj));
		}
	    }*/

		jkmp = pow(2, k-p) - 1;
		for( j = 1; j <= jkmp; ++j )
		{
		   jr = q*(j*r - 1);
		   jrmr2 = q*(j*r - r2 - 1);
		   jrpr2 = q*(j*r + r2 - 1);

		   for( ii = 0; ii < q; ++ii )
		   {
			*(vec_dum+ii) = *(b+bpm1+jrmr2+ii) + *(b+bpm1+jrpr2+ii); 
		   }

		   for( ii = 0; ii < q; ++ii )
		   {
			check = dotX( &fdum, (F+dpm1+q*ii), (vec_dum), q);
			check = dotX( &fdum2, (D+dpm1+q*ii), (b+bpm1+jr), q);
			*(b+bp+jr+ii) = fdum - fdum2;
			/*printf( "fffff %4d %4d %4d %16.6e %16.6e\n", p, ii, j, fdum, fdum2);*/
		   }
		}
	}

/* This is algorithm 4.5.15
*/
	dkm1 = qsq*(k-1);
	bkm1 = n2*(k-1) + q*(pow(2, k-1) - 1);
	xkm1 = q*(pow(2, k-1) - 1);

	memcpy( c, (b+bkm1), q*sizeof(double));
	    /*for( ii = 0; ii < q; ++ii )
			printf( "fffff %4d %16.6e %16.6e\n", ii, *(b+bkm1+ii), *(c + ii));*/

	    printf( "\n\n");
#if !CFD_FLAG
	if( q == 2)
	{
		check = solve2X2( (D+dkm1), c, (x+xkm1) );
		if(!check) printf( " Problems with solve2X2\n");
	}

	if( q == 3)
	{
		check = solve3X3( (D+dkm1), c, (x+xkm1) );
		if(!check) printf( " Problems with solve3X3\n");
	}

	if( q > 3 || q < 2)
#endif
	{
		check = decomp( (D+dkm1), q );
		/*check=ludcmp((D+dkm1), q, indx, &d, vv);*/
	    /*for( ii = 0; ii < q; ++ii )
			printf( "iiii %4d %16.6e %16.6e\n", ii, *(b+bkm1+ii), *(c + ii));*/
		check = backsub( (D+dkm1), q, c );
	    /*for( ii = 0; ii < q; ++ii )
	    {
		printf( "f1f1f1 %4d %4d %16.6e\n", dkm1, ii, *(c + ii));
		for( jj = 0; jj < q; ++jj )
		{
			printf( "fffff %4d %4d %16.6e\n", ii, jj, *(D+dkm1+q*ii+jj));
		}
	    }*/
		/*check=lubksb((D+dkm1), q, indx, c);*/
		memcpy( (x+xkm1), c, q*sizeof(double));
	}

	for( p = k-2; p >= 0; --p )
	{
		dp = qsq*p;
		bp = n2*p;

/* Do LU decomposition over D(q,q) */

#if !CFD_FLAG
		if( q > 3 || q < 2)
#endif
			check = decomp( (D+dp), q );

		r = pow(2, p);

/* I break the next loop of j in the algorithm into 3 parts:

      1) j = 1
      2) 2 = j < 2^(k-p-1)
      3) j = 2^(k-p-1)

*/
		j = 1;
		jr = q*(2*j*r - 1);
		jm1r = q*((2*j-1)*r - 1);
		jm2r = q*((2*j-2)*r - 1);

		for( ii = 0; ii < q; ++ii )
		{
		    check = dotX( &fdum, (F+dp+q*ii), (x+jr), q);
		    *(c+ii) = *(b+bp+jm1r+ii) - fdum;
			/*printf( "fffff %4d %4d %4d %16.6e %16.6e %16.6e\n", p, ii, j, fdum, *(F+dp+q*ii), *(b+bp+jm1r+ii));*/
		}

/* Solve D(q,q)*x(q) = c(q) */

#if !CFD_FLAG
		if( q == 2)
		{
		    check = solve2X2( (D+dp), c, (x+jm1r) );
		    if(!check) printf( " Problems with solve2X2\n");
		}

		if( q == 3)
		{
		    check = solve3X3( (D+dp), c, (x+jm1r) );
		    if(!check) printf( " Problems with solve3X3\n");
		}

		if( q > 3 || q < 2)
#endif
		{
		    check = backsub( (D+dp), q, c );
		    memcpy( (x+jm1r), c, q*sizeof(double));
		}

		jkmpm1 = pow(2, k-p-1);

		for( j = 2; j < jkmpm1; ++j )
		{
		    jr = q*(2*j*r - 1);
		    jm1r = q*((2*j-1)*r - 1);
		    jm2r = q*((2*j-2)*r - 1);

		    for( ii = 0; ii < q; ++ii )
		    {
			*(vec_dum+ii) =  *(x+jr+ii) + *(x+jm2r+ii);
		    }

		    for( ii = 0; ii < q; ++ii )
		    {
			check = dotX( &fdum, (F+dp+q*ii), vec_dum, q);
			*(c+ii) = *(b+bp+jm1r+ii) - fdum;
		    }

/* Solve D(q,q)*x(q) = c(q) */

#if !CFD_FLAG
		    if( q == 2)
		    {
			check = solve2X2( (D+dp), c, (x+jm1r) );
			if(!check) printf( " Problems with solve2X2\n");
		    }

		    if( q == 3)
		    {
			check = solve3X3( (D+dp), c, (x+jm1r) );
			if(!check) printf( " Problems with solve3X3\n");
		    }

		    if( q > 3 || q < 2)
#endif
		    {
			check = backsub( (D+dp), q, c );
			memcpy( (x+jm1r), c, q*sizeof(double));
		    }

		}

		if(jkmpm1 >= 2)
		{
		    j = jkmpm1;
		    jr = q*(2*j*r - 1);
		    jm1r = q*((2*j-1)*r - 1);
		    jm2r = q*((2*j-2)*r - 1);

		    for( ii = 0; ii < q; ++ii )
		    {
			check = dotX( &fdum, (F+dp+q*ii), (x+jm2r), q);
			*(c+ii) = *(b+bp+jm1r+ii) - fdum;
		    }

/* Solve D(q,q)*x(q) = c(q) */

#if !CFD_FLAG
		    if( q == 2)
		    {
			check = solve2X2( (D+dp), c, (x+jm1r) );
			if(!check) printf( " Problems with solve2X2\n");
		    }

		    if( q == 3)
		    {
			check = solve3X3( (D+dp), c, (x+jm1r) );
			if(!check) printf( " Problems with solve3X3\n");
		    }

		    if( q > 3 || q < 2)
#endif
		    {
			check = backsub( (D+dp), q, c );
			memcpy( (x+jm1r), c, q*sizeof(double));
		    }
		}
	}

	free(D);
	free(F);
	free(mat_dum);
	free(b);
	free(vec_dum);
	free(c);

	return 1;
}

