/*
    This function calculates the determinant of a matrix A
    as well as solves the system:

          A*x = b; 

    for A(3x3) and A(2x2)

    SLFCFD source file
    Version:  1.0
    Copyright (C) 2006  San Le 

    The source code contained in this file is released under the
    terms of the GNU Library General Public License.
 
*/

#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#define SMALL       1.0e-20

double determinant3X3( double *A)
{
/*
    This function calculates the determinant of a 3x3 matrix.

	Updated 4/24/06
*/

	int i,j,k;
	double temp[3], det[1];

	*(temp) = *(A+4)*(*(A+8)) - *(A+7)*(*(A+5));
	*(temp+1) = *(A+6)*(*(A+5)) - *(A+3)*(*(A+8));
	*(temp+2) = *(A+3)*(*(A+7)) - *(A+6)*(*(A+4));
	*(det) = *(A)*(*(temp)) + *(A+1)*(*(temp+1)) + *(A+2)*(*(temp+2));

	/*printf("determinant of A %f\n", det);*/
	if( fabs( *(det)) < SMALL )
	{
		printf("Singular Matrix\n");
		return 0;
	}

	return *(det);

}

int solve2X2( double *A, double *b, double *x )
{
/*
    This function solves a 2x2 linear system:

        A*x = b

    using Cramer's rule.  It was taken from Calculus and Analytic Geometry, 4th Ed., 
    by Al Shenk, page 664.

	Updated 4/24/06
*/
	int i,j,k;
	double det[3];

	*(det) = *(A)*(*(A+3)) - *(A+1)*(*(A+2));

	/*printf("determinant of A %f\n", *(det));*/
	if( fabs( *(det)) < SMALL )
	{
		printf("Singular Matrix\n");
		return 0;
	}

	*(det+1) = *(b)*(*(A+3)) - *(A+1)*(*(b+1));
	*(det+2) = *(A)*(*(b+1)) - *(b)*(*(A+2));

	*(x) = *(det+1)/( *(det));
	*(x+1) = *(det+2)/( *(det));

	return 1;

}


int solve3X3( double *A, double *b, double *x )
{
/*
    This function solves a 3x3 linear system:

        A*x = b

    using Cramer's rule.  It was taken from Calculus and Analytic Geometry, 4th Ed., 
    by Al Shenk, page 664.

	Updated 4/24/06
*/
	int i,j,k;
	double temp[3], det[4];

	*(temp) = *(A+4)*(*(A+8)) - *(A+7)*(*(A+5));
	*(temp+1) = *(A+6)*(*(A+5)) - *(A+3)*(*(A+8));
	*(temp+2) = *(A+3)*(*(A+7)) - *(A+6)*(*(A+4));
	*(det) = *(A)*(*(temp)) + *(A+1)*(*(temp+1)) + *(A+2)*(*(temp+2));

	/*printf("determinant of A %f\n", *(det));*/
	if( fabs( *(det)) < SMALL )
	{
		printf("Singular Matrix\n");
		return 0;
	}

	*(temp) = *(A+4)*(*(A+8)) - *(A+7)*(*(A+5));
	*(temp+1) = *(b+2)*(*(A+5)) - *(b+1)*(*(A+8));
	*(temp+2) = *(b+1)*(*(A+7)) - *(b+2)*(*(A+4));
	*(det+1) = *(b)*(*(temp)) + *(A+1)*(*(temp+1)) + *(A+2)*(*(temp+2));

	*(temp) = *(b+1)*(*(A+8)) - *(b+2)*(*(A+5));
	*(temp+1) = *(A+6)*(*(A+5)) - *(A+3)*(*(A+8));
	*(temp+2) = *(A+3)*(*(b+2)) - *(A+6)*(*(b+1));
	*(det+2) = *(A)*(*(temp)) + *(b)*(*(temp+1)) + *(A+2)*(*(temp+2));

	*(temp) = *(A+4)*(*(b+2)) - *(A+7)*(*(b+1));
	*(temp+1) = *(A+6)*(*(b+1)) - *(A+3)*(*(b+2));
	*(temp+2) = *(A+3)*(*(A+7)) - *(A+6)*(*(A+4));
	*(det+3) = *(A)*(*(temp)) + *(A+1)*(*(temp+1)) + *(b)*(*(temp+2));

	*(x) = *(det+1)/( *(det));
	*(x+1) = *(det+2)/( *(det));
	*(x+2) = *(det+3)/( *(det));

	return 1;

}


