/* This program uses the finite difference method to calculate a compressible
   CFD flow field.  It reads in the data and calculates the various
   quantities for each point in the rectangular CFD mesh.

   The algorithm for the code below(predictor/corrector) is based on the
   Forney-Lax-Wendroff 2-step.  The main reference is:

Forney, L. J., Particle Impaction on Axially Symmetric Supersonic Flow, Journal of
   Aerosol Science and Technology 15, page 49, 1991.

It is similar to the MacCormack 2-Step seen on page 261 of:

Hirsch, C., Numerical Computation of Internal and External Flows Vol. 2: Computational
   Methods for Inviscid and Viscous Flows, John Wiley & Sons, Chichester, 1994. 

This same reference on page 135 gives a good discussion of the conservative form of
Euler's Equations.

I have extrapolated the 2-D version to 3-D.

                     Updated 1/7/10

    SLFCFD source file
    Version:  1.2
    Copyright (C) 2005-2010  San Le

    The source code contained in this file is released under the
    terms of the GNU Library General Public License.

*/

#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include "cfdconst.h"
#include "cfdstruct.h"
#include "../AMR/amr.h"

#define SMALL               1.0e-20

int GnuplotWriter( BODY *, char *, FLOWLINES, FLOWLINES, FLOWLINES );

int ConnectSurfwriter ();

int writer_SLFFEA_3D( BOUND , MATL *, COM_NODE *);

int writer_3D( BOUND , MATL *, char *, COM_NODE *);

int writer_SLFFEA_2D( BOUND , MATL *, COM_NODE *);

int writer_2D( BOUND , MATL *, char *, COM_NODE *);

int pathlines(int , FLOWLINES ,  COM_NODE * );

int streaklines(int , FLOWLINES ,  COM_NODE * );

int flwReset( COM_NODE *node);

int applyBC_noslip( BOUND , MATL *, COM_NODE *);

int applyBC( BOUND , MATL *, COM_NODE *);

int flow_anim_write( int , int , ANIM_DAT *, FLOWLINES , GR_NODE * );

int flow_anim_files( ANIM_DAT *, int , char *, FLOWLINES , FLOWLINES , FLOWLINES  );

int InitCalc (COM_NODE *node);

int ForneyLaxWendroffPredictor ( BOUND , MATL *, COM_NODE *);

int ForneyLaxWendroffCorrector ( BOUND , MATL *, COM_NODE *);

int create_body( BODY *);

int bcset_noslip( BOUND *, BODY *, int *, COM_NODE * );

int bcset( BOUND *, int *, COM_NODE * );

int flwreader( MATL *, char *, COM_NODE *, FILE *, FLOWLINES, FLOWLINES, FLOWLINES );

int cfdMemory( double **, int , int **, int , MATL **, int , XYZF **, int , XYZI **, int ,
	XYZI2 **, int , XY_ZX_YZ **, int , COM_NODE **, int, SUBGRID **, int );

int Flowlines_File( FLOWLINES * );

int parse_options(int , char **, ANIM_DAT *, char *, int );

int dof, nmat, numel, numnp, sof, nx, ny, nz, num_bodies;
double dx, dy, dz;
int flag_2D = 1, flag_3D = 0, print_flag = 0, debug_flag = 0, reverse_flag = 0, codetype_flag = 1;
int incompressible_flag = 0;
int mach_read_flag;
double tolerance;
double dt, time_current;	 /* time step for time integration */
double err = 0.0;                /* err is not needed for compressible calculations, but is
                                   declared as a global in other files. */
int nstep, cstep;
int anim_counter;
int imax, jmax;
int stream_nstep = 2500;

int main (int argc, char **argv)
{

	int i, j, k, dum, node_num, bod_num;
	int check, name_length, counter, pathline_counter, MemoryCounter, wait_flag;
	COM_NODE *node, *mem_node;
	GR_NODE *gr_node;
	SUBGRID *subgrid1, *subgrid2, *subgrid3, *mem_SUBGRID;
	XYZI *mem_XYZI;
	XYZI2 *mem_XYZI2;
	XYZF *mem_XYZF;
	XY_ZX_YZ *mem_XY_ZX_YZ;
	int *mem_int, sofmi, sofmf, sofmXYZF, sofmXYZI, sofmXYZI2, sofmXY_ZX_YZ,
		sofmCOM_NODE, sofmSUBGRID, ptr_inc;
	MATL *matl;
	double *mem_double;
	double time, time_final;
	double fdum;
	char name[30], buf[ BUFSIZ ], *ccheck, one_char;
	FILE *o1, *o_check;
	FILE *fddata;
	ANIM_DAT anim[1];
	BOUND bc;
	BODY *body;
	FLOWLINES streak, path, stream, flowline;
	long timec;
	long timef;

	double dns, dnsP, dnsC;
	double prs, prsP, prsC;
	double ene, eneP, eneC;
	double temp, tempP, tempC;
	double enthal, enthalP, enthalC;
	double vx, vy, vz, vxP, vyP, vzP, vxC, vyC, vzC;
	double mach;
	double gam, R;
	double vol;      /* volume of a cell = dx*dy*dz */
	int imn, imx;    /* range of values for the i=index */
	int jmn, jmx;    /* range of values for the j=index */
	int kmn, kmx;    /* range of values for the k=index */

	memset(name,0,30*sizeof(char));
	memset(anim[0].node_name,0,30*sizeof(char));
	memset(anim[0].flowline_name,0,30*sizeof(char));
	memset(anim[0].color_name,0,30*sizeof(char));
	memset(anim[0].flowtype_name,0,30*sizeof(char));
	memset(anim[0].velocity_name,0,30*sizeof(char));

/*   o1 contains all the structural data
     streak.file_in is the initial points of the streaklines
     streak.file_out is the initial points of the streaklines
     path.file_in is the initial points of the streaklines
     path.file_out is the initial points of the streaklines
     stream.file_in is the initial points of the streamlines
     stream.file_out is the initial points of the streamlines
*/

/* Check the command line options passed by the reader. */

	codetype_flag = 1;
	check = parse_options( argc , argv, anim, name, codetype_flag );

	name_length = strlen(name);
	if( name_length > 25) name_length = 25;

	o1 = fopen( name,"r" );
	if(o1 == NULL ) {
		printf("Can't find file %30s\n", name);
		exit(1);
	}

	fgets( buf, BUFSIZ, o1 );
	fscanf( o1, "%d %d %d", &numnp, &nx, &ny);
	imax = nx;
	jmax = ny;
	nmat = 1;

/* Check if there is additional coordinate z data and that this is a 3-D file.  */

	flag_3D = 0;
	flag_2D = 1;
	nz = 1;
	while(( one_char = (unsigned char) fgetc(o1)) != '\n')
	{
		if(one_char != ' ' )
		{
		    ungetc( one_char, o1);
		    fscanf( o1,"%d", &nz);
		    flag_3D = 1;
		    flag_2D = 0;
		    break;
		}
	}
	numnp = nx*ny*nz;
	dof=numnp;
	fscanf( o1,"\n");


	if(flag_2D)
	{
	    name_length = strlen(name);
	    if( name_length > 25) name_length = 25;

/* Set the name of the input pathline file */

	    memset(path.name_in,0,30*sizeof(char));

	    ccheck = strncpy(path.name_in, name, name_length);
	    if(!ccheck) printf( " Problems with strncpy \n");

	    ccheck = strncpy(path.name_in+name_length, ".path", 5);
	    if(!ccheck) printf( " Problems with strncpy \n");

/* Check for the existence of the pathline input file as well as its properties */

	    path.read_flag = 0;
	    check = Flowlines_File( &path );
	    if(!ccheck) printf( " Problems with Flowlines_File \n");

/* Set the name of the input streakline file */

	    memset(streak.name_in,0,30*sizeof(char));

	    streak.read_flag = 0;
	    ccheck = strncpy(streak.name_in, name, name_length);
	    if(!ccheck) printf( " Problems with strncpy \n");

	    ccheck = strncpy(streak.name_in+name_length, ".streak", 7);
	    if(!ccheck) printf( " Problems with strncpy \n");

/* Check for the existence of the streakline input file as well as its properties */

	    check = Flowlines_File( &streak );
	    if(!ccheck) printf( " Problems with Flowlines_File \n");

/* Set the name of the input streamline file */

	    memset(stream.name_in,0,30*sizeof(char));

	    stream.read_flag = 0;
	    ccheck = strncpy(stream.name_in, name, name_length);
	    if(!ccheck) printf( " Problems with strncpy \n");

	    ccheck = strncpy(stream.name_in+name_length, ".stream", 7);
	    if(!ccheck) printf( " Problems with strncpy \n");

/* Check for the existence of the streamline input file as well as its properties */

	    check = Flowlines_File( &stream );
	    if(!ccheck) printf( " Problems with Flowlines_File \n");

	}

	if(flag_3D) printf( "total nodes:%d   nodes in x:%d nodes in y:%d nodes in z:%d\n", numnp, nx, ny, nz);
	else        printf( "total nodes:%d   nodes in x:%d nodes in y:%d\n", numnp, nx, ny);
	fgets( buf, BUFSIZ, o1 );
	printf( "\n");

	fscanf( o1, "%lf %d %d %lf\n", &dt, &nstep, &cstep, &time_current);
	if( nstep < cstep)    
	{
	    printf("Number of steps must be equal or greater than current step.\n");
	    exit(1);
	}
	if( nstep == 0)
	{
	    printf("Number of steps cannot be zero.\n");
	    exit(1);
	}

	printf( "Time step:     %16.8e\n", dt);
	printf( "number of steps:  %4d\n", nstep);
	printf( "current step:     %4d\n", cstep);
	printf( "current time:  %16.8e\n", time_current);
	printf( "\n");

	fgets( buf, BUFSIZ, o1 );
	printf( "\n");

	if( stream_nstep < nstep) stream_nstep = nstep;
	if(flag_2D)
	{
	    path.nstep_out = nstep;
	    streak.nstep_out = nstep;
	    stream.nstep_out = stream_nstep;
	}

/*   Begin allocation of meomory */

	MemoryCounter = 0;

/* For the doubles */
	sofmf=1;
	MemoryCounter += sofmf*sizeof(double);
	printf( "\n Memory requrement for doubles is %15d bytes\n",MemoryCounter);

/* For the integers */
	sofmi= 2 + 4*numnp;
	MemoryCounter += sofmi*sizeof(int);
	printf( "\n Memory requrement for integers is %15d bytes\n",MemoryCounter);

/* Note that I allocate 2*numnp for the nodal boundaries rather than just numnp 
   because some nodes may have more than 1 boundary condition, specifically the
   no-slip corner boundary which results in a node being no-slip in more than 1
   direction.  So I can have, for instance:

      bc[0].noslp.xp[3] = 5;
      bc[0].noslp.yp[3] = 5;

   This potential for having more than one boundary condition is why I have allocated
   2*numnp.
*/

/* For the XYZF doubles */
	sofmXYZF = streak.num_pt0 + streak.num_pt0*(streak.nstep_out+1) +
		path.num_pt0 + path.num_pt0*(path.nstep_out+1) +
		stream.num_pt0 + stream.num_pt0*(stream.nstep_out+1);
	MemoryCounter += sofmXYZF*sizeof(XYZF);
	printf( "\n Memory requrement for XYZF integers is %15d bytes\n",MemoryCounter);

/* For the XYZI integers */
	sofmXYZI = 1;
	MemoryCounter += sofmXYZI*sizeof(XYZI);
	printf( "\n Memory requrement for XYZI integers is %15d bytes\n",MemoryCounter);

/* For the XYZI2 integers */
	sofmXYZI2=7;
	MemoryCounter += sofmXYZI2*sizeof(XYZI2);
	printf( "\n Memory requrement for XYZI2 integers is %15d bytes\n",MemoryCounter);

/* For the XY_ZX_YZ integers */
	sofmXY_ZX_YZ=1;
	MemoryCounter += sofmXY_ZX_YZ*sizeof(XY_ZX_YZ);
	printf( "\n Memory requrement for XY_ZX_YZ integers is %15d bytes\n",MemoryCounter);

/* For the COM_NODE structs */
	sofmCOM_NODE=numnp + 1;
	MemoryCounter += sofmCOM_NODE*sizeof(COM_NODE);
	printf( "\n Memory requrement for COM_NODE integers is %15d bytes\n",MemoryCounter);

/* For the SUBGRID structs */
	sofmSUBGRID=3;
	MemoryCounter += sofmSUBGRID*sizeof(SUBGRID);
	printf( "\n Memory requrement for SUBGRID integers is %15d bytes\n",MemoryCounter);

	check = cfdMemory( &mem_double, sofmf, &mem_int, sofmi, &matl, nmat, &mem_XYZF, sofmXYZF,
		&mem_XYZI, sofmXYZI, &mem_XYZI2, sofmXYZI2, &mem_XY_ZX_YZ, sofmXY_ZX_YZ, &node,
		sofmCOM_NODE, &mem_SUBGRID, sofmSUBGRID );
	if(!check) printf( " Problems with cfdMemory\n");

/* For the GR_NODE structs when an animation is to be made */

	if(anim[0].flag) gr_node = (GR_NODE *)calloc(numnp,sizeof(GR_NODE));

/* For the integers */
	                                      ptr_inc = 0;
	bc.num_none=(mem_int+ptr_inc);        ptr_inc += 1;
	bc.num_prs0=(mem_int+ptr_inc);        ptr_inc += 1;

/* Note that bcset below has to be corrected for mem_int so that it has the correct
   offset for mem_int, i.e.:

	check = bcset( &bc, (mem_int+2), node);

	check = bcset_noslip( &bc, body, (mem_int+2+2*numnp), node);
*/

/* For the XYZF doubles */
	                                      ptr_inc = 0;
	streak.pt0=(mem_XYZF+ptr_inc);        ptr_inc += streak.num_pt0;
	streak.pt=(mem_XYZF+ptr_inc);         ptr_inc += streak.num_pt0*(streak.nstep_out+1);
	path.pt0=(mem_XYZF+ptr_inc);          ptr_inc += path.num_pt0;
	path.pt=(mem_XYZF+ptr_inc);           ptr_inc += path.num_pt0*(path.nstep_out+1);
	stream.pt0=(mem_XYZF+ptr_inc);        ptr_inc += stream.num_pt0;
	stream.pt=(mem_XYZF+ptr_inc);         ptr_inc += stream.num_pt0*(stream.nstep_out+1);


/* For the XYZI integers */
	                                      ptr_inc = 0;
	bc.num_vel0=(mem_XYZI+ptr_inc);       ptr_inc += 1;

/* For the XYZI2 integers */
	                                      ptr_inc = 0;
	bc.num_symm=(mem_XYZI2+ptr_inc);      ptr_inc += 1;
	bc.num_far=(mem_XYZI2+ptr_inc);       ptr_inc += 1;
	bc.num_free=(mem_XYZI2+ptr_inc);      ptr_inc += 1;
	bc.num_inflow=(mem_XYZI2+ptr_inc);    ptr_inc += 1;
	bc.num_frestr=(mem_XYZI2+ptr_inc);    ptr_inc += 1;
	bc.num_free2=(mem_XYZI2+ptr_inc);     ptr_inc += 1;
	bc.num_noslp=(mem_XYZI2+ptr_inc);     ptr_inc += 1;

/* For the XY_ZX_YZ integers */
	                                          ptr_inc = 0;
	bc.num_noslp_prs=(mem_XY_ZX_YZ+ptr_inc);  ptr_inc += 1;

/* For the SUBGRID structs */
	                                      ptr_inc = 0;
	subgrid1=(mem_SUBGRID+ptr_inc);       ptr_inc += 1;
	subgrid2=(mem_SUBGRID+ptr_inc);       ptr_inc += 1;
	subgrid3=(mem_SUBGRID+ptr_inc);       ptr_inc += 1;

	timec = clock();
	timef = 0;

	check = flwreader( matl, name, node, o1, path, streak, stream);
	if(!check) printf( " Problems with flwreader\n");

	gam = matl[0].gam;
	R = matl[0].R;

	/*for( i = 0; i < nx; ++i) {
	   for( j = 0; j < ny; ++j) {
	     for( k = 0; k < nz; ++k) {
		node_num = ny*nz*i + nz*j + k;
		printf("%7d %7d %7d %7d %7d\n",
			node[node_num].i, node[node_num].j, node[node_num].k, node_num, node[node_num].bc_type);
	     }
	   }
	}*/

/* Set the boundary condition variable, bc, as well as allocate memory for each individual
   boundary condition
*/

	check = bcset( &bc, (mem_int+2), node);
	if(!check) printf( " Problems with bcset\n");

	if(flag_2D)
	{
	    for( i = 0; i < streak.num_pt0; ++i )
	    {
		if(print_flag) printf(  "streak %4d   %16.8e %16.8e\n", i, streak.pt0[i].x, streak.pt0[i].y);
	    }

	    for( i = 0; i < path.num_pt0; ++i )
	    {
/* Set the initial pathline values. */

		if(print_flag) printf(  "path %4d   %16.8e %16.8e\n", i, path.pt0[i].x, path.pt0[i].y);
		path.pt[i].x = path.pt0[i].x;
		path.pt[i].y = path.pt0[i].y;

	    }

	    for( i = 0; i < stream.num_pt0; ++i )
	    {
/* Set the initial streamline values. */

		if(print_flag) printf(  "stream %4d   %16.8e %16.8e\n", i, stream.pt0[i].x, stream.pt0[i].y);
		stream.pt[i].x = stream.pt0[i].x;
		stream.pt[i].y = stream.pt0[i].y;

	    }

/* Allocate memory for the bodies in the flow.  */

	    if(num_bodies)
	    {
		body = (BODY *)calloc(num_bodies,sizeof(BODY));
		if(!body )
		{
		    printf( "failed to allocate memory for body structures\n ");
		    exit(1);
		}
		dum = 2*bc.num_noslp[0].xm; /* twice the amount necessary, just in case. */
		for( bod_num = 0; bod_num < num_bodies; ++bod_num )
		{
		    body[bod_num].nd = (J_I1_I2_INDX *)calloc(dum,sizeof(J_I1_I2_INDX));
		    body[bod_num].coord = (double *)calloc(2*dum*nsd,sizeof(double));
		}
	    }
	    check = bcset_noslip( &bc, body, (mem_int+2+2*numnp), node);
	    if(!check) printf( " Problems with bcset_noslip\n");

/* This is only used for generating points which draw the body.  It is not needed for calculation. */

	    if(num_bodies)
	    {
		check = create_body( body);
		if(!check) printf( " Problems with create_body\n");
	    }
	}


/*
	for( i = 0; i < bc.num_none[0]; ++i) {
		printf("%d\n", bc.none[i]);
	}
	for( i = 0; i < bc.num_symm[0].xp; ++i) {
		printf("%d\n", bc.symm.xp[i]);
	}
	for( i = 0; i < bc.num_symm[0].ym; ++i) {
		printf("%d\n", bc.symm.ym[i]);
	}
*/

/* Initialize flow field */

	check = InitCalc( node);
	if(!check) printf( " Problems with InitCalc\n");

/* Begin calculation of flow over time */

	time_final = ((double)nstep)*dt;

	time = ((double)cstep)*dt;
	counter = cstep;
	anim_counter = 0;
	if(counter == 0) pathline_counter = counter;
	else pathline_counter = counter-1;

	if( anim[0].flag)
	{

/* Read in data from "fdview.anim" if it exists. This contains the data for the desired
   quantities to display.
*/
	    check = flow_anim_files( anim, codetype_flag, name, path, streak, stream);

	    for( i = 0; i < nx; ++i )
	    {
		for( j = 0; j < ny; ++j )
		{
		    node_num = ny*i + j;

		    gr_node[node_num].i = i;
		    gr_node[node_num].j = j;

		    gr_node[node_num].bc_type = node[node_num].bc_type;
		    gr_node[node_num].nsid    = node[node_num].nsid;
		}
	    }

/* Open a file containing the initial node data which is needed to initialize
   the graphics.
*/
	    ccheck = strncpy(anim[0].node_name, "anim_node", 9);
	    anim[0].node_out = fopen( anim[0].node_name, "w" );
	    fwrite( gr_node, sizeof(GR_NODE), numnp, anim[0].node_out);
	    fclose( anim[0].node_out );              /* Close animation file */

/* Modify the names for animation input and output.  */

	    ccheck = strncpy(anim[0].node_name, "anim_node0", 10);
	    ccheck = strncpy(anim[0].flowline_name, "anim_flowline0", 14);

	}


	while (time < time_final - dt/2.0 )
	{
		check = ForneyLaxWendroffPredictor( bc, matl, node);
		if(!check) printf( " Problems with ForneyLaxWendroffPredictor\n");

		check = applyBC( bc, matl, node);
		if(!check) printf( " Problems with applyBC\n");

		check = applyBC_noslip( bc, matl, node);
		if(!check) printf( " Problems with applyBC_noslip\n");

		check = ForneyLaxWendroffCorrector( bc, matl, node);
		if(!check) printf( " Problems with ForneyLaxWendroffCorrector\n");

		check = applyBC( bc, matl, node);
		if(!check) printf( " Problems with applyBC\n");

		check = applyBC_noslip( bc, matl, node);
		if(!check) printf( " Problems with applyBC_noslip\n");

		check = flwReset( node);
		if(!check) printf( " Problems with flwReset\n");


		if(flag_2D)
		{
/* Update the value of all the streaklines points. */

		    if(streak.read_flag) check = streaklines( counter, streak, node);

/* Calculate the next point of the pathlines. */

		    if(path.read_flag) check = pathlines( pathline_counter, path, node);
		}

		time += dt;
		++counter;
		++pathline_counter;
		printf( " Time: %16.8e  step: %5d\n", time, counter);

/* The block of code below pretains to the animation.

         ------------ ANIMATION BEGIN ------------
*/

		if( anim[0].flag)
		{
		    for( i = 0; i < nx; ++i )
		    {
			for( j = 0; j < ny; ++j )
			{
			    node_num = ny*i + j;
			    gr_node[node_num].dns     =    dns = node[node_num].dns;
			    gr_node[node_num].temp    =   temp = node[node_num].temp;
			    gr_node[node_num].ene     =    ene = node[node_num].ene;
			    gr_node[node_num].enthal  = enthal = node[node_num].enthal;
			    gr_node[node_num].vx      =     vx = node[node_num].vx;
			    gr_node[node_num].vy      =     vy = node[node_num].vy;
			    gr_node[node_num].prs     =    prs = node[node_num].prs;

			    fdum = (vx*vx + vy*vy);
			    gr_node[node_num].vmag = sqrt(fdum);
			    gr_node[node_num].mach = gr_node[node_num].vmag/(R*gam*temp + SMALL);
			}
		    }

		    if( anim[0].flowtype_choice )
		    {
			if( anim[0].flowtype_choice == 1 ) flowline = path;
			if( anim[0].flowtype_choice == 2 ) flowline = streak;
			if( anim[0].flowtype_choice == 3 )
			{
			    if(stream.read_flag) check = streamlines( stream_nstep, stream, node);
			    flowline = stream;
			}
		    }

		    check = flow_anim_write( anim_counter, counter, anim, flowline, gr_node );
		    ++anim_counter;
		    if( anim_counter > 9) anim_counter = 0;

		}
/* The big block of code below pretains to the animation.

         ------------ ANIMATION END ------------
*/

	}

/* Calculate the streamlines at this point in time. */

	if(flag_2D) if(stream.read_flag) check = streamlines( stream_nstep, stream, node);

	cstep = counter;
	time_current = dt*(double)(counter);

/* Write out the streakline and obstruction data to a file which can be read
   into "gnuplot".
*/

	if(flag_2D)
	{
	    check = GnuplotWriter( body, name, path, streak, stream);
	    if(!check) printf( " Problems with GnuplotWriter\n");
	}

/* Calculate Mach Number */

	for ( node_num = 0; node_num < numnp; node_num++)
	{
		temp   = node[node_num].temp;
		vx     = node[node_num].vx;
		vy     = node[node_num].vy;
		vz     = node[node_num].vz;

		fdum = (vx*vx + vy*vy + vz*vz);
		node[node_num].vmag = sqrt(fdum);
		fdum = (vx*vx + vy*vy + vz*vz)/(R*gam*temp + SMALL);
		node[node_num].mach = sqrt(fdum);
	}


	if(flag_2D)
	{	
		check = writer_2D(bc, matl, name, node);
		if(!check) printf( " Problems with writer_2D\n");
#if 0
		check = writer_SLFFEA_2D(bc, matl, node);
		if(!check) printf( " Problems with writer_SLFFEA_2D\n");
#endif
	}

	if(flag_3D)
	{	
		check = writer_3D(bc, matl, name, node);
		if(!check) printf( " Problems with writer_3D\n");

		check = ConnectSurfwriter();
		if(!check) printf( " Problems with ConnectSurfwriter\n");
#if 0
		check = writer_SLFFEA_3D(bc, matl, node);
		if(!check) printf( " Problems with writer_SLFFEA_3D\n");
#endif
	}

	timec = clock();
	printf("\n\n elapsed CPU = %lf\n\n",( (double)timec)/800.);

	free(matl);
	free(mem_double);
	free(mem_int);
	free(mem_XYZF);
	free(mem_XYZI);
	free(mem_XYZI2);
	free(mem_XY_ZX_YZ);
	free(node);
	free(mem_SUBGRID);

	return 1;
}
