/* This file contains the structures of the finite difference code for
   a compressible CFD flow field. 

                     Updated 3/31/08

    SLFCFD source file
    Version:  1.2
    Copyright (C) 2005, 2006, 2007, 2008  San Le

    The source code contained in this file is released under the
    terms of the GNU Library General Public License.

*/

#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "cfdconst.h"

typedef struct {
	int *xmym, *xpym, *xpyp, *xmyp;
	int *zmxm, *zpxm, *zpxp, *zmxp;
	int *ymzm, *ypzm, *ypzp, *ymzp;
	int *xmymzm, *xpymzm, *xpypzm, *xmypzm;
	int *xmymzp, *xpymzp, *xpypzp, *xmypzp;
} XY_ZX_YZP;

typedef struct {
	int xmym, xpym, xpyp, xmyp;
	int zmxm, zpxm, zpxp, zmxp;
	int ymzm, ypzm, ypzp, ymzp;
	int xmymzm, xpymzm, xpypzm, xmypzm;
	int xmymzp, xpymzp, xpypzp, xmypzp;
} XY_ZX_YZ;

typedef struct {
	int *xp, *xm, *yp, *ym, *zp, *zm;
} XYZI2P;

typedef struct {
	int xp, xm, yp, ym, zp, zm;
} XYZI2;

typedef struct {
	double x, y, z;
} XYZF;

typedef struct {
	int *x, *y, *z;
} XYZIP;

typedef struct {
	int x, y, z;
} XYZI;

typedef struct {
/* This NODE structure is for the compressible code.  */
	int i, j, k;
	double dns, prs, temp, enthal, ene, vx, vy, vz, vmag;
	double dnsP, prsP, tempP, enthalP, eneP, vxP, vyP, vzP;   /* Calculated from predictor step. */
	double dnsC, prsC, tempC, enthalC, eneC, vxC, vyC, vzC;   /* Calculated from corrector step. */
	double mach;
	double F[5], FP[5], FC[5], G[5], GP[5], GC[5], H[5], HP[5], HC[5], U[5], UP[5], UC[5];
	int bc_type;
	int refine_flag;
	int nsid; /* This gives the numerical ID of the body in the fluid. The default is zero.
	             For the first body, all of its associate noslip nodes are set to 1. For the
	             second body, its noslip nodes would be set to 2, and so on. */
} COM_NODE;

typedef struct {
/* This NODE structure is for the incompressible code.  */
	int i, j, k;
	double prs, vx, vy, vz, vmag;
	double vxQ, vyQ, vzQ;              /* Calculated from quickest algorithm. */
	double vx0, vy0, vz0;              /* Temporary values that are sometimes needed. */
	double vxFvx, vxFvy, vyFvy, vyFvx; /* Face velecities */
	int bc_type;
	int refine_flag;
	int nsid; /* This gives the numerical ID of the body in the fluid. The default is zero.
	             For the first body, all of its associate noslip nodes are set to 1. For the
	             second body, its noslip nodes would be set to 2, and so on. */
} INC_NODE;

typedef struct {
/* This NODE structure is specific to the graphics code. It contains all the variables of
   the compressible and incompressible NODE structures, but none of the extra temporary
   time step quantities.
*/
	int i, j, k;
	double dns, prs, temp, enthal, ene, vx, vy, vz, mach, vmag;
	int bc_type;
	int nsid; /* This gives the numerical ID of the body in the fluid. The default is zero.
	             For the first body, all of its associate noslip nodes are set to 1. For the
	             second body, its noslip nodes would be set to 2, and so on. */
} GR_NODE;


typedef struct {
/* This NODE structure contains all the variables of the compressible and incompressible
   NODE structures, including the extra temporary time step quantities.  I don't use it yet
   but have it in case I have to do a calculation that involves both types of flows.
*/
	int i, j, k;
	double dns, prs, temp, enthal, ene, vx, vy, vz, vmag;
	double vxQ, vyQ, vzQ;              /* Calculated from quickest algorithm. */
	double vx0, vy0, vz0;              /* Temporary values that are sometimes needed. */
	double vxFvx, vxFvy, vyFvy, vyFvx; /* Face velecities */
	double dnsP, prsP, tempP, enthalP, eneP, vxP, vyP, vzP;   /* Calculated from predictor step. */
	double dnsC, prsC, tempC, enthalC, eneC, vxC, vyC, vzC;   /* Calculated from corrector step. */
	double mach;
	double F[5], FP[5], FC[5], G[5], GP[5], GC[5], H[5], HP[5], HC[5], U[5], UP[5], UC[5];
	int bc_type;
	int refine_flag;
	int nsid; /* This gives the numerical ID of the body in the fluid. The default is zero.
	             For the first body, all of its associate noslip nodes are set to 1. For the
	             second body, its noslip nodes would be set to 2, and so on. */
} GR_NODE2;


typedef struct {
	int *num_none;
	XYZI2 *num_symm;
	XYZI2 *num_symmI;
	XYZI2 *num_far;
	XYZI2 *num_free;
	XYZI2 *num_inflow;
	XYZI2 *num_frestr;
	XYZI2 *num_free2;
	XYZI2 *num_noslp;
	int *num_prs0;
	XYZI *num_vel0;
	XY_ZX_YZ *num_noslp_prs;
	int *none;
	XYZI2P symm;
	XYZI2P symmI;
	XYZI2P far;
	XYZI2P free;
	XYZI2P inflow;
	XYZI2P frestr;
	XYZI2P free2;
	XYZI2P noslp;
	int *prs0;
	XYZIP vel0;   /* These are for nodes internal to an object where all veloctities are set to zero. */
	XY_ZX_YZP noslp_prs;  /* These are for pressure corner nodes internal to an object. */
} BOUND;

typedef struct {
	double R;          /* Ideal Gas Constant */
	double gam;        /* Ratio of specific heats */
	double Re;         /* Reynolds Number */
	double q;
	double alpha;
} MATL;

typedef struct {
	int j;
	int i1;
	int i2;
	int index;
} J_I1_I2_INDX;       /* These contain the indicies needed to record the parameters
                      of the body.  */

typedef struct {
	J_I1_I2_INDX *nd;
	double *coord;
	int num_points;
} BODY;            /* This structure is for recording the outline of a body in the flow.  It
                      is used to record the nodes and draw the body in a display.  */

typedef struct {
	XYZF *pt;
	XYZF *pt0;
	int num_pt0;
	int nstep_in;
	int nstep_out;
	FILE *file_in;
	FILE *file_out;
	char name_in[41];
	char name_out[41];
	int read_flag;
	int new_flag;
} FLOWLINES;


typedef struct {
	int flag;                /* Tells whether to make an animation */
	int save_flag;           /* Tells whether to save the animation */
	int color_choice;
	int flowtype_choice;
	int velocity_choice;
	int node_print_flag;

	FILE *node_in;
	FILE *node_out;
	char node_name[30];
	FILE *flowline_in;
	FILE *flowline_out;
	char flowline_name[30];
	FILE *name;
	char color_name[30];
	char flowtype_name[30];
	char velocity_name[30];
} ANIM_DAT;


