/* This program performs the predictor step for the program which
   uses the finite difference method to calculate a compressible CFD flow field.
   The algorithm for the code below is based on the Forney-Lax-Wendroff 2-step. 
   The main reference is:

Forney, L. J., Particle Impaction on Axially Symmetric Supersonic Flow, Journal of
   Aerosol Science and Technology 15, page 49, 1991.

It is similar to the MacCormack 2-Step seen on page 261 of:

Hirsch, C., Numerical Computation of Internal and External Flows Vol. 2: Computational
   Methods for Inviscid and Viscous Flows, John Wiley & Sons, Chichester, 1994. 

This same reference on page 135 gives a good discussion of the conservative form of
Euler's Equations.

I have extrapolated the 2-D version to 3-D.

                     Updated 2/2/06

    SLFCFD source file
    Version:  1.2
    Copyright (C) 2005, 2006, 2007, 2008  San Le

    The source code contained in this file is released under the
    terms of the GNU Library General Public License.

*/

#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "cfdconst.h"
#include "cfdstruct.h"

extern dof, numel, numnp, sof, nx, ny, nz;
extern int flag_2D, flag_3D;
extern double dt, dx, dy, dz;

int ForneyLaxWendroffPredictor ( BOUND bc, MATL *matl, COM_NODE *node)
{
/*

  Use the key below to understand the indexing.

	 ce: node (i, j, k )
	 xm1: node (i - 1, j, k)
	 xp1: node (i + 1, j, k)
	 ym1: node (i, j - 1, k)
	 yp1: node (i, j + 1, k)
	 zm1: node (i, j, k - 1)
	 zp1: node (i, j, k + 1)
*/

	int i, j, k, i2, dum, node_num;
	int ce, xp1, xm1, yp1, ym1, zp1, zm1;
	double dns, prs, temp, enthal, ene, vx, vy, vz;
	double dnsP, prsP, tempP, enthalP, eneP, vxP, vyP, vzP;
	double gam, R;

	double Fce[5], Fxm1[5], Fxp1[5], Fym1[5], Fyp1[5], Fzm1[5], Fzp1[5];
	double Gce[5], Gxm1[5], Gxp1[5], Gym1[5], Gyp1[5], Gzm1[5], Gzp1[5];
	double Hce[5], Hxm1[5], Hxp1[5], Hym1[5], Hyp1[5], Hzm1[5], Hzp1[5];
	double Uce[5], Uxm1[5], Uxp1[5], Uym1[5], Uyp1[5], Uzm1[5], Uzp1[5];

	double taux = dt/dx;
	double tauy = dt/dy;
	double tauz = dt/dz;
	double fdum;

	gam = matl[0].gam;
	R = matl[0].R;

	if (flag_3D)
	{
	    for ( dum=0; dum < bc.num_none[0]; dum++)
	    {
		node_num = bc.none[dum];
		i = node[node_num].i;
		j = node[node_num].j;
		k = node[node_num].k;
		ce  = nz*ny*i + nz*j + k;
		xm1 = nz*ny*(i-1) + nz*j + k;
		xp1 = nz*ny*(i+1) + nz*j + k;
		ym1 = nz*ny*i + nz*(j-1) + k;
		yp1 = nz*ny*i + nz*(j+1) + k;
		zm1 = nz*ny*i + nz*j + (k-1);
		zp1 = nz*ny*i + nz*j + (k+1);

		if( xm1 < 0 || ym1 < 0 || zm1 < 0)
		{
		    printf("There is a problem with your mesh.\n");
		    printf("One of the surrounding nodes in either the x-, y-, or z-\n");
		    printf("position is negative.\n");
		    exit(1);
		}

		fdum = PT166;
		//printf("rrrr %4d %4d %4d vz\n", i, j, k);

		for ( i2 = 0; i2 < 5; i2++)
		{
/*     Based on the calculation for UP[i2] below, I probably don't need all the positions for
       F, G, H, and U to be calculated.  Remember to comment them out later.
*/

#if 0
		    Fce[i2]  = node[ce].F[i2];   Gce[i2]  = node[ce].G[i2];   Hce[i2]  = node[ce].H[i2];
		    Fxm1[i2] = node[xm1].F[i2];  Gxm1[i2] = node[xm1].G[i2];  Hxm1[i2] = node[xm1].H[i2];
		    Fxp1[i2] = node[xp1].F[i2];  Gxp1[i2] = node[xp1].G[i2];  Hxp1[i2] = node[xp1].H[i2];
		    Fym1[i2] = node[ym1].F[i2];  Gym1[i2] = node[ym1].G[i2];  Hym1[i2] = node[ym1].H[i2];
		    Fyp1[i2] = node[yp1].F[i2];  Gyp1[i2] = node[yp1].G[i2];  Hyp1[i2] = node[yp1].H[i2];
		    Fzm1[i2] = node[zm1].F[i2];  Gzm1[i2] = node[zm1].G[i2];  Hzm1[i2] = node[zm1].H[i2];
		    Fzp1[i2] = node[zp1].F[i2];  Gzp1[i2] = node[zp1].G[i2];  Hzp1[i2] = node[zp1].H[i2];

		    Uce[i2]  = node[ce].U[i2]; 
		    Uxm1[i2] = node[xm1].U[i2];
		    Uxp1[i2] = node[xp1].U[i2];
		    Uym1[i2] = node[ym1].U[i2];
		    Uyp1[i2] = node[yp1].U[i2];
		    Uzm1[i2] = node[zm1].U[i2];
		    Uzp1[i2] = node[zp1].U[i2];
#endif

		    Fce[i2]  = node[ce].F[i2];   Gce[i2]  = node[ce].G[i2];   Hce[i2]  = node[ce].H[i2];
		    Fxp1[i2] = node[xp1].F[i2];
		                                 Gyp1[i2] = node[yp1].G[i2];
		                                                              Hzp1[i2] = node[zp1].H[i2];
		    Uxm1[i2] = node[xm1].U[i2];
		    Uxp1[i2] = node[xp1].U[i2];
		    Uym1[i2] = node[ym1].U[i2];
		    Uyp1[i2] = node[yp1].U[i2];
		    Uzm1[i2] = node[zm1].U[i2];
		    Uzp1[i2] = node[zp1].U[i2];

		    //UP[i2] = Uce[i2] - taux*(Fxp1[i2] - Fce[i2]) - tauy*(Gyp1[i2] - Gce[i2]) - tauz*(Hzp1[i2] - Hce[i2]);
		    //UP[i2] = PT166*(Uxm1[i2] + Uxp1[i2] + Uym1[i2] + Uyp1[i2] + Uzm1[i2] + Uzp1[i2]) -
		    //    taux*(Fxp1[i2] - Fce[i2]) - tauy*(Gyp1[i2] - Gce[i2]) - tauz*(Hzp1[i2] - Hce[i2]);

		    node[ce].UP[i2] = fdum*(Uxm1[i2] + Uxp1[i2] + Uym1[i2] + Uyp1[i2] + Uzm1[i2] + Uzp1[i2]) -
		            taux*(Fxp1[i2] - Fce[i2]) -
		            tauy*(Gyp1[i2] - Gce[i2]) -
		            tauz*(Hzp1[i2] - Hce[i2]);

		    //printf("rrrr vz %16.8e %16.8e %16.8e %16.8e %16.8e %16.8e %16.8e\n",
		    //        UP[i2], Uxm1[i2], Uxp1[i2], Uym1[i2], Uyp1[i2], Uzm1[i2], Uzp1[i2]);
		}

		//printf("rrrr vz %16.8e %16.8e\n", vx, vy);

		//printf("rrrr vz %16.8e %16.8e %16.8e %16.8e %16.8e %16.8e %16.8e\n",
		//        UP[3], Uxm1[3], Uxp1[3], Uym1[3], Uyp1[3], Uzm1[3], Uzp1[3]);
		//printf("jjjj vz %16.8e %16.8e %16.8e %16.8e %16.8e %16.8e\n",
		//        Fxp1[3], Fce[3], Gyp1[3], Gce[3], Hzp1[3], Hce[3]);

		node[ce].dnsP    = dnsP    = node[ce].UP[0];
		node[ce].vxP     = vxP     = node[ce].UP[1]/dnsP;
		node[ce].vyP     = vyP     = node[ce].UP[2]/dnsP;
		node[ce].vzP     = vzP     = node[ce].UP[3]/dnsP;
		node[ce].eneP    = eneP    = node[ce].UP[4]/dnsP;
		node[ce].tempP   = tempP   = (1.0/R)*(gam - 1.0) * ( eneP - 0.5*(vxP*vxP + vyP*vyP + vzP*vzP) );
		node[ce].enthalP = enthalP = R*gam*tempP/(gam - 1.0) + 0.5*(vxP*vxP + vyP*vyP + vzP*vzP);
		node[ce].prsP    = prsP    = R*dnsP*tempP;

		/* set F Predictor */                  /* set G Predictor */                  /* set H Predictor */

		node[ce].FP[0] = dnsP*vxP;             node[ce].GP[0] = dnsP*vyP;             node[ce].HP[0] = dnsP*vzP;
		node[ce].FP[1] = dnsP*vxP*vxP + prsP;  node[ce].GP[1] = dnsP*vyP*vxP;         node[ce].HP[1] = dnsP*vzP*vxP;
		node[ce].FP[2] = dnsP*vxP*vyP;         node[ce].GP[2] = dnsP*vyP*vyP + prsP;  node[ce].HP[2] = dnsP*vzP*vyP;
		node[ce].FP[3] = dnsP*vxP*vzP;         node[ce].GP[3] = dnsP*vyP*vzP;         node[ce].HP[3] = dnsP*vzP*vzP + prsP;
		node[ce].FP[4] = dnsP*vxP*enthalP;     node[ce].GP[4] = dnsP*vyP*enthalP;     node[ce].HP[4] = dnsP*vzP*enthalP;
		
	    }
	}

	if (flag_2D)
	{
	    for ( dum=0; dum < bc.num_none[0]; dum++)
	    {
		node_num = bc.none[dum];
		i = node[node_num].i;
		j = node[node_num].j;
		ce  = ny*i + j;
		xm1 = ny*(i-1) + j;
		xp1 = ny*(i+1) + j;
		ym1 = ny*i + (j-1);
		yp1 = ny*i + (j+1);

		if( xm1 < 0 || ym1 < 0 )
		{
		    printf("There is a problem with your mesh.\n");
		    printf("One of the surrounding nodes in either the x- or y-\n");
		    printf("position is negative.\n");
		    exit(1);
		}

		fdum = 0.25;

		for ( i2 = 0; i2 < 5; i2++)
		{
/*     Based on the calculation for UP[i2] below, I probably don't need all the positions for
       F, G, H, and U to be calculated.  Remember to comment them out later.
*/
#if 0
		    Fce[i2]  = node[ce].F[i2];   Gce[i2]  = node[ce].G[i2];
		    Fxm1[i2] = node[xm1].F[i2];  Gxm1[i2] = node[xm1].G[i2];
		    Fxp1[i2] = node[xp1].F[i2];  Gxp1[i2] = node[xp1].G[i2];
		    Fym1[i2] = node[ym1].F[i2];  Gym1[i2] = node[ym1].G[i2];
		    Fyp1[i2] = node[yp1].F[i2];  Gyp1[i2] = node[yp1].G[i2];

		    Uce[i2]  = node[ce].U[i2]; 
		    Uxm1[i2] = node[xm1].U[i2];
		    Uxp1[i2] = node[xp1].U[i2];
		    Uym1[i2] = node[ym1].U[i2];
		    Uyp1[i2] = node[yp1].U[i2];
#endif
		    Fce[i2]  = node[ce].F[i2];   Gce[i2]  = node[ce].G[i2];
		    Fxp1[i2] = node[xp1].F[i2];
		                                 Gyp1[i2] = node[yp1].G[i2];
		    Uxm1[i2] = node[xm1].U[i2];
		    Uxp1[i2] = node[xp1].U[i2];
		    Uym1[i2] = node[ym1].U[i2];
		    Uyp1[i2] = node[yp1].U[i2];

		    node[ce].UP[i2] = fdum*(Uxm1[i2] + Uxp1[i2] + Uym1[i2] + Uyp1[i2]) -
		            taux*(Fxp1[i2] - Fce[i2]) -
		            tauy*(Gyp1[i2] - Gce[i2]);

		    //printf("rrrr vz %16.8e %16.8e %16.8e %16.8e %16.8e %16.8e %16.8e\n",
		    //        UP[i2], Uxm1[i2], Uxp1[i2], Uym1[i2], Uyp1[i2], Uzm1[i2], Uzp1[i2]);
		}

		//printf("rrrr vz %16.8e %16.8e\n", vx, vy);

		//printf("rrrr vz %16.8e %16.8e %16.8e %16.8e %16.8e %16.8e %16.8e\n",
		//        UP[3], Uxm1[3], Uxp1[3], Uym1[3], Uyp1[3], Uzm1[3], Uzp1[3]);
		//printf("jjjj vz %16.8e %16.8e %16.8e %16.8e %16.8e %16.8e\n",
		//        Fxp1[3], Fce[3], Gyp1[3], Gce[3], Hzp1[3], Hce[3]);

		node[ce].dnsP    = dnsP    = node[ce].UP[0];
		node[ce].vxP     = vxP     = node[ce].UP[1]/dnsP;
		node[ce].vyP     = vyP     = node[ce].UP[2]/dnsP;
		node[ce].eneP    = eneP    = node[ce].UP[4]/dnsP;
		node[ce].tempP   = tempP   = (1.0/R)*(gam - 1.0) * ( eneP - 0.5*(vxP*vxP + vyP*vyP) );
		node[ce].enthalP = enthalP = R*gam*tempP/(gam - 1.0) + 0.5*(vxP*vxP + vyP*vyP);
		node[ce].prsP    = prsP    = R*dnsP*tempP;

		/* set F Predictor */                  /* set G Predictor */

		node[ce].FP[0] = dnsP*vxP;             node[ce].GP[0] = dnsP*vyP;
		node[ce].FP[1] = dnsP*vxP*vxP + prsP;  node[ce].GP[1] = dnsP*vyP*vxP;
		node[ce].FP[2] = dnsP*vxP*vyP;         node[ce].GP[2] = dnsP*vyP*vyP + prsP;
		node[ce].FP[3] = 0.0;                  node[ce].GP[3] = 0.0;
		node[ce].FP[4] = dnsP*vxP*enthalP;     node[ce].GP[4] = dnsP*vyP*enthalP;
		
	    }
	}

	return 1;
}



