/* This program draws the streamlines for the program which uses the
   finite difference method to calculate a CFD flow field.

                     Updated 6/8/08

    SLFCFD source file
    Version:  1.2
    Copyright (C) 2008  San Le

    The source code contained in this file is released under the
    terms of the GNU Library General Public License.

*/

#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "cfdconst.h"
#include "cfdstruct.h"
#include "../common/shape_struct.h"

#if COMPRESSIBLE
#define NODE    COM_NODE
#endif
#if INCOMPRESSIBLE
#define NODE    INC_NODE
#endif
#if GUI
#define NODE    GR_NODE
#endif

#define SMALL               1.0e-20


int cfdshape(SHAPE *, double , double );

extern double dx, dy, dz;
extern double dt, time_current;
extern int nstep, cstep;
extern int imax, jmax;


int streamlines(int n2, FLOWLINES stream, NODE *node)
{

	int i, i2, j, k, k2, dum, dum2, loop_flag;
	int ce, xm1, xm2, xp1, xp2, ym1, ym2, yp1, yp2, xm1ym1, xm1yp1, xp1ym1, xp1yp1;
	int index_i, index_j, num_pt0;
	double perx, pery, xi, xi2, eta, eta2, w, fdum, fdum2, fdum3,
		vec_dum[2], vec_dum2[2], vec_dum_mag, vec_dum2_mag;
	SHAPE sh;

	w = 1.0;
	num_pt0 = stream.num_pt0;

/* The variables below generate the streamlines
*/
	for( k = 0; k < num_pt0; ++k )
	{
	    loop_flag = 0;
	    for( k2 = 0; k2 < n2; ++k2 )
	    {

		/*printf(  "stream %4d   %16.8e %16.8e\n", i, stream.pt0[num_pt0*n2+k].x, stream.pt0[num_pt0*n2+k].y);*/

/* Calculate the next point of the streamline for each one based on the last point. */

		i = (int)(stream.pt[num_pt0*k2+k].x/dx);
		j = (int)(stream.pt[num_pt0*k2+k].y/dy);

		ce = jmax*i + j;
		xm1 = jmax*(i-1) + j;
		xp1 = jmax*(i+1) + j;
		ym1 = jmax*i + j-1;
		yp1 = jmax*i + j+1;
		xp1yp1 = jmax*(i+1) + j+1;
		xm1yp1 = jmax*(i-1) + j+1;
		xp1ym1 = jmax*(i+1) + j-1;

		if( i < imax-1 && i > 0 && j < jmax-1 && j > 0 && loop_flag < 3 ) 
		{
		    perx = stream.pt[num_pt0*k2+k].x/dx - (double)(i);
		    pery = stream.pt[num_pt0*k2+k].y/dy - (double)(j);

		    xi  = 2.0*perx - 1.0;
		    eta = 2.0*pery - 1.0;

/* Because I am using a staggared grid for u(i,j) and v(i,j) which do not fall on node
   points, xi and eta have to be determined based on this shift.

   First, -1 <= xi <= 1.0, and xi = 0.0 is where the node of u lies.  This means that I check whether
   at xi > 0.0. in which case I interpolate between node[ce].vx and node[xp1].vx.  If not, then the interpolation
   is between node[xm1].vx and node[ce].vx. 

   In addition, once it is determined which side of 0.0 xi lies on, I will continue using xi for
   interpolation purposes for u.  The only consideration I have to make is either I do xi + 1.0 if
   xi < 0.0 (because I'm interpolating from node[xm1].vx to node[ce].vx or xi - 1.0 if xi > 0.0 because I'm
   interpolating from node[ce].vx to node[xp1].vx. 

   The same is done for v with respect to eta.

   On October 9th, 2007, I rechecked the algorithm below and it seemed to still be good.
*/
		    if( xi > 0.0 )
		    {
			xi2 = xi - 1.0;
			cfdshape( &sh, xi2, eta);

			fdum = sh.Nhat[0].dx0*node[ce].vx +
			  sh.Nhat[1].dx0*node[xp1].vx +
			  sh.Nhat[2].dx0*node[xp1yp1].vx +
			  sh.Nhat[3].dx0*node[yp1].vx;
		    }
		    else
		    {
			xi2 = xi + 1.0;
			cfdshape( &sh, xi2, eta);

			fdum = sh.Nhat[0].dx0*node[xm1].vx +
			  sh.Nhat[1].dx0*node[ce].vx +
			  sh.Nhat[2].dx0*node[yp1].vx +
			  sh.Nhat[3].dx0*node[xm1yp1].vx;
		    }

		    stream.pt[num_pt0*(k2+1)+k].x =
			stream.pt[num_pt0*k2+k].x + w*dt*fdum;

		    /*printf(  "stream %4d %4d  %10.4e %10.4e %16.8e %16.8e %16.8e ",
			k2, k, w, dt, fdum, stream.pt0[num_pt0*k2+k].x, stream.pt[num_pt0*(k2+1)+k].x);*/

		    if( eta > 0.0 )
		    {
			eta2 = eta - 1.0;
			cfdshape( &sh, xi, eta2);

			fdum = sh.Nhat[0].dx0*node[ce].vy +
			  sh.Nhat[1].dx0*node[xp1].vy +
			  sh.Nhat[2].dx0*node[xp1yp1].vy +
			  sh.Nhat[3].dx0*node[yp1].vy;
		    }
		    else
		    {
			eta2 = eta + 1.0;
			cfdshape( &sh, xi, eta2);

			fdum = sh.Nhat[0].dx0*node[ym1].vy +
			  sh.Nhat[1].dx0*node[xp1ym1].vy +
			  sh.Nhat[2].dx0*node[xp1].vy +
			  sh.Nhat[3].dx0*node[ce].vy;
		    }

		    stream.pt[num_pt0*(k2+1)+k].y =
			stream.pt[num_pt0*k2+k].y + w*dt*fdum;


/* The streamlines stopped getting drawn after they loop around once.  This is determined by
   taking the vector cross product of adjacent points with the first 2 points.
*/

		    if( k2 == 0)
		    {
			vec_dum[0] = stream.pt[num_pt0+k].x - stream.pt[k].x;
			vec_dum[1] = stream.pt[num_pt0+k].y - stream.pt[k].y;
			vec_dum_mag = vec_dum[0]*vec_dum[0] + vec_dum[1]*vec_dum[1];
			vec_dum_mag = sqrt(vec_dum_mag) + SMALL;
			vec_dum[0] /= vec_dum_mag;
			vec_dum[1] /= vec_dum_mag;
		    }
		    else
		    {
			vec_dum2[0] = stream.pt[num_pt0*(k2+1)+k].x - stream.pt[num_pt0*k2+k].x;
			vec_dum2[1] = stream.pt[num_pt0*(k2+1)+k].y - stream.pt[num_pt0*k2+k].y;
			vec_dum2_mag = vec_dum2[0]*vec_dum2[0] + vec_dum2[1]*vec_dum2[1];
			vec_dum2_mag = sqrt(vec_dum2_mag) + SMALL;
			vec_dum2[0] /= vec_dum2_mag;
			vec_dum2[1] /= vec_dum2_mag;

			fdum3 = vec_dum_mag*vec_dum2_mag;

/* Take the dot product of the vectors. */

			fdum = vec_dum[0]*vec_dum2[0] + vec_dum[1]*vec_dum2[1]; 

/* If fum < 0.0, then the loop is moving in the oppostite direction of where it started. */

			if(fdum < 0.0 && fabs(fdum) > 0.95 )
			{
			    loop_flag = 1;
			/*if(k > 40)
		    printf(  " RR %4d %4d %4d %16.8e %16.8f %16.8f %16.8f %16.8f %16.8f %16.8f %16.8f %16.8f\n",
			k, k2, loop_flag, fdum,
			stream.pt[num_pt0*k2+k].x, stream.pt[num_pt0*k2+k].y,
			vec_dum[0], vec_dum[1], vec_dum2[0], vec_dum2[1], vec_dum_mag, vec_dum2_mag);*/
			}

/* If fum > 0.0, then the loop is moving in the oppostite direction of where it started. */

			if(loop_flag == 1)
			{

/* If the streamline has looped around and now points in the same direction as the where it started,
   stop drawing the streamline.
*/
			    if(fdum > 0.0 && fabs(fdum) > 0.95 )
			    {
				loop_flag = 2;
			/*if(k > 40)
		    printf(  " ZZZZ %4d %4d %4d %16.8e %16.8f %16.8f\n", k, k2, loop_flag,
			fdum, stream.pt[num_pt0*k2+k].x, stream.pt[num_pt0*k2+k].y);*/
			    }
			}

			if(loop_flag == 2)
			{
			    fdum  = stream.pt[num_pt0*k2+k].x - stream.pt[k].x;
			    fdum2 = stream.pt[num_pt0*k2+k].y - stream.pt[k].y;
			    fdum3 = sqrt(fdum*fdum + fdum2*fdum2);
			    if(fdum3 < dx/2.0 )
			    {
				loop_flag = 3;
			    }
			}


		    }

		    /*printf(  " RR %16.8e %16.8e %16.8e\n",
			 fdum, stream.pt0[num_pt0*k2+k].y, stream.pt[num_pt0*(k2+1)+k].y);*/

		    /*stream.pt[num_pt0*k2+k].x =
			stream.pt[num_pt0*k2+k].x + w*dt*(perx*(node[xp1].vx - node[ce].vx) + node[ce].vx);
		    stream.pt[num_pt0*k2+k].y =
			stream.pt[num_pt0*k2+k].y + w*dt*(perx*(node[yp1].vy - node[ce].vy) + node[ce].vy);*/
		} 
		else
		{
		    stream.pt[num_pt0*k2+k].x = -100000.0;
		    stream.pt[num_pt0*k2+k].y = -100000.0;
		}
	    }
	}

	return 1;

}


