/* This program reapplies the boundary conditions for the program
   which uses the finite difference method to calculate an incomprssible
   laminar CFD flow field.  

    SLFCFD source file
    Version:  1.2
    Copyright (C) 2006  San Le

    The source code contained in this file is released under the
    terms of the GNU Library General Public License.

*/

#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "../compressible/cfdconst.h"
#include "../compressible/cfdstruct.h"
#include "../AMR/amr.h"

#define PT666 1.0/3.0

extern double dx, dy, dz, Re;
extern double dt, time_current;
extern double q, alp;
extern int debug_flag, reverse_flag;

int solve3X3( double *, double *, double * );

int applyBC_inflow( BOUND bc, INC_NODE *node, SUBGRID subgrid)
{
/* Below is the bondary condition for the inflow at the left.
   It reprsents linearally extrapolating the pressure from the
   2 nodes that are usually in the opposite direction of the
   boundary which it is near.  I don't use this function, so I am
   keeping it as a placeholder.

                     Updated 4/18/08
*/
	int i, j, k, dum, node_num;
	int ce, xm1, xm2, xm3, xp1, xp2, xp3, ym1, ym2, ym3, yp1, yp2, yp3;
	int nx, ny;
	double vx_xm1, vx_xp1, vy_ym1, vy_yp1;

	ny = subgrid.ny;
	nx = subgrid.nx;

	for( dum = 0; dum < bc.num_inflow[0].xp; ++dum )
	{
	    node_num = bc.inflow.xp[dum];
	    i = node[node_num].i;
	    j = node[node_num].j;
	    ce = ny*i+j;
	    xm1  = ny*(i-1) + j;
	    xm2  = ny*(i-2) + j;
	    xm3  = ny*(i-3) + j;
	    yp1 = ny*i + j+1;
	    ym1 = ny*i + j-1;

	    vx_xp1 = (18.0*node[ce].vxQ - 9.0*node[xm1].vxQ + 2.0*node[xm2].vx)/11.0;
	    node[ce].prs = (node[yp1].prs + 2.0*node[xm1].prs + node[ym1].prs)/4.0 -
		dx/dt*(vx_xp1 - node[xm1].vxQ + node[ce].vyQ - node[ym1].vyQ)/4.0;
	}

	for( dum = 0; dum < bc.num_inflow[0].xm; ++dum )
	{
	    node_num = bc.inflow.xm[dum];
	    i = node[node_num].i;
	    j = node[node_num].j;
	    ce = ny*i+j;
	    xp1  = ny*(i+1) + j;
	    xp2  = ny*(i+2) + j;
	    xp3  = ny*(i+3) + j;
	    yp1 = ny*i + j+1;
	    ym1 = ny*i + j-1;

	    vx_xm1 = (18.0*node[ce].vxQ - 9.0*node[xp1].vxQ + 2.0*node[xp2].vx)/11.0;
	    node[ce].prs = (node[yp1].prs + 2.0*node[xp1].prs + node[ym1].prs)/4.0 -
		dx/dt*(node[ce].vxQ - vx_xm1 + node[ce].vyQ - node[ym1].vyQ)/4.0;
	}

	for( dum = 0; dum < bc.num_inflow[0].yp; ++dum )
	{
	    node_num = bc.inflow.yp[dum];
	    i = node[node_num].i;
	    j = node[node_num].j;
	    ce = ny*i+j;
	    xm1  = ny*(i-1) + j;
	    xp1  = ny*(i+1) + j;
	    ym1 = ny*i + j-1;
	    ym2 = ny*i + j-2;
	    ym3 = ny*i + j-3;

	    vy_yp1 = (18.0*node[ce].vyQ - 9.0*node[ym1].vyQ + 2.0*node[ym2].vy)/11.0;
	    node[ce].prs = (node[xp1].prs + 2.0*node[ym1].prs + node[xm1].prs)/4.0 -
		dx/dt*(node[ce].vxQ - node[xm1].vxQ + vy_yp1 - node[ym1].vyQ)/4.0;
	}

	for( dum = 0; dum < bc.num_inflow[0].ym; ++dum )
	{
	    node_num = bc.inflow.ym[dum];
	    i = node[node_num].i;
	    j = node[node_num].j;
	    ce = ny*i+j;
	    xp3  = ny*(i-1) + j;
	    xp1  = ny*(i+1) + j;
	    yp1 = ny*i + j+1;
	    yp2 = ny*i + j+2;
	    yp3 = ny*i + j+3;

	    vy_ym1 = (18.0*node[ce].vyQ - 9.0*node[yp1].vyQ + 2.0*node[yp2].vy)/11.0;
	    node[ce].prs = (node[xp1].prs + 2.0*node[yp1].prs + node[xm1].prs)/4.0 -
		dx/dt*(node[ce].vxQ - node[xm1].vxQ + node[ce].vyQ - vy_ym1)/4.0;
	}

	return 1;
}


#if NEW

int applyBC_free2( BOUND bc, INC_NODE *node, SUBGRID subgrid)
{
/* Below is the bondary condition for the outflow opposite the inflow.
   In the compressible flow, this took the form of a simple free boundary
   condition.

                     Updated 6/8/08
*/
	int i, j, k, dum, node_num, velocity_flag;
	int ce, xm1, xm2, xm3, xp1, xp2, ym1, ym2, yp1, yp2, xm1ym1, xm1yp1, xp1ym1, xp1yp1;
	double vxFvx_xp1, vxFvy_xp1;
	double CvxFvxR, CvxFvxL, CvyFvxT, CvyFvxB, CvxFvyR, CvxFvyL, CvyFvyT, CvyFvyB, px, py;
	double gx, gy;
	double vx_xp1, vx_xp2, vx_xm2, vx_yp1, vx_yp2, vx_ym2,
		vy_xp1, vy_xp2, vy_xm2, vy_yp1, vy_yp2, vy_ym2, vy_xp1ym1;
	double du2dx2A, du2dy2A, dv2dx2A, dv2dy2A, du2dx2B, du2dy2B, dv2dx2B, dv2dy2B;
	int nx, ny;

	ny = subgrid.ny;
	nx = subgrid.nx;


	gx=alp*dt/(dx*dx);
	gy=alp*dt/(dy*dy);
	gx=0.0;
	gy=0.0;

/* The reason I have set gx and gy to zero below is because in Dr. Moore's email,
   she says,

     "In the velocity computation, the quadratic differencing is used except
     in the last row, which is NI for v and NI+1 for u.  In this row,
     diffusion terms are ignored and convection is based on first-order
     upwinding."

   I think I removed them because they represented the diffusion terms since they
   are associated with the second derivatives of u and v with respect to x and y.
*/


	for( dum = 0; dum < bc.num_free2[0].xp; ++dum )
	{
		node_num = bc.free2.xp[dum];
		i = node[node_num].i;
		j = node[node_num].j;

		ce  = ny*i + j;
		xm1 = ny*(i-1) + j;
		xm2 = ny*(i-2) + j;
		xp1 = ny*(i+1) + j;
		ym1 = ny*i + (j-1);
		ym2 = ny*i + (j-2);
		yp1 = ny*i + (j+1);
		yp2 = ny*i + (j+2);
		xm1ym1 = ny*(i-1) + (j-1);
		xm1yp1 = ny*(i-1) + (j+1);
		xp1ym1 = ny*(i+1) + (j-1);
		xp1yp1 = ny*(i+1) + (j+1);

		vx_xp1 = 2.0*node[ce].vx - node[xm1].vx;
		vx_xp1 = (18.0*node[ce].vx - 9.0*node[xm1].vx + 2.0*node[xm2].vx)/11.0;
		vy_xp1 = 2.0*node[ce].vy - node[xm1].vy;
		vy_xp1 = (18.0*node[ce].vy - 9.0*node[xm1].vy + 2.0*node[xm2].vy)/11.0;
		vy_xp1ym1 = 2.0*node[ym1].vy - node[xm1ym1].vy;

		vy_xp2 = (18.0*vy_xp1 - 9.0*node[ce].vy + 2.0*node[xm1].vy)/11.0;
		vx_xp2 = (18.0*vx_xp1 - 9.0*node[ce].vx + 2.0*node[xm1].vx)/11.0;

		vy_ym2 = (18.0*node[ym1].vy - 9.0*node[ce].vy + 2.0*node[yp1].vy)/11.0;
		vx_ym2 = (18.0*node[ym1].vx - 9.0*node[ce].vx + 2.0*node[yp1].vx)/11.0;
		if( j > 1)
		{
		    vy_ym2 = node[ym2].vy;
		    vx_ym2 = node[ym2].vx;
		}

		vy_yp2 = (18.0*node[yp1].vy - 9.0*node[ce].vy + 2.0*node[ym1].vy)/11.0;
		vx_yp2 = (18.0*node[yp1].vx - 9.0*node[ce].vx + 2.0*node[ym1].vx)/11.0;
		if( j < ny-2)
		{
		    vy_yp2 = node[yp2].vy;
		    vx_yp2 = node[yp2].vx;
		}

		vxFvx_xp1 = .5*(vx_xp1 + node[ce].vx) - q/3.0*(node[xm1].vx - 2.0*node[ce].vx + vx_xp1);
		vxFvy_xp1 = .5*(node[yp1].vx + node[ce].vx) - q/3.0*(node[ym1].vx - 2.0*node[ce].vx + node[yp1].vx);

#if 0
/*
   Unlike in the true quickest algorithm, I do not use the face velocities at the boundary.
   This improves the solution and makes it more stable.  I cannot justify this numerically,
   as it is something I just stumbled on.  Note also that I use the same points for both
   vx and vy.  If you look in the quickest algorithm, the lines below are given by:
*/

		CvxFvxR = dt*node[xp1].vxFvx/dx;  /* for the right (actually, dt*vxFvx_xp1/dx */
		CvxFvxL = dt*node[ce].vxFvx/dx;   /* for the left */
		CvyFvxT = dt*node[yp1].vyFvx/dy;  /* for the top */
		CvyFvxB = dt*node[ce].vyFvx/dy;   /* for the bottom */

		CvxFvyR = dt*node[xp1].vxFvy/dx;  /* for the right (actually, dt*vxFvy_xp1/dx */
		CvxFvyL = dt*node[ce].vxFvy/dx;   /* for the left */
		CvyFvyT = dt*node[yp1].vyFvy/dy;  /* for the top */
		CvyFvyB = dt*node[ce].vyFvy/dy;   /* for the bottom */
#endif

		CvxFvxR = dt*node[ce].vx/dx;
		CvxFvxL = dt*node[xm1].vx/dx;
		CvyFvxT = dt*node[ce].vy/dy;
		CvyFvxB = dt*node[ym1].vy/dy;

		CvxFvyR = dt*node[ce].vx/dx;
		CvxFvyL = dt*node[xm1].vx/dx;
		CvyFvyT = dt*node[ce].vy/dy;
		CvyFvyB = dt*node[ym1].vy/dy;

		node[ce].vxQ = node[ce].vx
		     -CvxFvxR*(.5*(node[ce].vx+vx_xp1) - .5*CvxFvxR*(vx_xp1 - node[ce].vx)
		     -( (1 - CvxFvxR*CvxFvxR)/6 - gx)*(node[xm1].vx - 2*node[ce].vx + vx_xp1))

		     +CvxFvxL*(.5*(node[xm1].vx + node[ce].vx) - .5*CvxFvxL*(node[ce].vx - node[xm1].vx)
		     -( (1 - CvxFvxL*CvxFvxL)/6 - gx)*(node[xm2].vx - 2*node[xm1].vx + node[ce].vx))

		     -CvyFvxT*(.5*(node[ce].vx + node[yp1].vx) - .5*CvyFvxT*(node[yp1].vx - node[ce].vx)
		     -( (1 - CvyFvxT*CvyFvxT)/6 - gy)*(node[ym1].vx - 2*node[ce].vx + node[yp1].vx))

		     +CvyFvxB*(.5*(node[ym1].vx + node[ce].vx) - .5*CvyFvxB*(node[ce].vx - node[ym1].vx)
		     -( (1-CvyFvxB*CvyFvxB)/6 - gy)*(vx_ym2 - 2*node[ym1].vx + node[ce].vx))

		     +gx*(node[xm1].vx - 2*node[ce].vx + vx_xp1)
		     +gy*(node[ym1].vx - 2*node[ce].vx + node[yp1].vx);

		node[ce].vyQ = node[ce].vy
		     -CvxFvyR*(.5*(node[ce].vy + vy_xp1) - .5*CvxFvyR*(vy_xp1 - node[ce].vy)
		     -( (1 - CvxFvyR*CvxFvyR)/6 - gx)*(node[xm1].vy - 2*node[ce].vy + vy_xp1))

		     +CvxFvyL*(.5*(node[xm1].vy + node[ce].vy) - .5*CvxFvyL*(node[ce].vy - node[xm1].vy)
		     -( (1-CvxFvyL*CvxFvyL)/6 - gx)*(node[xm2].vy - 2*node[xm1].vy + node[ce].vy))

		     -CvyFvyT*(.5*(node[ce].vy + node[yp1].vy) -.5*CvyFvyT*(node[yp1].vy - node[ce].vy)
		     -( (1 - CvyFvyT*CvyFvyT)/6 - gy)*(node[ym1].vy - 2*node[ce].vy + node[yp1].vy))

		     +CvyFvyB*(.5*(node[ym1].vy + node[ce].vy) - .5*CvyFvyB*(node[ce].vy - node[ym1].vy)
		     -( (1 - CvyFvyB*CvyFvyB)/6 - gy)*(vy_ym2 - 2*node[ym1].vy + node[ce].vy))

		     +gx*(node[xm1].vy - 2*node[ce].vy + vy_xp1)
		     +gy*(node[ym1].vy - 2*node[ce].vy + node[yp1].vy);

/* I don't shift the derivatives with respect to x because that would involve vx_xp2 and vy_xp2 which
   is beyond the point I want to extrapolate.
*/

		/*velocity_flag = 0;*/

		{
		   du2dx2A = node[xm1].vx - 2.0*node[ce].vx + vx_xp1;
		   du2dx2B = node[xm2].vx - 2.0*node[xm1].vx + node[ce].vx;
		   dv2dx2A = node[xm1].vy - 2.0*node[ce].vy + vy_xp1;
		   dv2dx2B = node[xm2].vy - 2.0*node[xm1].vy + node[ce].vy;
		}
		if( node[ce].vyQ >= 0.0 )
		{
		   du2dy2A = node[ym1].vx - 2.0*node[ce].vx + node[yp1].vx;
		   du2dy2B = vx_ym2 - 2.0*node[ym1].vx + node[ce].vx;
		   dv2dy2A = node[ym1].vy - 2.0*node[ce].vy + node[yp1].vy;
		   dv2dy2B = vy_ym2 - 2.0*node[ym1].vy + node[ce].vy;
		}
		if( node[ce].vyQ < 0.0 )
		{
		   du2dy2A = node[ce].vx - 2.0*node[yp1].vx + vx_yp2;
		   du2dy2B = node[ym1].vx - 2.0*node[ce].vx + node[yp1].vx;
		   dv2dy2A = node[ce].vy - 2.0*node[yp1].vy + vy_yp2;
		   dv2dy2B = node[ym1].vy - 2.0*node[ce].vy + node[yp1].vy;
		   velocity_flag = 1;
		}

		/*printf(  "uuuuu %d %d %d %d %16.8e %16.8e\n", ce, xm1, xm2, velocity_flag, dv2dy2A, dv2dy2B);*/

		if( velocity_flag )
		{
		   node[ce].vxQ = node[ce].vx
		     - CvxFvxR*(.5*(node[ce].vx + vx_xp1) - .5*CvxFvxR*(vx_xp1 - node[ce].vx)
		     - ( (1.0 - CvxFvxR*CvxFvxR)/6.0 - gx )*du2dx2A)

		     + CvxFvxL*(.5*(node[xm1].vx + node[ce].vx) - .5*CvxFvxL*(node[ce].vx - node[xm1].vx)
		     - ( (1.0 - CvxFvxL*CvxFvxL)/6.0 - gx )*du2dx2B)

		     - CvyFvxT*(.5*(node[ce].vx + node[yp1].vx) - .5*CvyFvxT*(node[yp1].vx - node[ce].vx)
		     - ( (1.0 - CvyFvxT*CvyFvxT)/6.0 - gy )*du2dy2A)

		     + CvyFvxB*(.5*(node[ym1].vx + node[ce].vx) - .5*CvyFvxB*(node[ce].vx - node[ym1].vx)
		     - ( (1.0 - CvyFvxB*CvyFvxB)/6.0 - gy )*du2dy2B)

		     + gx*(node[xm1].vx - 2.0*node[ce].vx + vx_xp1)
		     + gy*(node[ym1].vx - 2.0*node[ce].vx + node[yp1].vx);

		   node[ce].vyQ = node[ce].vy
		     - CvxFvyR*(.5*(node[ce].vy + vy_xp1) - .5*CvxFvyR*(vy_xp1 - node[ce].vy)
		     - ( (1.0 - CvxFvyR*CvxFvyR)/6.0 - gx )*dv2dx2A)

		     + CvxFvyL*(.5*(node[xm1].vy + node[ce].vy) - .5*CvxFvyL*(node[ce].vy - node[xm1].vy)
		     - ( (1.0 - CvxFvyL*CvxFvyL)/6.0 - gx )*dv2dx2B)

		     - CvyFvyT*(.5*(node[ce].vy + node[yp1].vy) - .5*CvyFvyT*(node[yp1].vy - node[ce].vy)
		     - ( (1.0 - CvyFvyT*CvyFvyT)/6.0 - gy )*dv2dy2A)

		     + CvyFvyB*(.5*(node[ym1].vy + node[ce].vy) - .5*CvyFvyB*(node[ce].vy - node[ym1].vy)
		     - ( (1.0 - CvyFvyB*CvyFvyB)/6.0 - gy )*dv2dy2B)

		     + gx*(node[xm1].vy - 2.0*node[ce].vy + vy_xp1)
		     + gy*(node[ym1].vy - 2.0*node[ce].vy + node[yp1].vy);
		}


		/*node[ce].prs = 2.0*node[xm1].prs - node[xm2].prs;
		node[ce].prs = 2.0*node[xm2].prs - node[xm3].prs;
		node[ce].prs = node[xm1].prs;*/
	}


	for( dum = 0; dum < bc.num_free2[0].yp; ++dum )
	{
		node_num = bc.free2.yp[dum];
		i = node[node_num].i;
		j = node[node_num].j;

		ce  = ny*i + j;
		xm1 = ny*(i-1) + j;
		xm2 = ny*(i-2) + j;
		xp1 = ny*(i+1) + j;
		xp2 = ny*(i+2) + j;
		ym1 = ny*i + (j-1);
		ym2 = ny*i + (j-2);
		yp1 = ny*i + (j+1);
		xm1ym1 = ny*(i-1) + (j-1);
		xm1yp1 = ny*(i-1) + (j+1);
		xp1ym1 = ny*(i+1) + (j-1);
		xp1yp1 = ny*(i+1) + (j+1);

		vx_yp1 = 2.0*node[ce].vx - node[ym1].vx;
		vx_yp1 = (18.0*node[ce].vx - 9.0*node[ym1].vx + 2.0*node[ym2].vx)/11.0;
		vy_yp1 = 2.0*node[ce].vy - node[ym1].vy;
		vy_yp1 = (18.0*node[ce].vy - 9.0*node[ym1].vy + 2.0*node[ym2].vy)/11.0;

		vy_xm2 = (18.0*node[xm1].vy - 9.0*node[ce].vy + 2.0*node[xp1].vy)/11.0;
		vx_xm2 = (18.0*node[xm1].vx - 9.0*node[ce].vx + 2.0*node[xp1].vx)/11.0;
		if( i > 1)
		{
		    vy_xm2 = node[xm2].vy;
		    vx_xm2 = node[xm2].vx;
		}

		vy_xp2 = (18.0*node[xp1].vy - 9.0*node[ce].vy + 2.0*node[xm1].vy)/11.0;
		vx_xp2 = (18.0*node[xp1].vx - 9.0*node[ce].vx + 2.0*node[xm1].vx)/11.0;
		if( i < nx-2)
		{
		    vy_xp2 = node[xp2].vy;
		    vx_xp2 = node[xp2].vx;
		}

#if 0
/*
   Unlike in the true quickest algorithm, I do not use the face velocities at the boundary.
   This improves the solution and makes it more stable.  I cannot justify this numerically,
   as it is something I just stumbled on.  Note also that I use the same points for both
   vx and vy.  If you look in the quickest algorithm, the lines below are given by:
*/

		CvxFvxR = dt*node[xp1].vxFvx/dx;  /* for the right */
		CvxFvxL = dt*node[ce].vxFvx/dx;   /* for the left */
		CvyFvxT = dt*node[yp1].vyFvx/dy;  /* for the top (actually, dt*vyFvx_yp1/dx */
		CvyFvxB = dt*node[ce].vyFvx/dy;   /* for the bottom */

		CvxFvyR = dt*node[xp1].vxFvy/dx;  /* for the right */
		CvxFvyL = dt*node[ce].vxFvy/dx;   /* for the left */
		CvyFvyT = dt*node[yp1].vyFvy/dy;  /* for the top (actually, dt*vyFvy_yp1/dx */
		CvyFvyB = dt*node[ce].vyFvy/dy;   /* for the bottom */

#endif

		CvxFvxR = dt*node[ce].vx/dx;
		CvxFvxL = dt*node[xm1].vx/dx;
		CvyFvxT = dt*node[ce].vy/dy;
		CvyFvxB = dt*node[ym1].vy/dy;

		CvxFvyR = dt*node[ce].vx/dx;
		CvxFvyL = dt*node[xm1].vx/dx;
		CvyFvyT = dt*node[ce].vy/dy;
		CvyFvyB = dt*node[ym1].vy/dy;

		node[ce].vxQ = node[ce].vx
		     -CvxFvxR*(.5*(node[ce].vx+node[xp1].vx)-.5*CvxFvxR*(node[xp1].vx-node[ce].vx)
		     -( (1-CvxFvxR*CvxFvxR)/6 -gx )*(node[xm1].vx-2*node[ce].vx+node[xp1].vx))

		     +CvxFvxL*(.5*(node[xm1].vx+node[ce].vx)-.5*CvxFvxL*(node[ce].vx-node[xm1].vx)
		     -( (1-CvxFvxL*CvxFvxL)/6 -gx )*(vx_xm2-2*node[xm1].vx+node[ce].vx))

		     -CvyFvxT*(.5*(node[ce].vx+vx_yp1)-.5*CvyFvxT*(vx_yp1-node[ce].vx)
		     -( (1-CvyFvxT*CvyFvxT)/6 -gy )*(node[ym1].vx-2*node[ce].vx+vx_yp1))

		     +CvyFvxB*(.5*(node[ym1].vx+node[ce].vx)-.5*CvyFvxB*(node[ce].vx-node[ym1].vx)
		     -( (1-CvyFvxB*CvyFvxB)/6 -gy )*(node[ym2].vx-2*node[ym1].vx+node[ce].vx))

		     +gx*(node[xm1].vx-2*node[ce].vx+node[xp1].vx)
		     +gy*(node[ym1].vx-2*node[ce].vx+vx_yp1);

		node[ce].vyQ = node[ce].vy
		     -CvxFvyR*(.5*(node[ce].vy+node[xp1].vy)-.5*CvxFvyR*(node[xp1].vy-node[ce].vy)
		     -( (1-CvxFvyR*CvxFvyR)/6 -gx )*(node[xm1].vy-2*node[ce].vy+node[xp1].vy))

		     +CvxFvyL*(.5*(node[xm1].vy+node[ce].vy)-.5*CvxFvyL*(node[ce].vy-node[xm1].vy)
		     -( (1-CvxFvyL*CvxFvyL)/6 -gx )*(vy_xm2-2*node[xm1].vy+node[ce].vy))

		     -CvyFvyT*(.5*(node[ce].vy+vy_yp1)-.5*CvyFvyT*(vy_yp1-node[ce].vy)
		     -( (1-CvyFvyT*CvyFvyT)/6 -gy )*(node[ym1].vy-2*node[ce].vy+vy_yp1))

		     +CvyFvyB*(.5*(node[ym1].vy+node[ce].vy)-.5*CvyFvyB*(node[ce].vy-node[ym1].vy)
		     -( (1-CvyFvyB*CvyFvyB)/6 -gy )*(node[ym2].vy-2*node[ym1].vy+node[ce].vy))

		     +gx*(node[xm1].vy-2*node[ce].vy+node[xp1].vy)
		     +gy*(node[ym1].vy-2*node[ce].vy+vy_yp1);

		velocity_flag = 0;

		if( node[ce].vxQ >= 0.0 )
		{
		   du2dx2A = node[xm1].vx - 2.0*node[ce].vx + node[xp1].vx;
		   du2dx2B = vx_xm2 - 2.0*node[xm1].vx + node[ce].vx;
		   dv2dx2A = node[xm1].vy - 2.0*node[ce].vy + node[xp1].vy;
		   dv2dx2B = vy_xm2 - 2.0*node[xm1].vy + node[ce].vy;
		}
		if( node[ce].vxQ < 0.0 )
		{
		   du2dx2A = node[ce].vx - 2.0*node[xp1].vx + vx_xp2;
		   du2dx2B = node[xm1].vx - 2.0*node[ce].vx + node[xp1].vx;
		   dv2dx2A = node[ce].vy - 2.0*node[xp1].vy + vy_xp2;
		   dv2dx2B = node[xm1].vy - 2.0*node[ce].vy + node[xp1].vy;
		   velocity_flag = 1;
		}
/* I don't shift the derivatives with respect to y because that would involve vx_yp2 and vy_yp2 which
   is beyond the point I want to extrapolate.
*/

		{
		   du2dy2A = node[ym1].vx - 2.0*node[ce].vx + vx_yp1;
		   du2dy2B = node[ym2].vx - 2.0*node[ym1].vx + node[ce].vx;
		   dv2dy2A = node[ym1].vy - 2.0*node[ce].vy + vy_yp1;
		   dv2dy2B = node[ym2].vy - 2.0*node[ym1].vy + node[ce].vy;
		}
		/*printf(  "uuuuu %d %d %d %d %16.8e %16.8e\n", ce, xm1, xm2, velocity_flag, du2dy2A, du2dy2B);*/

		if( velocity_flag )
		{
		   node[ce].vxQ = node[ce].vx
		     - CvxFvxR*(.5*(node[ce].vx + node[xp1].vx) - .5*CvxFvxR*(node[xp1].vx - node[ce].vx)
		     - ( (1.0 - CvxFvxR*CvxFvxR)/6.0 - gx )*du2dx2A)

		     + CvxFvxL*(.5*(node[xm1].vx + node[ce].vx) - .5*CvxFvxL*(node[ce].vx - node[xm1].vx)
		     - ( (1.0 - CvxFvxL*CvxFvxL)/6.0 - gx )*du2dx2B)

		     - CvyFvxT*(.5*(node[ce].vx + vx_yp1) - .5*CvyFvxT*(vx_yp1 - node[ce].vx)
		     - ( (1.0 - CvyFvxT*CvyFvxT)/6.0 - gy )*du2dy2A)

		     + CvyFvxB*(.5*(node[ym1].vx + node[ce].vx) - .5*CvyFvxB*(node[ce].vx - node[ym1].vx)
		     - ( (1.0 - CvyFvxB*CvyFvxB)/6.0 - gy )*du2dy2B)

		     + gx*(node[xm1].vx - 2.0*node[ce].vx + node[xp1].vx)
		     + gy*(node[ym1].vx - 2.0*node[ce].vx + vx_yp1);

		   node[ce].vyQ = node[ce].vy
		     - CvxFvyR*(.5*(node[ce].vy + node[xp1].vy) - .5*CvxFvyR*(node[xp1].vy - node[ce].vy)
		     - ( (1.0 - CvxFvyR*CvxFvyR)/6.0 - gx )*dv2dx2A)

		     + CvxFvyL*(.5*(node[xm1].vy + node[ce].vy) - .5*CvxFvyL*(node[ce].vy - node[xm1].vy)
		     - ( (1.0 - CvxFvyL*CvxFvyL)/6.0 - gx )*dv2dx2B)

		     - CvyFvyT*(.5*(node[ce].vy + vy_yp1) - .5*CvyFvyT*(vy_yp1 - node[ce].vy)
		     - ( (1.0 - CvyFvyT*CvyFvyT)/6.0 - gy )*dv2dy2A)

		     + CvyFvyB*(.5*(node[ym1].vy + node[ce].vy) - .5*CvyFvyB*(node[ce].vy - node[ym1].vy)
		     - ( (1.0 - CvyFvyB*CvyFvyB)/6.0 - gy )*dv2dy2B)

		     + gx*(node[xm1].vy - 2.0*node[ce].vy + node[xp1].vy)
		     + gy*(node[ym1].vy - 2.0*node[ce].vy + vy_yp1);
		}


	}

	return 1;

}

#endif


#if NEW2

int applyBC_free2( BOUND bc, INC_NODE *node, SUBGRID subgrid)
{
/* Below is the bondary condition for the outflow opposite the inflow.
   In the compressible flow, this took the form of a simple free boundary
   condition.

                     Updated 6/10/08
*/
	int i, j, k, dum, node_num, velocity_flag;
	int ce, xm1, xm2, xm3, xp1, xp2, ym1, ym2, yp1, yp2, xm1ym1, xm1yp1, xp1ym1, xp1yp1;
	double vxFvx_xp1, vxFvy_xp1;
	double CvxFvxR, CvxFvxL, CvyFvxT, CvyFvxB, CvxFvyR, CvxFvyL, CvyFvyT, CvyFvyB, px, py;
	double gx, gy;
	double vx_xp1, vx_xp2, vx_xm2, vx_yp1, vx_yp2, vx_ym2,
		vy_xp1, vy_xp2, vy_xm2, vy_yp1, vy_yp2, vy_ym2, vy_xp1ym1;
	double du2dx2A, du2dy2A, dv2dx2A, dv2dy2A, du2dx2B, du2dy2B, dv2dx2B, dv2dy2B;
	int nx, ny;

	ny = subgrid.ny;
	nx = subgrid.nx;


	gx=alp*dt/(dx*dx);
	gy=alp*dt/(dy*dy);
	gx=0.0;
	gy=0.0;

/* The reason I have set gx and gy to zero below is because in Dr. Moore's email,
   she says,

     "In the velocity computation, the quadratic differencing is used except
     in the last row, which is NI for v and NI+1 for u.  In this row,
     diffusion terms are ignored and convection is based on first-order
     upwinding."

   I think I removed them because they represented the diffusion terms since they
   are associated with the second derivatives of u and v with respect to x and y.
*/


	for( dum = 0; dum < bc.num_free2[0].xp; ++dum )
	{
		node_num = bc.free2.xp[dum];
		i = node[node_num].i;
		j = node[node_num].j;

		ce  = ny*i + j;
		xm1 = ny*(i-1) + j;
		xm2 = ny*(i-2) + j;
		xp1 = ny*(i+1) + j;
		ym1 = ny*i + (j-1);
		ym2 = ny*i + (j-2);
		yp1 = ny*i + (j+1);
		yp2 = ny*i + (j+2);
		xm1ym1 = ny*(i-1) + (j-1);
		xm1yp1 = ny*(i-1) + (j+1);
		xp1ym1 = ny*(i+1) + (j-1);
		xp1yp1 = ny*(i+1) + (j+1);

		vx_xp1 = 2.0*node[ce].vx - node[xm1].vx;
		vx_xp1 = (18.0*node[ce].vx - 9.0*node[xm1].vx + 2.0*node[xm2].vx)/11.0;
		vy_xp1 = 2.0*node[ce].vy - node[xm1].vy;
		vy_xp1 = (18.0*node[ce].vy - 9.0*node[xm1].vy + 2.0*node[xm2].vy)/11.0;
		vy_xp1ym1 = 2.0*node[ym1].vy - node[xm1ym1].vy;

		vy_xp2 = (18.0*vy_xp1 - 9.0*node[ce].vy + 2.0*node[xm1].vy)/11.0;
		vx_xp2 = (18.0*vx_xp1 - 9.0*node[ce].vx + 2.0*node[xm1].vx)/11.0;

		vy_ym2 = (18.0*node[ym1].vy - 9.0*node[ce].vy + 2.0*node[yp1].vy)/11.0;
		vx_ym2 = (18.0*node[ym1].vx - 9.0*node[ce].vx + 2.0*node[yp1].vx)/11.0;
		if( j > 1)
		{
		    vy_ym2 = node[ym2].vy;
		    vx_ym2 = node[ym2].vx;
		}

		vy_yp2 = (18.0*node[yp1].vy - 9.0*node[ce].vy + 2.0*node[ym1].vy)/11.0;
		vx_yp2 = (18.0*node[yp1].vx - 9.0*node[ce].vx + 2.0*node[ym1].vx)/11.0;
		if( j < ny-2)
		{
		    vy_yp2 = node[yp2].vy;
		    vx_yp2 = node[yp2].vx;
		}

		vxFvx_xp1 = .5*(vx_xp1 + node[ce].vx) - q/3.0*(node[xm1].vx - 2.0*node[ce].vx + vx_xp1);
		vxFvy_xp1 = .5*(node[yp1].vx + node[ce].vx) - q/3.0*(node[ym1].vx - 2.0*node[ce].vx + node[yp1].vx);

#if 0
/*
   Unlike in the true quickest algorithm, I do not use the face velocities at the boundary.
   This improves the solution and makes it more stable.  I cannot justify this numerically,
   as it is something I just stumbled on.  Note also that I use the same points for both
   vx and vy.  If you look in the quickest algorithm, the lines below are given by:
*/

		CvxFvxR = dt*node[xp1].vxFvx/dx;  /* for the right (actually, dt*vxFvx_xp1/dx */
		CvxFvxL = dt*node[ce].vxFvx/dx;   /* for the left */
		CvyFvxT = dt*node[yp1].vyFvx/dy;  /* for the top */
		CvyFvxB = dt*node[ce].vyFvx/dy;   /* for the bottom */

		CvxFvyR = dt*node[xp1].vxFvy/dx;  /* for the right (actually, dt*vxFvy_xp1/dx */
		CvxFvyL = dt*node[ce].vxFvy/dx;   /* for the left */
		CvyFvyT = dt*node[yp1].vyFvy/dy;  /* for the top */
		CvyFvyB = dt*node[ce].vyFvy/dy;   /* for the bottom */
#endif

		CvxFvxR = dt*node[ce].vx/dx;
		CvxFvxL = dt*node[xm1].vx/dx;
		CvyFvxT = dt*node[ce].vy/dy;
		CvyFvxB = dt*node[ym1].vy/dy;

		CvxFvyR = dt*node[ce].vx/dx;
		CvxFvyL = dt*node[xm1].vx/dx;
		CvyFvyT = dt*node[ce].vy/dy;
		CvyFvyB = dt*node[ym1].vy/dy;

/* I don't shift the derivatives with respect to x because that would involve vx_xp2 and vy_xp2 which
   is beyond the point I want to extrapolate.
*/

		/*velocity_flag = 0;*/

		{
		   du2dx2A = node[xm1].vx - 2.0*node[ce].vx + vx_xp1;
		   du2dx2B = node[xm2].vx - 2.0*node[xm1].vx + node[ce].vx;
		   dv2dx2A = node[xm1].vy - 2.0*node[ce].vy + vy_xp1;
		   dv2dx2B = node[xm2].vy - 2.0*node[xm1].vy + node[ce].vy;
		}
		if( node[ce].vy >= 0.0 )
		{
		   du2dy2A = node[ym1].vx - 2.0*node[ce].vx + node[yp1].vx;
		   du2dy2B = vx_ym2 - 2.0*node[ym1].vx + node[ce].vx;
		   dv2dy2A = node[ym1].vy - 2.0*node[ce].vy + node[yp1].vy;
		   dv2dy2B = vy_ym2 - 2.0*node[ym1].vy + node[ce].vy;
		}
		if( node[ce].vy < 0.0 )
		{
		   du2dy2A = node[ce].vx - 2.0*node[yp1].vx + vx_yp2;
		   du2dy2B = node[ym1].vx - 2.0*node[ce].vx + node[yp1].vx;
		   dv2dy2A = node[ce].vy - 2.0*node[yp1].vy + vy_yp2;
		   dv2dy2B = node[ym1].vy - 2.0*node[ce].vy + node[yp1].vy;
		   velocity_flag = 1;
		}

		/*printf(  "uuuuu %d %d %d %d %16.8e %16.8e\n", ce, xm1, xm2, velocity_flag, dv2dy2A, dv2dy2B);*/

		node[ce].vxQ = node[ce].vx
		     - CvxFvxR*(.5*(node[ce].vx + vx_xp1) - .5*CvxFvxR*(vx_xp1 - node[ce].vx)
		     - ( (1.0 - CvxFvxR*CvxFvxR)/6.0 - gx )*du2dx2A)

		     + CvxFvxL*(.5*(node[xm1].vx + node[ce].vx) - .5*CvxFvxL*(node[ce].vx - node[xm1].vx)
		     - ( (1.0 - CvxFvxL*CvxFvxL)/6.0 - gx )*du2dx2B)

		     - CvyFvxT*(.5*(node[ce].vx + node[yp1].vx) - .5*CvyFvxT*(node[yp1].vx - node[ce].vx)
		     - ( (1.0 - CvyFvxT*CvyFvxT)/6.0 - gy )*du2dy2A)

		     + CvyFvxB*(.5*(node[ym1].vx + node[ce].vx) - .5*CvyFvxB*(node[ce].vx - node[ym1].vx)
		     - ( (1.0 - CvyFvxB*CvyFvxB)/6.0 - gy )*du2dy2B)

		     + gx*(node[xm1].vx - 2.0*node[ce].vx + vx_xp1)
		     + gy*(node[ym1].vx - 2.0*node[ce].vx + node[yp1].vx);

		node[ce].vyQ = node[ce].vy
		     - CvxFvyR*(.5*(node[ce].vy + vy_xp1) - .5*CvxFvyR*(vy_xp1 - node[ce].vy)
		     - ( (1.0 - CvxFvyR*CvxFvyR)/6.0 - gx )*dv2dx2A)

		     + CvxFvyL*(.5*(node[xm1].vy + node[ce].vy) - .5*CvxFvyL*(node[ce].vy - node[xm1].vy)
		     - ( (1.0 - CvxFvyL*CvxFvyL)/6.0 - gx )*dv2dx2B)

		     - CvyFvyT*(.5*(node[ce].vy + node[yp1].vy) - .5*CvyFvyT*(node[yp1].vy - node[ce].vy)
		     - ( (1.0 - CvyFvyT*CvyFvyT)/6.0 - gy )*dv2dy2A)

		     + CvyFvyB*(.5*(node[ym1].vy + node[ce].vy) - .5*CvyFvyB*(node[ce].vy - node[ym1].vy)
		     - ( (1.0 - CvyFvyB*CvyFvyB)/6.0 - gy )*dv2dy2B)

		     + gx*(node[xm1].vy - 2.0*node[ce].vy + vy_xp1)
		     + gy*(node[ym1].vy - 2.0*node[ce].vy + node[yp1].vy);


		/*node[ce].prs = 2.0*node[xm1].prs - node[xm2].prs;
		node[ce].prs = 2.0*node[xm2].prs - node[xm3].prs;
		node[ce].prs = node[xm1].prs;*/
	}


	for( dum = 0; dum < bc.num_free2[0].yp; ++dum )
	{
		node_num = bc.free2.yp[dum];
		i = node[node_num].i;
		j = node[node_num].j;

		ce  = ny*i + j;
		xm1 = ny*(i-1) + j;
		xm2 = ny*(i-2) + j;
		xp1 = ny*(i+1) + j;
		xp2 = ny*(i+2) + j;
		ym1 = ny*i + (j-1);
		ym2 = ny*i + (j-2);
		yp1 = ny*i + (j+1);
		xm1ym1 = ny*(i-1) + (j-1);
		xm1yp1 = ny*(i-1) + (j+1);
		xp1ym1 = ny*(i+1) + (j-1);
		xp1yp1 = ny*(i+1) + (j+1);

		vx_yp1 = 2.0*node[ce].vx - node[ym1].vx;
		vx_yp1 = (18.0*node[ce].vx - 9.0*node[ym1].vx + 2.0*node[ym2].vx)/11.0;
		vy_yp1 = 2.0*node[ce].vy - node[ym1].vy;
		vy_yp1 = (18.0*node[ce].vy - 9.0*node[ym1].vy + 2.0*node[ym2].vy)/11.0;

		vy_xm2 = (18.0*node[xm1].vy - 9.0*node[ce].vy + 2.0*node[xp1].vy)/11.0;
		vx_xm2 = (18.0*node[xm1].vx - 9.0*node[ce].vx + 2.0*node[xp1].vx)/11.0;
		if( i > 1)
		{
		    vy_xm2 = node[xm2].vy;
		    vx_xm2 = node[xm2].vx;
		}

		vy_xp2 = (18.0*node[xp1].vy - 9.0*node[ce].vy + 2.0*node[xm1].vy)/11.0;
		vx_xp2 = (18.0*node[xp1].vx - 9.0*node[ce].vx + 2.0*node[xm1].vx)/11.0;
		if( i < nx-2)
		{
		    vy_xp2 = node[xp2].vy;
		    vx_xp2 = node[xp2].vx;
		}

#if 0
/*
   Unlike in the true quickest algorithm, I do not use the face velocities at the boundary.
   This improves the solution and makes it more stable.  I cannot justify this numerically,
   as it is something I just stumbled on.  Note also that I use the same points for both
   vx and vy.  If you look in the quickest algorithm, the lines below are given by:
*/

		CvxFvxR = dt*node[xp1].vxFvx/dx;  /* for the right */
		CvxFvxL = dt*node[ce].vxFvx/dx;   /* for the left */
		CvyFvxT = dt*node[yp1].vyFvx/dy;  /* for the top (actually, dt*vyFvx_yp1/dx */
		CvyFvxB = dt*node[ce].vyFvx/dy;   /* for the bottom */

		CvxFvyR = dt*node[xp1].vxFvy/dx;  /* for the right */
		CvxFvyL = dt*node[ce].vxFvy/dx;   /* for the left */
		CvyFvyT = dt*node[yp1].vyFvy/dy;  /* for the top (actually, dt*vyFvy_yp1/dx */
		CvyFvyB = dt*node[ce].vyFvy/dy;   /* for the bottom */

#endif

		CvxFvxR = dt*node[ce].vx/dx;
		CvxFvxL = dt*node[xm1].vx/dx;
		CvyFvxT = dt*node[ce].vy/dy;
		CvyFvxB = dt*node[ym1].vy/dy;

		CvxFvyR = dt*node[ce].vx/dx;
		CvxFvyL = dt*node[xm1].vx/dx;
		CvyFvyT = dt*node[ce].vy/dy;
		CvyFvyB = dt*node[ym1].vy/dy;

		if( node[ce].vx >= 0.0 )
		{
		   du2dx2A = node[xm1].vx - 2.0*node[ce].vx + node[xp1].vx;
		   du2dx2B = vx_xm2 - 2.0*node[xm1].vx + node[ce].vx;
		   dv2dx2A = node[xm1].vy - 2.0*node[ce].vy + node[xp1].vy;
		   dv2dx2B = vy_xm2 - 2.0*node[xm1].vy + node[ce].vy;
		}
		if( node[ce].vx < 0.0 )
		{
		   du2dx2A = node[ce].vx - 2.0*node[xp1].vx + vx_xp2;
		   du2dx2B = node[xm1].vx - 2.0*node[ce].vx + node[xp1].vx;
		   dv2dx2A = node[ce].vy - 2.0*node[xp1].vy + vy_xp2;
		   dv2dx2B = node[xm1].vy - 2.0*node[ce].vy + node[xp1].vy;
		   velocity_flag = 1;
		}
/* I don't shift the derivatives with respect to y because that would involve vx_yp2 and vy_yp2 which
   is beyond the point I want to extrapolate.
*/

		{
		   du2dy2A = node[ym1].vx - 2.0*node[ce].vx + vx_yp1;
		   du2dy2B = node[ym2].vx - 2.0*node[ym1].vx + node[ce].vx;
		   dv2dy2A = node[ym1].vy - 2.0*node[ce].vy + vy_yp1;
		   dv2dy2B = node[ym2].vy - 2.0*node[ym1].vy + node[ce].vy;
		}
		/*printf(  "uuuuu %d %d %d %d %16.8e %16.8e\n", ce, xm1, xm2, velocity_flag, du2dy2A, du2dy2B);*/

		node[ce].vxQ = node[ce].vx
		     - CvxFvxR*(.5*(node[ce].vx + node[xp1].vx) - .5*CvxFvxR*(node[xp1].vx - node[ce].vx)
		     - ( (1.0 - CvxFvxR*CvxFvxR)/6.0 - gx )*du2dx2A)

		     + CvxFvxL*(.5*(node[xm1].vx + node[ce].vx) - .5*CvxFvxL*(node[ce].vx - node[xm1].vx)
		     - ( (1.0 - CvxFvxL*CvxFvxL)/6.0 - gx )*du2dx2B)

		     - CvyFvxT*(.5*(node[ce].vx + vx_yp1) - .5*CvyFvxT*(vx_yp1 - node[ce].vx)
		     - ( (1.0 - CvyFvxT*CvyFvxT)/6.0 - gy )*du2dy2A)

		     + CvyFvxB*(.5*(node[ym1].vx + node[ce].vx) - .5*CvyFvxB*(node[ce].vx - node[ym1].vx)
		     - ( (1.0 - CvyFvxB*CvyFvxB)/6.0 - gy )*du2dy2B)

		     + gx*(node[xm1].vx - 2.0*node[ce].vx + node[xp1].vx)
		     + gy*(node[ym1].vx - 2.0*node[ce].vx + vx_yp1);

		node[ce].vyQ = node[ce].vy
		     - CvxFvyR*(.5*(node[ce].vy + node[xp1].vy) - .5*CvxFvyR*(node[xp1].vy - node[ce].vy)
		     - ( (1.0 - CvxFvyR*CvxFvyR)/6.0 - gx )*dv2dx2A)

		     + CvxFvyL*(.5*(node[xm1].vy + node[ce].vy) - .5*CvxFvyL*(node[ce].vy - node[xm1].vy)
		     - ( (1.0 - CvxFvyL*CvxFvyL)/6.0 - gx )*dv2dx2B)

		     - CvyFvyT*(.5*(node[ce].vy + vy_yp1) - .5*CvyFvyT*(vy_yp1 - node[ce].vy)
		     - ( (1.0 - CvyFvyT*CvyFvyT)/6.0 - gy )*dv2dy2A)

		     + CvyFvyB*(.5*(node[ym1].vy + node[ce].vy) - .5*CvyFvyB*(node[ce].vy - node[ym1].vy)
		     - ( (1.0 - CvyFvyB*CvyFvyB)/6.0 - gy )*dv2dy2B)

		     + gx*(node[xm1].vy - 2.0*node[ce].vy + node[xp1].vy)
		     + gy*(node[ym1].vy - 2.0*node[ce].vy + vy_yp1);

	}

	return 1;

}

#endif


int applyBC_freestream( BOUND bc, INC_NODE *node, SUBGRID subgrid)
{
/* Below is the bondary condition for the upper and lower part of the
   grid.  It reprsents saying the velocity is the freestream
   velocity at the top and bottom of the grid, i.e. dv/dn and du/dn
   are zero where n is the normal to the top of the grid.

                     Updated 4/8/08
*/

	int i, j, k, dum, node_num;
	int ce, xm1, xm2, xp1, xp2, ym1, ym2, yp1, yp2;
	int nx, ny;

	ny = subgrid.ny;
	nx = subgrid.nx;

	for( dum = 0; dum < bc.num_frestr[0].xp; ++dum )
	{
		node_num = bc.frestr.xp[dum];
		i = node[node_num].i;
		j = node[node_num].j;

		ce  = ny*i + j;
		xm1  = ny*(i-1) + j;

		node[ce].vxQ = node[xm1].vxQ;
		node[ce].vyQ = node[xm1].vyQ;
	}

	for( dum = 0; dum < bc.num_frestr[0].xm; ++dum )
	{
		node_num = bc.frestr.xm[dum];
		i = node[node_num].i;
		j = node[node_num].j;

		ce  = ny*i + j;
		xp1  = ny*(i+1) + j;

		node[ce].vxQ = node[xp1].vxQ;
		node[ce].vyQ = node[xp1].vyQ;
	}

	for( dum = 0; dum < bc.num_frestr[0].yp; ++dum )
	{
		node_num = bc.frestr.yp[dum];
		i = node[node_num].i;
		j = node[node_num].j;

		ce  = ny*i + j;
		ym1  = ny*i + j-1;

		node[ce].vxQ = node[ym1].vxQ;
		node[ce].vyQ = node[ym1].vyQ;
	}

	for( dum = 0; dum < bc.num_frestr[0].ym; ++dum )
	{
		node_num = bc.frestr.ym[dum];
		i = node[node_num].i;
		j = node[node_num].j;

		ce  = ny*i + j;
		yp1  = ny*i + j+1;

		node[ce].vxQ = node[yp1].vxQ;
		node[ce].vyQ = node[yp1].vyQ;
	}

	 return 1;

}

int applyBC_free( BOUND bc, INC_NODE *node, SUBGRID subgrid)
{
/*
   Set free boundary conditions.  This function is not used.  It was commonly
   used in my compressible code, but the boundary conditions for the
   incompressible code are much more complex.   I leave it in as a future
   placeholder, but note that it is very incomplete and only written for the "xm"
   case.

                     Updated 4/8/08
*/

	int i, j, k, dum, node_num;
	int ce, xp1, xp2, xm1, xm2, yp1, yp2, ym1, ym2;
	int nx, ny;

	ny = subgrid.ny;
	nx = subgrid.nx;

	for( dum = 0; dum < bc.num_free[0].xp; ++dum)
	{
	    node_num = bc.free.xp[dum];
	    i = node[node_num].i;
	    j = node[node_num].j;

	    ce = ny*i + j;
	    xm1  = ny*(i-1) + j;
	    xm2  = ny*(i-2) + j;

	    node[ce].vxQ = 2.0*node[xm1].vxQ - node[xm2].vxQ;
	    node[ce].vyQ = 2.0*node[xm1].vyQ - node[xm2].vyQ;

	}

	return 1;
}


int applyBC_noslip( BOUND bc, INC_NODE *node, SUBGRID subgrid)
{
/* The no slip BC is applied.  Because I am using a staggard grid, there are many nodes
   which do not fall on the edge nodes of the obstruction.  So for some nodes, I have to use
   extrapolation.

   The boundary condtions for noslip also apply to some pressure nodes.  The physical condition is
   that the pressure gradient normal to the surface is zero.  This is given on page 340-341,
   section 17.1.3 in:

     Fletcher, Clive A. J., Computational Techniques for Fluid Dynamics: Vol I, 2nd
        Ed., Springer-Verlag, Berlin, 1991.

   I do not apply it exactly as givin by Fletcher because what I currently do gives better
   results.  Overall, this pressure condition slightly improves the results.

   All of these boundary conditions are mixed together below, except for setting the velocity
   to zero inside the body.
   
                     Updated 9/2/08
*/

	int i, j, k, dum, node_num;
	int ce, xm1, xm2, xp1, xp2, ym1, ym2, yp1, yp2, xm1ym1, xm1yp1, xm2ym1, xp1ym1, xp1yp1;
	int nx, ny;

	ny = subgrid.ny;
	nx = subgrid.nx;


#if 1
/* Set nodes internal to the body to zero */

	for( dum = 0; dum < bc.num_vel0[0].x; ++dum)
	{
		node_num = bc.vel0.x[dum];
		i = node[node_num].i;
		j = node[node_num].j;
		ce =  ny*i + j;
		node[ce].vx = 0.0;
		node[ce].vxQ = 0.0;
	}

	for( dum = 0; dum < bc.num_vel0[0].y; ++dum)
	{
		node_num = bc.vel0.y[dum];
		i = node[node_num].i;
		j = node[node_num].j;
		ce =  ny*i + j;
		node[ce].vy = 0.0;
		node[ce].vyQ = 0.0;
	}
#endif

	if( bc.num_noslp[0].xp > 0)
	{
/* noslip in x + */

	    dum = 0;
	    node_num = bc.noslp.xp[dum];
	    i = node[node_num].i;
	    j = node[node_num].j;
	    ce  = ny*i + j;
	    xm1 = ny*(i-1) + j;
	    xp1 = ny*(i+1) + j;

	    node[ce].vx = 0.0;     /* For the noslip vx velocities */
	    node[ce].vxQ = 0.0;

	    node[ce].vy  = -node[xp1].vy/3.0;   /* For the vy velocities immediately to the left */
	    node[ce].vyQ = -node[xp1].vy/3.0;   /* of the noslip velocities */

	    node[xp1].vxFvy = 0.0;   /* For the vxFvy velocities */

	    node[ce].prs = node[xp1].prs + 2.0*node[xm1].vxQ*alp/dx;   /* For the pressure immediately
	                                                                  to the left */

	    for( dum = 1; dum < bc.num_noslp[0].xp-1; ++dum)
	    {
		node_num = bc.noslp.xp[dum];
		i = node[node_num].i;
		j = node[node_num].j;
		ce  = ny*i + j;
		xm1 = ny*(i-1) + j;
		xp1 = ny*(i+1) + j;

		node[ce].vx = 0.0;
		node[ce].vxQ = 0.0;

		node[ce].vy  = -node[xp1].vy/3.0;   /* For the vy velocities immediately to the left */
		node[ce].vyQ = -node[xp1].vy/3.0;   /* of the noslip velocities */

		node[xp1].vxFvy = 0.0;   /* For the vxFvy velocities */

		node[ce].vyFvx  = 0.0;   /* For the vyFvx velocities */

		node[ce].prs = node[xp1].prs + 2.0*node[xm1].vxQ*alp/dx;   /* For the pressure immediately
	                                                                      to the left */
	    }

	    dum = bc.num_noslp[0].xp-1;
	    node_num = bc.noslp.xp[dum];
	    i = node[node_num].i;
	    j = node[node_num].j;
	    ce  = ny*i + j;
	    xm1 = ny*(i-1) + j;
	    xp1 = ny*(i+1) + j;

	    node[ce].vx = 0.0;
	    node[ce].vxQ = 0.0;

	    node[ce].vyFvx  = 0.0;   /* For the vyFvx velocities */

	    node[ce].prs = node[xp1].prs + 2.0*node[xm1].vxQ*alp/dx;   /* For the pressure immediately
	                                                                  to the left */
	}

	if( bc.num_noslp[0].xm > 0)
	{

/* noslip in x - */

	    dum = 0;
	    node_num = bc.noslp.xm[dum];
	    i = node[node_num].i;
	    j = node[node_num].j;
	    ce  = ny*i + j;
	    xm1 = ny*(i-1) + j;
	    xp1 = ny*(i+1) + j;
	    xp2 = ny*(i+2) + j;

	    node[ce].vx = 0.0;     /* For the noslip vx velocities */
	    node[ce].vxQ = 0.0;

	    node[xp1].vy  = -node[ce].vy/3.0;   /* For the vy velocities immediately to the right */
	    node[xp1].vyQ = -node[ce].vy/3.0;   /* of the noslip velocities */

	    node[xp1].vxFvy = 0.0;   /* For the vxFvy velocities */

	    node[xp1].prs = node[ce].prs + 2.0*node[xp1].vxQ*alp/dx;   /* For the pressure immediately
	                                                                  to the right */

	    for( dum = 1; dum < bc.num_noslp[0].xm-1; ++dum)
	    {
		node_num = bc.noslp.xm[dum];
		i = node[node_num].i;
		j = node[node_num].j;
		ce  = ny*i + j;
		xm1 = ny*(i-1) + j;
		xp1 = ny*(i+1) + j;
		xp2 = ny*(i+2) + j;

		node[ce].vx = 0.0;
		node[ce].vxQ = 0.0;

		node[xp1].vy  = -node[ce].vy/3.0;   /* For the vy velocities immediately to the right */
		node[xp1].vyQ = -node[ce].vy/3.0;   /* of the noslip velocities */

		node[xp1].vxFvy = 0.0;   /* For the vxFvy velocities */

		node[ce].vyFvx  = 0.0;   /* For the vyFvx velocities */

		node[xp1].prs = node[ce].prs + 2.0*node[xp1].vxQ*alp/dx;   /* For the pressure immediately
	                                                                      to the right */
	    }

	    dum = bc.num_noslp[0].xm-1;
	    node_num = bc.noslp.xm[dum];
	    i = node[node_num].i;
	    j = node[node_num].j;
	    ce  = ny*i + j;
	    xm1 = ny*(i-1) + j;
	    xp1 = ny*(i+1) + j;
	    xp2 = ny*(i+2) + j;

	    node[ce].vx = 0.0;
	    node[ce].vxQ = 0.0;

	    node[ce].vyFvx  = 0.0;   /* For the vyFvx velocities */

	    node[xp1].prs = node[ce].prs + 2.0*node[xp1].vxQ*alp/dx;   /* For the pressure immediately
	                                                                  to the right */
	}

	if( bc.num_noslp[0].yp > 0)
	{

/* noslip in y + */

	    dum = 0;
	    node_num = bc.noslp.yp[dum];
	    i = node[node_num].i;
	    j = node[node_num].j;
	    ce  = ny*i + j;
	    ym1 = ny*i + j-1;
	    yp1 = ny*i + j+1;

	    node[ce].vy = 0.0;     /* For the noslip vx velocities */
	    node[ce].vyQ = 0.0;

	    node[ce].vx  = -node[yp1].vx/3.0;   /* For the vy velocities immediately bolow the top */
	    node[ce].vxQ = -node[yp1].vx/3.0;   /* of the noslip velocities */

	    node[yp1].vyFvx = 0.0;   /* For the vxFvy velocities */

	    node[ce].prs = node[yp1].prs + 2.0*node[ym1].vyQ*alp/dy;   /* For the pressure immediately
	                                                                  bolow the top edge */

	    for( dum = 1; dum < bc.num_noslp[0].yp-1; ++dum)
	    {
		node_num = bc.noslp.yp[dum];
		i = node[node_num].i;
		j = node[node_num].j;
		ce  = ny*i + j;
		ym1 = ny*i + j-1;
		yp1 = ny*i + j+1;

		node[ce].vy = 0.0;
		node[ce].vyQ = 0.0;

		node[ce].vx  = -node[yp1].vx/3.0;   /* For the vy velocities immediately bolow the top */
		node[ce].vxQ = -node[yp1].vx/3.0;   /* of the noslip velocities */

		node[yp1].vyFvx = 0.0;   /* For the vxFvy velocities */

		node[ce].vxFvy  = 0.0;   /* For the vyFvx velocities */

		node[ce].prs = node[yp1].prs + 2.0*node[ym1].vyQ*alp/dy;   /* For the pressure immediately
	                                                                      bolow the top edge */
	    }

	    dum = bc.num_noslp[0].yp-1;
	    node_num = bc.noslp.yp[dum];
	    i = node[node_num].i;
	    j = node[node_num].j;
	    ce  = ny*i + j;
	    ym1 = ny*i + j-1;
	    yp1 = ny*i + j+1;

	    node[ce].vy = 0.0;
	    node[ce].vyQ = 0.0;

	    node[ce].vxFvy  = 0.0;   /* For the vyFvx velocities */

	    node[ce].prs = node[yp1].prs + 2.0*node[ym1].vyQ*alp/dy;   /* For the pressure immediately
	                                                                  bolow the top edge */

	}
	if( bc.num_noslp[0].ym > 0)
	{

/*     noslip in y - */

	    dum = 0;
	    node_num = bc.noslp.ym[dum];
	    i = node[node_num].i;
	    j = node[node_num].j;
	    ce  = ny*i + j;
	    ym1 = ny*i + j-1;
	    yp1 = ny*i + j+1;
	    yp2 = ny*i + j+2;

	    node[ce].vy = 0.0;     /* For the noslip vx velocities */
	    node[ce].vyQ = 0.0;

	    node[yp1].vx  = -node[ce].vx/3.0;   /* For the vy velocities immediately above the bottom edge */
	    node[yp1].vxQ = -node[ce].vx/3.0;   /* of the noslip velocities */

	    node[yp1].vyFvx = 0.0;   /* For the vxFvy velocities */

	    node[yp1].prs = node[ce].prs + 2.0*node[yp1].vyQ*alp/dy;   /* For the pressure immediately
	                                                                  above the bottom edge */

	    for( dum = 1; dum < bc.num_noslp[0].ym-1; ++dum)
	    {
		node_num = bc.noslp.ym[dum];
		i = node[node_num].i;
		j = node[node_num].j;
		ce  = ny*i + j;
		ym1 = ny*i + j-1;
		yp1 = ny*i + j+1;
		yp2 = ny*i + j+2;

		node[ce].vy = 0.0;
		node[ce].vyQ = 0.0;

		node[yp1].vx  = -node[ce].vx/3.0;   /* For the vy velocities immediately above the bottom edge */
		node[yp1].vxQ = -node[ce].vx/3.0;   /* of the noslip velocities */

		node[yp1].vyFvx = 0.0;   /* For the vxFvy velocities */

		node[ce].vxFvy  = 0.0;   /* For the vyFvx velocities */

		node[yp1].prs = node[ce].prs + 2.0*node[yp1].vyQ*alp/dy;   /* For the pressure immediately
	                                                                      above the bottom edge */
	    }

	    dum = bc.num_noslp[0].ym-1;
	    node_num = bc.noslp.ym[dum];
	    i = node[node_num].i;
	    j = node[node_num].j;
	    ce  = ny*i + j;
	    ym1 = ny*i + j-1;
	    yp1 = ny*i + j+1;
	    yp2 = ny*i + j+2;

	    node[ce].vy = 0.0;
	    node[ce].vyQ = 0.0;

	    node[ce].vxFvy  = 0.0;   /* For the vyFvx velocities */

	    node[yp1].prs = node[ce].prs + 2.0*node[yp1].vyQ*alp/dy;   /* For the pressure immediately
	                                                                  above the bottom edge */
	}


/* These are for the corner nodes where there is overlap between noslip in x and y. */

	for( dum = 0; dum < bc.num_noslp_prs[0].xpyp; ++dum )
	{
	    node_num = bc.noslp_prs.xpyp[0];
	    i = node[node_num].i;
	    j = node[node_num].j;
	    ce  = ny*i + j;
	    xm1 = ny*(i-1) + j;
	    xp1 = ny*(i+1) + j;
	    ym1 = ny*i + j-1;
	    yp1 = ny*i + j+1;
	    node[ce].prs = .5*(node[xp1].prs + 2.0*node[xm1].vxQ*alp/dx +
		node[yp1].prs + 2.0*node[ym1].vyQ*alp/dy);
	}
	for( dum = 0; dum < bc.num_noslp_prs[0].xpym; ++dum )
	{
	    node_num = bc.noslp_prs.xpym[0];
	    i = node[node_num].i;
	    j = node[node_num].j;
	    ce  = ny*i + j;
	    xm1 = ny*(i-1) + j;
	    xp1 = ny*(i+1) + j;
	    ym1 = ny*i + j-1;
	    yp1 = ny*i + j+1;
	    node[ce].prs = .5*(node[xp1].prs + 2.0*node[xm1].vxQ*alp/dx +
		node[ym1].prs + 2.0*node[ce].vyQ*alp/dy);
	}
	for( dum = 0; dum < bc.num_noslp_prs[0].xmyp; ++dum )
	{
	    node_num = bc.noslp_prs.xmyp[0];
	    i = node[node_num].i;
	    j = node[node_num].j;
	    ce  = ny*i + j;
	    xm1 = ny*(i-1) + j;
	    xp1 = ny*(i+1) + j;
	    ym1 = ny*i + j-1;
	    yp1 = ny*i + j+1;
	    node[ce].prs = .5*(node[xm1].prs + 2.0*node[ce].vxQ*alp/dx +
		node[yp1].prs + 2.0*node[ym1].vyQ*alp/dy);
	}
	for( dum = 0; dum < bc.num_noslp_prs[0].xmym; ++dum )
	{
	    node_num = bc.noslp_prs.xmym[0];
	    i = node[node_num].i;
	    j = node[node_num].j;
	    ce  = ny*i + j;
	    xm1 = ny*(i-1) + j;
	    xp1 = ny*(i+1) + j;
	    ym1 = ny*i + j-1;
	    yp1 = ny*i + j+1;
	    node[ce].prs = .5*(node[xm1].prs + 2.0*node[ce].vxQ*alp/dx +
		node[ym1].prs + 2.0*node[ce].vyQ*alp/dy);
	}

	return 1;

}

int reset( INC_NODE *node, SUBGRID subgrid)
{
/*
                     Updated 4/8/08
*/
	int i, j, k, dum;
	int ce;
	int nx, ny;

	ny = subgrid.ny;
	nx = subgrid.nx;

	for( i = 0; i < nx; ++i )
	{
	    for( j = 0; j < ny; ++j )
	    {
/*  Now the previous values of vx and vy are set to equal vxQ and vyQ.
*/
	      
		ce  = ny*i + j;
		node[ce].vx = node[ce].vxQ;
		node[ce].vy = node[ce].vyQ;
	    }
	}

     return 1;
}



int reset0( BOUND bc, INC_NODE *node, SUBGRID subgrid)
{
/*
The velocities vx0 and vy0 are used in the subroutine sor as a type of reference
velocity while sor updates vxQ and vyQ.  This slightly improves stability.

                     Updated 4/8/08
*/
	int i, j, k, dum, node_num;
	int ce;
	int nx, ny;

	ny = subgrid.ny;
	nx = subgrid.nx;

/* For the 2nd farfield boundary condition. */

	for( dum = 0; dum < bc.num_free2[0].xp; ++dum )
	{
		node_num = bc.free2.xp[dum];
		node[node_num].vx0 = node[node_num].vxQ;
		node[node_num].vy0 = node[node_num].vyQ;
	}
	for( dum = 0; dum < bc.num_free2[0].xm; ++dum )
	{
		node_num = bc.free2.xm[dum];
		node[node_num].vx0 = node[node_num].vxQ;
		node[node_num].vy0 = node[node_num].vyQ;
	}
	for( dum = 0; dum < bc.num_free2[0].yp; ++dum )
	{
		node_num = bc.free2.yp[dum];
		node[node_num].vx0 = node[node_num].vxQ;
		node[node_num].vy0 = node[node_num].vyQ;
	}
	for( dum = 0; dum < bc.num_free2[0].ym; ++dum )
	{
		node_num = bc.free2.ym[dum];
		node[node_num].vx0 = node[node_num].vxQ;
		node[node_num].vy0 = node[node_num].vyQ;
	}

/* For the freestream boundary condition. */

	for( dum = 0; dum < bc.num_frestr[0].xp; ++dum )
	{
		node_num = bc.frestr.xp[dum];
		node[node_num].vx0 = node[node_num].vxQ;
		node[node_num].vy0 = node[node_num].vyQ;
	}
	for( dum = 0; dum < bc.num_frestr[0].xm; ++dum )
	{
		node_num = bc.frestr.xm[dum];
		node[node_num].vx0 = node[node_num].vxQ;
		node[node_num].vy0 = node[node_num].vyQ;
	}
	for( dum = 0; dum < bc.num_frestr[0].yp; ++dum )
	{
		node_num = bc.frestr.yp[dum];
		node[node_num].vx0 = node[node_num].vxQ;
		node[node_num].vy0 = node[node_num].vyQ;
	}
	for( dum = 0; dum < bc.num_frestr[0].ym; ++dum )
	{
		node_num = bc.frestr.ym[dum];
		node[node_num].vx0 = node[node_num].vxQ;
		node[node_num].vy0 = node[node_num].vyQ;
	}

     return 1;
}


