/* This program calculates the next values of the velocities vx and v.

   The algorithm for the code is based on a modified version of the QUICKEST SCHEME.

Davis, R. W., and E. F. Moore, Journal of Fluid Mechanics, 116, 475 (1982).

Davis, R. W., E. F. Moore, and L. P. Purtell, Physics of Fluids, 27, 46 (1984).

   More is given on this method on page 347 of:

Fletcher, Clive A. J., Computational Techniques for Fluid Dynamics: Vol II, second
        edition, John Wiley & Sons Inc., New York, 1991.

                     Updated 6/10/08

    SLFCFD source file
    Version:  1.2
    Copyright (C) 2006  San Le

    The source code contained in this file is released under the
    terms of the GNU Library General Public License.

*/

#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include "../compressible/cfdconst.h"
#include "../compressible/cfdstruct.h"
#include "../AMR/amr.h"

extern double q, alp, Re, umax;
extern umax_i, umax_j;

int face_velocities( BOUND bc, INC_NODE *node, SUBGRID subgrid)
{
	int i, j, k, dum, dum2, node_num;
	int nx, ny;
	int ce, xm1, xm2, xp1, xp2, ym1, ym2, yp1, yp2;
	int xm1ym1, xm1yp1, xm1yp2, xp1ym1, xp1yp1, xp2ym1, xm1ym2, xm2ym1;
	double vx_xp1, vx_xm2, vy_xp1, vy_xm2, vx_yp1, vx_ym2, vy_yp1, vy_ym2;
	double vy_xp1ym1, vx_xm1yp1, vy_xp2ym1, vx_xm1yp2;
	double fdum;

	nx = subgrid.nx;
	ny = subgrid.ny;

	for( i = 1; i < nx; ++i )
	{
	    for( j = 1; j < ny; ++j )
	    {
		ce  = ny*i + j;
		xm1 = ny*(i-1) + j;
		xm2 = ny*(i-2) + j;
		xp1 = ny*(i+1) + j;
		xp2 = ny*(i+2) + j;
		ym1 = ny*i + (j-1);
		ym2 = ny*i + (j-2);
		yp1 = ny*i + (j+1);
		yp2 = ny*i + (j+2);
		xm1ym1 = ny*(i-1) + (j-1);
		xm1ym2 = ny*(i-1) + (j-2);
		xm1yp1 = ny*(i-1) + (j+1);
		xm1yp2 = ny*(i-1) + (j+2);
		xm2ym1 = ny*(i-2) + (j-1);
		xp1ym1 = ny*(i+1) + (j-1);
		xp1yp1 = ny*(i+1) + (j+1);
		xp2ym1 = ny*(i+2) + (j-1);

		if( i == 1)
		{
		    vy_xm2 = 2.0*node[xm1].vy - node[ce].vy;
		    vx_xm2 = 2.0*node[xm1].vx - node[ce].vx;
		    vy_xm2 = (18.0*node[xm1].vy - 9.0*node[ce].vy + 2.0*node[xp1].vy)/11.0;
		    vx_xm2 = (18.0*node[xm1].vx - 9.0*node[ce].vx + 2.0*node[xp1].vx)/11.0;
		}
		else
		{
		    vy_xm2 = node[xm2].vy;
		    vx_xm2 = node[xm2].vx;
		}

		if( j == 1)
		{
		    vy_ym2 = 2.0*node[ym1].vy - node[ce].vy;
		    vx_ym2 = 2.0*node[ym1].vx - node[ce].vx;
		    vy_ym2 = (18.0*node[ym1].vy - 9.0*node[ce].vy + 2.0*node[yp1].vy)/11.0;
		    vx_ym2 = (18.0*node[ym1].vx - 9.0*node[ce].vx + 2.0*node[yp1].vx)/11.0;
		}
		else
		{
		    vy_ym2 = node[ym2].vy;
		    vx_ym2 = node[ym2].vx;
		}

		if( i == nx-1)
		{
		    vy_xp1 = (18.0*node[ce].vy - 9.0*node[xm1].vy + 2.0*node[xm2].vy)/11.0;
		    vx_xp1 = (18.0*node[ce].vx - 9.0*node[xm1].vx + 2.0*node[xm2].vx)/11.0;
		    vy_xp1ym1 = (18.0*node[ym1].vy - 9.0*node[xm1ym1].vy + 2.0*node[xm2ym1].vy)/11.0;
		    vy_xp2ym1 = (18.0*vy_xp1ym1 - 9.0*node[ym1].vy + 2.0*node[xm1ym1].vy)/11.0;
		}
		else
		{
		    vy_xp1 = node[xp1].vy;
		    vx_xp1 = node[xp1].vx;
		    vy_xp1ym1 =  node[xp1ym1].vy;
		    if( i == nx-2) vy_xp2ym1 = (18.0*vy_xp1ym1 - 9.0*node[ym1].vy + 2.0*node[xm1ym1].vy)/11.0;
		    else vy_xp2ym1 = node[xp2ym1].vy;
		}

		if( j == ny-1)
		{
		    vy_yp1 = (18.0*node[ce].vy - 9.0*node[ym1].vy + 2.0*node[ym2].vy)/11.0;
		    vx_yp1 = (18.0*node[ce].vx - 9.0*node[ym1].vx + 2.0*node[ym2].vx)/11.0;
		    vx_xm1yp1 = (18.0*node[xm1].vx - 9.0*node[xm1ym1].vx + 2.0*node[xm1ym2].vx)/11.0;
		    vx_xm1yp2 = (18.0*vx_xm1yp1 - 9.0*node[xm1].vx + 2.0*node[xm1ym1].vx)/11.0;
		}
		else
		{
		    vy_yp1 = node[yp1].vy;
		    vx_yp1 = node[yp1].vx;
		    vx_xm1yp1 = node[xm1yp1].vx;
		    if( j == ny-2) vx_xm1yp2 = (18.0*vx_xm1yp1 - 9.0*node[xm1].vx + 2.0*node[xm1ym1].vx)/11.0;
		    else vx_xm1yp2 = node[xm1yp2].vx;
		}

/* The equations for upwinding vx and vy treat these velocities as the fluxed quantity at the center of
   the control volume.  But they are still placed at the staggered grid points which are different from
   the pressure.  This is confusing since the impression is that pressure is the only quantity located
   at the center of a cell.  But you should view vx and vy, which are the 2 fluxed quantities, as being at
   the center of their own control volumes, with the vxFvx, vyFvx, vxFvy, vyFvy, the veloctities on the
   faces of the control volume with:

   vxFvx = velocity vx on the vx control volume
   vyFvx = velocity vy on the vx control volume
   vxFvy = velocity vx on the vy control volume
   vyFvy = velocity vy on the vy control volume

   Also note that depending on whether node[ce].vxFvx and node[ce].vyFvy are positive or negative, I change
   the way node[ce].vxFvx, node[ce].vyFvx, node[ce].vyFvy, node[ce].vyFvy are calculated.  This is because
   when they are negative, the upwinding is now in the opposite direction.   The 1982 paper by Davis and
   Moore hinted at this change for the face velocities on page 479 equation (6) and (7).  Even though they
   did not seem to apply this themselves, I have applied it to both the face velocities as well as the main
   finite difference formula given on page 481, equation (16).  I elaborate more on this below for the main
   algorithm.

   San Le     Saturday, November 17th, 2007
*/

#if NEW
		node[ce].vxFvx = .5*(node[ce].vx + node[xm1].vx) - q/3.0*(vx_xm2 - 2.0*node[xm1].vx + node[ce].vx);
		if(node[ce].vxFvx < 0.0)
		    node[ce].vxFvx = .5*(node[ce].vx + node[xm1].vx) - q/3.0*(node[xm1].vx - 2.0*node[ce].vx + vx_xp1);

		if(node[ce].vxFvx < 0.0)
		    node[ce].vyFvx = .5*(vy_xp1ym1 + node[ym1].vy) - q/3.0*(node[ym1].vy - 2.0*vy_xp1ym1 + vy_xp2ym1);
		else
		    node[ce].vyFvx = .5*(vy_xp1ym1 + node[ym1].vy) - q/3.0*(node[xm1ym1].vy - 2.0*node[ym1].vy + vy_xp1ym1);

		node[ce].vyFvy = .5*(node[ce].vy + node[ym1].vy) - q/3.0*(vy_ym2 - 2.0*node[ym1].vy + node[ce].vy);
		if(node[ce].vyFvy < 0.0)
		    node[ce].vyFvy = .5*(node[ce].vy + node[ym1].vy) - q/3.0*(node[ym1].vy - 2.0*node[ce].vy + vy_yp1);

		if(node[ce].vyFvy < 0.0)
		    node[ce].vxFvy = .5*(vx_xm1yp1 + node[xm1].vx) - q/3.0*(node[xm1].vx - 2.0*vx_xm1yp1 + vx_xm1yp2);
		else
		    node[ce].vxFvy = .5*(vx_xm1yp1 + node[xm1].vx) - q/3.0*(node[xm1ym1].vx - 2.0*node[xm1].vx + vx_xm1yp1);
#endif

#if NEW2

/* Look below for more discussion of this modification in the algorithm.*/

		fdum = .5*(node[ce].vx + node[xm1].vx);
		if(fdum >= 0.0)
		{
		    node[ce].vxFvx = .5*(node[ce].vx + node[xm1].vx) - q/3.0*(vx_xm2 - 2.0*node[xm1].vx + node[ce].vx);
		    node[ce].vyFvx = .5*(vy_xp1ym1 + node[ym1].vy) - q/3.0*(node[xm1ym1].vy - 2.0*node[ym1].vy + vy_xp1ym1);
		}
		if(fdum < 0.0)
		{
		    node[ce].vxFvx = .5*(node[ce].vx + node[xm1].vx) - q/3.0*(node[xm1].vx - 2.0*node[ce].vx + vx_xp1);
		    node[ce].vyFvx = .5*(vy_xp1ym1 + node[ym1].vy) - q/3.0*(node[ym1].vy - 2.0*vy_xp1ym1 + vy_xp2ym1);
		}

		fdum = .5*(node[ce].vy + node[ym1].vy);
		if(fdum >= 0.0)
		{
		    node[ce].vyFvy = .5*(node[ce].vy + node[ym1].vy) - q/3.0*(vy_ym2 - 2.0*node[ym1].vy + node[ce].vy);
		    node[ce].vxFvy = .5*(vx_xm1yp1 + node[xm1].vx) - q/3.0*(node[xm1ym1].vx - 2.0*node[xm1].vx + vx_xm1yp1);
		}
		if(fdum < 0.0)
		{
		    node[ce].vyFvy = .5*(node[ce].vy + node[ym1].vy) - q/3.0*(node[ym1].vy - 2.0*node[ce].vy + vy_yp1);
		    node[ce].vxFvy = .5*(vx_xm1yp1 + node[xm1].vx) - q/3.0*(node[xm1].vx - 2.0*vx_xm1yp1 + vx_xm1yp2);
		}

#endif


#if ORIGINAL
		node[ce].vxFvx = .5*(node[ce].vx + node[xm1].vx) - q/3.0*(vx_xm2 - 2.0*node[xm1].vx + node[ce].vx);

		node[ce].vyFvx = .5*(vy_xp1ym1 + node[ym1].vy) - q/3.0*(node[xm1ym1].vy - 2.0*node[ym1].vy + vy_xp1ym1);

		node[ce].vyFvy = .5*(node[ce].vy + node[ym1].vy) - q/3.0*(vy_ym2 - 2.0*node[ym1].vy + node[ce].vy);

		node[ce].vxFvy = .5*(vx_xm1yp1 + node[xm1].vx) - q/3.0*(node[xm1ym1].vx - 2.0*node[xm1].vx + vx_xm1yp1);
#endif

	    }
	}

	return 1;
}



int quickest( BOUND bc, INC_NODE *node, SUBGRID subgrid)
{
	int i, j, k, dum, dum2, node_num, velocity_flag;
	int nx, ny, istart, iend, jstart, jend;
	int ce, xm1, xm2, xp1, xp2, ym1, ym2, yp1, yp2, xm1ym1, xm1yp1, xp1ym1, xp1yp1;
	double usq1, usq2, vsq1, vsq2, uv1, uvQ;
	double vx_xm2, vy_xm2, vx_ym2, vy_ym2, vx_xp2, vy_xp2, vx_yp2, vy_yp2;
	double CvxFvxR, CvxFvxL, CvyFvxT, CvyFvxB, CvxFvyR, CvxFvyL, CvyFvyT, CvyFvyB, px, py; 
	double gx, gy, perx, pery;
	double dx, dy, dt;
	double fdum1, fdum2, fdum3, fdum4, fdum5, fdum6, fdum7;
	double du2dx2A, du2dy2A, dv2dx2A, dv2dy2A, du2dx2B, du2dy2B, dv2dx2B, dv2dy2B;

	dx = subgrid.dx;
	dy = subgrid.dy;
	dt = subgrid.dt;
	nx = subgrid.nx;
	ny = subgrid.ny;

	gx=alp*dt/(dx*dx);
	gy=alp*dt/(dy*dy);

	for( i = 1; i < nx-1; ++i )
	{
	    for( j = 1; j < ny-1; ++j )
	    {
		ce  = ny*i + j;
		xm1 = ny*(i-1) + j;
		xm2 = ny*(i-2) + j;
		xp1 = ny*(i+1) + j;
		xp2 = ny*(i+2) + j;
		ym1 = ny*i + (j-1);
		ym2 = ny*i + (j-2);
		yp1 = ny*i + (j+1);
		yp2 = ny*i + (j+2);
		xm1ym1 = ny*(i-1) + (j-1);
		xm1yp1 = ny*(i-1) + (j+1);
		xp1ym1 = ny*(i+1) + (j-1);
		xp1yp1 = ny*(i+1) + (j+1);

		vy_xm2 = 2.0*node[xm1].vy - node[ce].vy;
		vx_xm2 = 2.0*node[xm1].vx - node[ce].vx;
		vy_xm2 = (18.0*node[xm1].vy - 9.0*node[ce].vy + 2.0*node[xp1].vy)/11.0;
		vx_xm2 = (18.0*node[xm1].vx - 9.0*node[ce].vx + 2.0*node[xp1].vx)/11.0;
		if( i > 1)
		{
		    vy_xm2 = node[xm2].vy;
		    vx_xm2 = node[xm2].vx;
		}

		vy_ym2 = 2.0*node[ym1].vy - node[ce].vy;
		vx_ym2 = 2.0*node[ym1].vx - node[ce].vx;
		vy_ym2 = (18.0*node[ym1].vy - 9.0*node[ce].vy + 2.0*node[yp1].vy)/11.0;
		vx_ym2 = (18.0*node[ym1].vx - 9.0*node[ce].vx + 2.0*node[yp1].vx)/11.0;
		if( j > 1)
		{
		    vy_ym2 = node[ym2].vy;
		    vx_ym2 = node[ym2].vx;
		}

		vy_yp2 = (18.0*node[yp1].vy - 9.0*node[ce].vy + 2.0*node[ym1].vy)/11.0;
		vx_yp2 = (18.0*node[yp1].vx - 9.0*node[ce].vx + 2.0*node[ym1].vx)/11.0;
		if( j < ny-2)
		{
		    vy_yp2 = node[yp2].vy;
		    vx_yp2 = node[yp2].vx;
		}

		vy_xp2 = (18.0*node[xp1].vy - 9.0*node[ce].vy + 2.0*node[xm1].vy)/11.0;
		vx_xp2 = (18.0*node[xp1].vx - 9.0*node[ce].vx + 2.0*node[xm1].vx)/11.0;
		if( i < nx-2)
		{
		    vy_xp2 = node[xp2].vy;
		    vx_xp2 = node[xp2].vx;
		}

/* The equations for upwinding vx and vy treat these velocities as the fluxed quantity at the center of
   the control volume.  But they are still placed at the staggered grid points which are different from
   the pressure.  This is confusing since the impression is that pressure is the only quantity located
   at the center of a cell.  But you should view vx and vy, which are the 2 fluxed quantities, as being at
   the center of their own control volumes, needing their own veloctities on the faces of the control
   volume.   These are given by:

   node[xp1].vxFvx    =  velocity in x on the right face for the vx control volume
   node[ce].vxFvx     =  velocity in x on the left face for the vx control volume
   node[yp1].vyFvx    =  velocity in y on the top face for the vx control volume
   node[ce].vyFvx     =  velocity in y on the bottom face for the vx control volume

   node[xp1].vxFvy    =  velocity in x on the right face for the vy control volume
   node[ce].vxFvy     =  velocity in x on the left face for the vy control volume
   node[yp1].vyFvy    =  velocity in y on the top face for the vy control volume
   node[ce].vyFvy     =  velocity in y on the bottom face for the vy control volume
*/

		CvxFvxR = dt*node[xp1].vxFvx/dx;  /* for the right */
		CvxFvxL = dt*node[ce].vxFvx/dx;   /* for the left */
		CvyFvxT = dt*node[yp1].vyFvx/dy;  /* for the top */
		CvyFvxB = dt*node[ce].vyFvx/dy;   /* for the bottom */

		CvxFvyR = dt*node[xp1].vxFvy/dx;  /* for the right */
		CvxFvyL = dt*node[ce].vxFvy/dx;   /* for the left */
		CvyFvyT = dt*node[yp1].vyFvy/dy;  /* for the top */
		CvyFvyB = dt*node[ce].vyFvy/dy;   /* for the bottom */

/* Below is the main part of the algorithm.  The significant change I made is in the
   terms for the 2nd derivaties.  Specifically:

        node[xm1].vx - 2.0*node[ce].vx + node[xp1].vx

        node[xm2].vx - 2.0*node[xm1].vx + node[ce].vx

        node[ym1].vx - 2.0*node[ce].vx + node[yp1].vx

        node[ym2].vx - 2.0*node[ym1].vx + node[ce].vx


        node[xm1].vy - 2.0*node[ce].vy + node[xp1].vy

        node[xm2].vy - 2.0*node[xm1].vy + node[ce].vy

        node[ym1].vy - 2.0*node[ce].vy + node[yp1].vy

        node[ym2].vy - 2.0*node[ym1].vy + node[ce].vy

   * Note that I use vx_xm2 instead of node[xm2].vx and vx_ym2 instead of node[ym2].vx because of the cases
     where x=1 or y =1.  This is also true for node[xm2].vy and node[ym2].vy.

   What happens now to the terms above is that when the calculated value of node[ce].vxQ and node[ce].vyQ
   are negative, the calculation of these 2nd derivative terms will change based on:

   if node[ce].vxQ < 0.0 

     node[xm1].vx - 2.0*node[ce].vx + node[xp1].vx changes to node[ce].vx - 2.0*node[xp1].vx + node[xp2].vx;

     node[xm2].vx - 2.0*node[xm1].vx + node[ce].vx changes to node[xm1].vx - 2.0*node[ce].vx + node[xp1].vx;

     node[xm1].vy - 2.0*node[ce].vy + node[xp1].vy changes to node[ce].vy - 2.0*node[xp1].vy + node[xp2].vy;

     node[xm2].vy - 2.0*node[xm1].vy + node[ce].vy changes to node[xm1].vy - 2.0*node[ce].vy + node[xp1].vy;

   if node[ce].vyQ < 0.0 

     node[ym1].vx - 2.0*node[ce].vx + node[yp1].vx changes to node[ce].vx - 2.0*node[yp1].vx + node[yp2].vx;

     node[ym2].vx - 2.0*node[ym1].vx + node[ce].vx changes to node[ym1].vx - 2.0*node[ce].vx + node[yp1].vx;

     node[ym1].vy - 2.0*node[ce].vy + node[yp1].vy changes to node[ce].vy - 2.0*node[yp1].vy + node[yp2].vy;

     node[ym2].vy - 2.0*node[ym1].vy + node[ce].vy changes to node[ym1].vy - 2.0*node[ce].vy + node[yp1].vy;


   * Note that I use vx_xp2 instead of node[xp2].vx and vx_yp2 instead of node[yp2].vy because of the cases
     where x = nx-2 or y = ny-2.  This is also true for node[xp2].vx and node[yp2].vy.

   and node[ce].vxQ and node[ce].vyQ are calculated again based on these changes.  The logic behind this
   modification in the algorithm is that when the flow is negative, the direction to retain
   upwinding is now in either the positive x or y direction.  So in the case of node[ce].vxQ < 0.0,
   all the terms are shifted one nodal point to the right.  The same is true for node[ce].vyQ < 0.0,
   with a nodal shift towards the top.  The 1982 paper by Davis and Moore hinted at this change
   for the face velocities on page 479 equation (6) and (7).  Even though they did not seem to
   apply this themselves, I have applied it to both the face velocities as well as the main
   finite difference formula given on page 481, equation (16).

   Prior to this modification in the algorithm, the code would become unstable for any Reynolds number
   greater than 100.  Previously over many months, I tried changing the boundary conditions, the
   pressure corrector using SOR, and the Courant velocity terms located on the faces of the
   control volume.  But those changes still resulted in unsuccessful simulations whereas this change
   improved things dramatically to the point where I have run calculations up to a Reynolds number
   of 4000.

   Interestingly, my actual motivation for the change was to try to create symmetry with the results
   for the flow which comes in at a 15 degree angle of attack.  When I changed the flow to
   -15 degrees, the results were very poor until I made the modification.

   Another fact compelling the change is that I noticed the streaklines and flow above the centerline
   where the obstruction is located was broken up with many regions having high concentrations of
   vorticity.  I looked back at the algorithm and felt that this was due to the terms node[ym2].vx and node[ym2].vy
   which, when calculating for velocities directly above the obstruction, node[ym2].vx and node[ym2].vy are
   located 2 nodes inside the obstruction.  Because velocities inside the obstruction are based on
   extrapolations, they are less accurate and this caused high velocity gradients to form on the edges
   of the box*.  For velocities directly below the obstruction, the only term appearing inside the box is
   [yp1] terms so there were less problems.  There are also [ym1] terms inside the obstruction
   for the velocities directly above it, but they don't seem as significant as node[ym2].vx and node[ym2].vy.

   Again, these changes made a big difference for both stability and symmetry.  I don't know if my
   reasoning is valid as it's based more on a numerical argument rather than a mathematical one.

   Incidentally, I have left in the original implementation of the algorithm for comparison.

   San Le     Saturday, November 17th, 2007

   * The box I am referring to is the when for when the single obstruction is a rectangle as given
     in the files a0, a-15, a15.

*/

#if NEW

		node[ce].vxQ = node[ce].vx
		     - CvxFvxR*(.5*(node[ce].vx + node[xp1].vx) - .5*CvxFvxR*(node[xp1].vx - node[ce].vx)
		     - ( (1.0 - CvxFvxR*CvxFvxR)/6.0 - gx )*(node[xm1].vx - 2.0*node[ce].vx + node[xp1].vx))

		     + CvxFvxL*(.5*(node[xm1].vx + node[ce].vx) - .5*CvxFvxL*(node[ce].vx - node[xm1].vx)
		     - ( (1.0 - CvxFvxL*CvxFvxL)/6.0 - gx )*(vx_xm2 - 2.0*node[xm1].vx + node[ce].vx))

		     - CvyFvxT*(.5*(node[ce].vx + node[yp1].vx) - .5*CvyFvxT*(node[yp1].vx - node[ce].vx)
		     - ( (1.0 - CvyFvxT*CvyFvxT)/6.0 - gy )*(node[ym1].vx - 2.0*node[ce].vx + node[yp1].vx))

		     + CvyFvxB*(.5*(node[ym1].vx + node[ce].vx) - .5*CvyFvxB*(node[ce].vx - node[ym1].vx)
		     - ( (1.0 - CvyFvxB*CvyFvxB)/6.0 - gy )*(vx_ym2 - 2.0*node[ym1].vx + node[ce].vx))

		     + gx*(node[xm1].vx - 2.0*node[ce].vx + node[xp1].vx)
		     + gy*(node[ym1].vx - 2.0*node[ce].vx + node[yp1].vx);

		node[ce].vyQ = node[ce].vy
		     - CvxFvyR*(.5*(node[ce].vy + node[xp1].vy) - .5*CvxFvyR*(node[xp1].vy - node[ce].vy)
		     - ( (1.0 - CvxFvyR*CvxFvyR)/6.0 - gx )*(node[xm1].vy - 2.0*node[ce].vy + node[xp1].vy))

		     + CvxFvyL*(.5*(node[xm1].vy + node[ce].vy) - .5*CvxFvyL*(node[ce].vy - node[xm1].vy)
		     - ( (1.0 - CvxFvyL*CvxFvyL)/6.0 - gx )*(vy_xm2 - 2.0*node[xm1].vy + node[ce].vy))

		     - CvyFvyT*(.5*(node[ce].vy + node[yp1].vy) - .5*CvyFvyT*(node[yp1].vy - node[ce].vy)
		     - ( (1.0 - CvyFvyT*CvyFvyT)/6.0 - gy )*(node[ym1].vy - 2.0*node[ce].vy + node[yp1].vy))

		     + CvyFvyB*(.5*(node[ym1].vy + node[ce].vy) - .5*CvyFvyB*(node[ce].vy - node[ym1].vy)
		     - ( (1.0 - CvyFvyB*CvyFvyB)/6.0 - gy )*(vy_ym2 - 2.0*node[ym1].vy + node[ce].vy))

		     + gx*(node[xm1].vy - 2.0*node[ce].vy + node[xp1].vy)
		     + gy*(node[ym1].vy - 2.0*node[ce].vy + node[yp1].vy);

		
		velocity_flag = 0;
		if( node[ce].vxQ >= 0.0 )
		{
		   du2dx2A = node[xm1].vx - 2.0*node[ce].vx + node[xp1].vx;
		   du2dx2B = vx_xm2 - 2.0*node[xm1].vx + node[ce].vx;
		   dv2dx2A = node[xm1].vy - 2.0*node[ce].vy + node[xp1].vy;
		   dv2dx2B = vy_xm2 - 2.0*node[xm1].vy + node[ce].vy;
		}
		if( node[ce].vxQ < 0.0 )
		{
		   du2dx2A = node[ce].vx - 2.0*node[xp1].vx + vx_xp2;
		   du2dx2B = node[xm1].vx - 2.0*node[ce].vx + node[xp1].vx;
		   dv2dx2A = node[ce].vy - 2.0*node[xp1].vy + vy_xp2;
		   dv2dx2B = node[xm1].vy - 2.0*node[ce].vy + node[xp1].vy;
		   velocity_flag = 1;
		}
		if( node[ce].vyQ >= 0.0 )
		{
		   du2dy2A = node[ym1].vx - 2.0*node[ce].vx + node[yp1].vx;
		   du2dy2B = vx_ym2 - 2.0*node[ym1].vx + node[ce].vx;
		   dv2dy2A = node[ym1].vy - 2.0*node[ce].vy + node[yp1].vy;
		   dv2dy2B = vy_ym2 - 2.0*node[ym1].vy + node[ce].vy;
		}
		if( node[ce].vyQ < 0.0 )
		{
		   du2dy2A = node[ce].vx - 2.0*node[yp1].vx + vx_yp2;
		   du2dy2B = node[ym1].vx - 2.0*node[ce].vx + node[yp1].vx;
		   dv2dy2A = node[ce].vy - 2.0*node[yp1].vy + vy_yp2;
		   dv2dy2B = node[ym1].vy - 2.0*node[ce].vy + node[yp1].vy;
		   velocity_flag = 1;
		}

		if( velocity_flag )
		{
		   node[ce].vxQ = node[ce].vx
		     - CvxFvxR*(.5*(node[ce].vx + node[xp1].vx) - .5*CvxFvxR*(node[xp1].vx - node[ce].vx)
		     - ( (1.0 - CvxFvxR*CvxFvxR)/6.0 - gx )*du2dx2A)

		     + CvxFvxL*(.5*(node[xm1].vx + node[ce].vx) - .5*CvxFvxL*(node[ce].vx - node[xm1].vx)
		     - ( (1.0 - CvxFvxL*CvxFvxL)/6.0 - gx )*du2dx2B)

		     - CvyFvxT*(.5*(node[ce].vx + node[yp1].vx) - .5*CvyFvxT*(node[yp1].vx - node[ce].vx)
		     - ( (1.0 - CvyFvxT*CvyFvxT)/6.0 - gy )*du2dy2A)

		     + CvyFvxB*(.5*(node[ym1].vx + node[ce].vx) - .5*CvyFvxB*(node[ce].vx - node[ym1].vx)
		     - ( (1.0 - CvyFvxB*CvyFvxB)/6.0 - gy )*du2dy2B)

		     + gx*(node[xm1].vx - 2.0*node[ce].vx + node[xp1].vx)
		     + gy*(node[ym1].vx - 2.0*node[ce].vx + node[yp1].vx);

		   node[ce].vyQ = node[ce].vy
		     - CvxFvyR*(.5*(node[ce].vy + node[xp1].vy) - .5*CvxFvyR*(node[xp1].vy - node[ce].vy)
		     - ( (1.0 - CvxFvyR*CvxFvyR)/6.0 - gx )*dv2dx2A)

		     + CvxFvyL*(.5*(node[xm1].vy + node[ce].vy) - .5*CvxFvyL*(node[ce].vy - node[xm1].vy)
		     - ( (1.0 - CvxFvyL*CvxFvyL)/6.0 - gx )*dv2dx2B)
  
		     - CvyFvyT*(.5*(node[ce].vy + node[yp1].vy) - .5*CvyFvyT*(node[yp1].vy - node[ce].vy)
		     - ( (1.0 - CvyFvyT*CvyFvyT)/6.0 - gy )*dv2dy2A)

		     + CvyFvyB*(.5*(node[ym1].vy + node[ce].vy) - .5*CvyFvyB*(node[ce].vy - node[ym1].vy)
		     - ( (1.0 - CvyFvyB*CvyFvyB)/6.0 - gy )*dv2dy2B)

		     + gx*(node[xm1].vy - 2.0*node[ce].vy + node[xp1].vy)
		     + gy*(node[ym1].vy - 2.0*node[ce].vy + node[yp1].vy);
		}
#endif


#if NEW2
/* A logical modification of the above algorithm is simply to base the shift in the 2nd derivatives
   on node[ce].vx and node[ce].vy rather than node[ce].vxQ and node[ce].vyQ.  This seems it would
   be faster since I wouldn't have to calculate node[ce].vxQ and node[ce].vyQ first and then checking
   if these quantities are negative( or node[ce].vxFvx and node[ce].vyFvy, see above.). After trying
   this, the results were mostly the same except for the -15 angle of attack problem for dt = 0.04,
   at 4800 and 7200 steps where it was somewhat worse.  It was actually a bit slower.

   But I have to admit that for the case of a2x15, it worked out to 3259 steps, and my previous change
   only works to 3119.

   San Le     Monday, June 6th, 2008
*/
		if( node[ce].vx >= 0.0 )
		{
		   du2dx2A = node[xm1].vx - 2.0*node[ce].vx + node[xp1].vx;
		   du2dx2B = vx_xm2 - 2.0*node[xm1].vx + node[ce].vx;
		   dv2dx2A = node[xm1].vy - 2.0*node[ce].vy + node[xp1].vy;
		   dv2dx2B = vy_xm2 - 2.0*node[xm1].vy + node[ce].vy;
		}
		if( node[ce].vx < 0.0 )
		{
		   du2dx2A = node[ce].vx - 2.0*node[xp1].vx + vx_xp2;
		   du2dx2B = node[xm1].vx - 2.0*node[ce].vx + node[xp1].vx;
		   dv2dx2A = node[ce].vy - 2.0*node[xp1].vy + vy_xp2;
		   dv2dx2B = node[xm1].vy - 2.0*node[ce].vy + node[xp1].vy;
		}
		if( node[ce].vy >= 0.0 )
		{
		   du2dy2A = node[ym1].vx - 2.0*node[ce].vx + node[yp1].vx;
		   du2dy2B = vx_ym2 - 2.0*node[ym1].vx + node[ce].vx;
		   dv2dy2A = node[ym1].vy - 2.0*node[ce].vy + node[yp1].vy;
		   dv2dy2B = vy_ym2 - 2.0*node[ym1].vy + node[ce].vy;
		}
		if( node[ce].vy < 0.0 )
		{
		   du2dy2A = node[ce].vx - 2.0*node[yp1].vx + vx_yp2;
		   du2dy2B = node[ym1].vx - 2.0*node[ce].vx + node[yp1].vx;
		   dv2dy2A = node[ce].vy - 2.0*node[yp1].vy + vy_yp2;
		   dv2dy2B = node[ym1].vy - 2.0*node[ce].vy + node[yp1].vy;
		}

		node[ce].vxQ = node[ce].vx
		     - CvxFvxR*(.5*(node[ce].vx + node[xp1].vx) - .5*CvxFvxR*(node[xp1].vx - node[ce].vx)
		     - ( (1.0 - CvxFvxR*CvxFvxR)/6.0 - gx )*du2dx2A)

		     + CvxFvxL*(.5*(node[xm1].vx + node[ce].vx) - .5*CvxFvxL*(node[ce].vx - node[xm1].vx)
		     - ( (1.0 - CvxFvxL*CvxFvxL)/6.0 - gx )*du2dx2B)

		     - CvyFvxT*(.5*(node[ce].vx + node[yp1].vx) - .5*CvyFvxT*(node[yp1].vx - node[ce].vx)
		     - ( (1.0 - CvyFvxT*CvyFvxT)/6.0 - gy )*du2dy2A)

		     + CvyFvxB*(.5*(node[ym1].vx + node[ce].vx) - .5*CvyFvxB*(node[ce].vx - node[ym1].vx)
		     - ( (1.0 - CvyFvxB*CvyFvxB)/6.0 - gy )*du2dy2B)

		     + gx*(node[xm1].vx - 2.0*node[ce].vx + node[xp1].vx)
		     + gy*(node[ym1].vx - 2.0*node[ce].vx + node[yp1].vx);

		node[ce].vyQ = node[ce].vy
		     - CvxFvyR*(.5*(node[ce].vy + node[xp1].vy) - .5*CvxFvyR*(node[xp1].vy - node[ce].vy)
		     - ( (1.0 - CvxFvyR*CvxFvyR)/6.0 - gx )*dv2dx2A)

		     + CvxFvyL*(.5*(node[xm1].vy + node[ce].vy) - .5*CvxFvyL*(node[ce].vy - node[xm1].vy)
		     - ( (1.0 - CvxFvyL*CvxFvyL)/6.0 - gx )*dv2dx2B)

		     - CvyFvyT*(.5*(node[ce].vy + node[yp1].vy) - .5*CvyFvyT*(node[yp1].vy - node[ce].vy)
		     - ( (1.0 - CvyFvyT*CvyFvyT)/6.0 - gy )*dv2dy2A)

		     + CvyFvyB*(.5*(node[ym1].vy + node[ce].vy) - .5*CvyFvyB*(node[ce].vy - node[ym1].vy)
		     - ( (1.0 - CvyFvyB*CvyFvyB)/6.0 - gy )*dv2dy2B)

		     + gx*(node[xm1].vy - 2.0*node[ce].vy + node[xp1].vy)
		     + gy*(node[ym1].vy - 2.0*node[ce].vy + node[yp1].vy);

#endif

#if ORIGINAL

/* Orginal algorithm. */

		node[ce].vxQ = node[ce].vx
		     - CvxFvxR*(.5*(node[ce].vx + node[xp1].vx) - .5*CvxFvxR*(node[xp1].vx - node[ce].vx)
		     - ( (1.0 - CvxFvxR*CvxFvxR)/6.0 - gx )*(node[xm1].vx - 2.0*node[ce].vx + node[xp1].vx))

		     + CvxFvxL*(.5*(node[xm1].vx + node[ce].vx) - .5*CvxFvxL*(node[ce].vx - node[xm1].vx)
		     - ( (1.0 - CvxFvxL*CvxFvxL)/6.0 - gx )*(vx_xm2 - 2.0*node[xm1].vx + node[ce].vx))

		     - CvyFvxT*(.5*(node[ce].vx + node[yp1].vx) - .5*CvyFvxT*(node[yp1].vx - node[ce].vx)
		     - ( (1.0 - CvyFvxT*CvyFvxT)/6.0 - gy )*(node[ym1].vx - 2.0*node[ce].vx + node[yp1].vx))

		     + CvyFvxB*(.5*(node[ym1].vx + node[ce].vx) - .5*CvyFvxB*(node[ce].vx - node[ym1].vx)
		     - ( (1.0 - CvyFvxB*CvyFvxB)/6.0 - gy )*(vx_ym2 - 2.0*node[ym1].vx + node[ce].vx))

		     + gx*(node[xm1].vx - 2.0*node[ce].vx + node[xp1].vx)
		     + gy*(node[ym1].vx - 2.0*node[ce].vx + node[yp1].vx);

		node[ce].vyQ = node[ce].vy
		     - CvxFvyR*(.5*(node[ce].vy + node[xp1].vy) - .5*CvxFvyR*(node[xp1].vy - node[ce].vy)
		     - ( (1.0 - CvxFvyR*CvxFvyR)/6.0 - gx )*(node[xm1].vy - 2.0*node[ce].vy + node[xp1].vy))

		     + CvxFvyL*(.5*(node[xm1].vy + node[ce].vy) - .5*CvxFvyL*(node[ce].vy - node[xm1].vy)
		     - ( (1.0 - CvxFvyL*CvxFvyL)/6.0 - gx )*(vy_xm2 - 2.0*node[xm1].vy + node[ce].vy))

		     - CvyFvyT*(.5*(node[ce].vy + node[yp1].vy) - .5*CvyFvyT*(node[yp1].vy - node[ce].vy)
		     - ( (1.0 - CvyFvyT*CvyFvyT)/6.0 - gy )*(node[ym1].vy - 2.0*node[ce].vy + node[yp1].vy))

		     + CvyFvyB*(.5*(node[ym1].vy + node[ce].vy) - .5*CvyFvyB*(node[ce].vy - node[ym1].vy)
		     - ( (1.0 - CvyFvyB*CvyFvyB)/6.0 - gy )*(vy_ym2 - 2.0*node[ym1].vy + node[ce].vy))

		     + gx*(node[xm1].vy - 2.0*node[ce].vy + node[xp1].vy)
		     + gy*(node[ym1].vy - 2.0*node[ce].vy + node[yp1].vy);

#endif


		if( umax < fabs(node[ce].vxQ) )
		{
		    umax = fabs(node[ce].vxQ);
		    umax_i = i;
		    umax_j = j;
		}
		if( umax < fabs(node[ce].vyQ) )
		{
		    umax = fabs(node[ce].vyQ);
		    umax_i = i;
		    umax_j = j;
		}

		/*if( abs(node[ce].vx) > 400. ) 
		{
		    printf(  "mmmmmm %d %d %d %9.5f %9.5f\n", i, j, ce, node[ce].vxQ, node[ce].vyQ);
		    goto here;
		}*/

		/*printf(  "%d %d %d %9.5f %9.5f\n", i, j, ce, node[ce].vxQ, node[ce].vyQ);*/

	    }
	}


	return 1;
}


