/* This program solves Poission's equation for the pressure using
   the SOR algorithm.  It contains the SOR function itself as well
   as boundary condition functions. 

                     Updated 4/14/08

    SLFCFD source file
    Version:  1.2
    Copyright (C) 2006  San Le

    The source code contained in this file is released under the
    terms of the GNU Library General Public License.

*/

#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "../compressible/cfdconst.h"
#include "../compressible/cfdstruct.h"
#include "../AMR/amr.h"

#define w               1.5
#define BIG             10.0e8

/*extern double dx, dy, dz, dt;*/
extern double err, Re, alp;
extern int imax, jmax;
extern int debug_flag, reverse_flag;


int sor( BOUND bc, INC_NODE *node, SUBGRID subgrid)
{
/* This subroutine solves Poission's equation for the pressure using
   the SOR algorithm.  f represents the pressure.

                     Updated 4/4/08
*/

	int i, j, k, m, dum, counter, node_num, check;
	int ce, xm1, xm2, xm3, xp1, xp2, xp3, ym1, ym2, ym3, yp1, yp2, yp3,
		xm1ym1, xm1yp1, xp1ym1, xp1ym2, xp1yp1;
	double pxm1, pxp1, pym1, pyp1, pxp1yp1, pxm1yp1, vy_ym1, vx_xm1;
	double bets, max_df;
	double usq1, usq2, vsq1, vsq2, uv1, uvQ, pi, df;
	int  nx, ny, max_itteration;
	double dx, dy, dz, dt;
	FILE *out;

	/*out = fopen( "o2", "w" );*/

	nx = subgrid.nx;
	ny = subgrid.ny;
	dx = subgrid.dx;
	dy = subgrid.dy;
	dt = subgrid.dt;

	max_df = 100000.0;
	bets = dx*dx/(dy*dy);
	pi = acos(-1.0);
	df = 0.0;

	max_df = 1000.0;
	counter = 0;
   
	max_itteration = 100000;

	for( k = 0; k < 1; ++k )
	{
	     while( max_df > err && counter < max_itteration )
	     {

		max_df = 0.0;
		for( j = 1; j < ny-1; ++j )
		{
		     for( i = 1; i < nx-1; ++i )
		     {
			ce  = ny*i + j;
			xm1 = ny*(i-1) + j;
			xp1 = ny*(i+1) + j;
			ym1 = ny*i + j-1;
			yp1 = ny*i + j+1;

/* The expression below is based on Equation 17.13 page 339 in

        Fletcher, Clive A. J., Computational Techniques for Fluid Dynamics: Vol II, 2nd
             Ed., Springer-Verlag, Berlin, 1991.

        where the F is replaced with uQ and G is vQ. 

        Unlike the MAC formulation, where F and G have a complex relationship with u and v (See
        equation Equation 17.9 and 17.11 page 337-338), Equation 17.22 on page 343 shows how
        the uncorrected velocity u = F.
*/

			df = 1./(2.+2*bets)*(node[xp1].prs+bets*node[yp1].prs+node[xm1].prs +
			     bets*node[ym1].prs-2.*(1.+bets)*node[ce].prs -
			     dx*dx/dt*((node[ce].vxQ-node[xm1].vxQ)/dx +
			     (node[ce].vyQ-node[ym1].vyQ)/dy));

			node[ce].prs = node[ce].prs + w*df;

			if( fabs(df) > max_df ) 
			{
			     max_df = fabs(df);
			}
		     }
		}


		check = apply_sorBC_free2( bc, node, subgrid);
		if(!check) printf( " Problems with apply_sorBC_free2\n");
		
		check = apply_sorBC_inflow( bc, node, subgrid);
		if(!check) printf( " Problems with apply_sorBC_inflow\n");

		check = apply_sorBC_freestream( bc, node, subgrid);
		if(!check) printf( " Problems with apply_sorBC_freestream\n");

		check = apply_sorBC_ends( bc, node, subgrid);
		if(!check) printf( " Problems with apply_sorBC_ends\n");

		++counter;
		if( max_df < err ) printf(  "SOR iteration counter: %4d df: %14.6e\n", counter, max_df);
		if( counter > max_itteration - 1 )
		{
		    printf(  "max iteration for pressure reached. counter: %4d error: %14.6e\n",
			counter, max_df);
		}
	     }
	}

	/*fclose(out);*/

	return 1;
}

int apply_sorBC_free2( BOUND bc, INC_NODE *node, SUBGRID subgrid)
{
	int i, j, k, m, dum, counter, node_num;
	int ce, xm1, xm2, xm3, xp1, xp2, xp3, ym1, ym2, ym3, yp1, yp2, yp3,
		xm1ym1, xm1yp1, xp1ym1, xp1ym2, xp1yp1;
	double pxp1, pyp1;
	double dx, dy, dz, dt;
	int  nx, ny;

	nx = subgrid.nx;
	ny = subgrid.ny;
	dx = subgrid.dx;
	dy = subgrid.dy;
	dt = subgrid.dt;

	for( dum = 0; dum < bc.num_free2[0].xp; ++dum )
	{
/* Below, I'm applying the BC at the right. See page 435-436 in:

      Hoffman, Joe D., Numerical Methods for Engineers and Scientists, 4th Ed.,
         McGraw-Hill Inc., New York, 1992.

*/
		node_num = bc.free2.xp[dum];
		i = node[node_num].i;
		j = node[node_num].j;

		ce   = ny*i + j;
		xm1  = ny*(i-1) + j;
		xm2  = ny*(i-2) + j;
		xm3  = ny*(i-3) + j;
		ym1  = ny*i + j-1;
		yp1  = ny*i + j+1;
		node[ce].prs = 2.0*node[xm2].prs - node[xm3].prs;
		node[ce].prs = node[xm1].prs;
		node[ce].prs = (node[yp1].prs + 2.0*node[xm1].prs + node[ym1].prs)/4.0 -
		    dx/dt*(node[ce].vxQ - node[xm1].vxQ + node[ce].vyQ - node[ym1].vyQ)/4.0;
		/*node[ce].prs = 0.0;*/
		pxp1 = 2.0*node[ce].prs - node[xm1].prs;
		pxp1 = (18.0*node[ce].prs - 9.0*node[xm1].prs + 2.0*node[xm2].prs)/11.0;

	   /*if(debug_flag) printf(  "pppp  %4d %4d %4d %16.8e %16.8e %16.8e %16.8e %16.8e\n",
		    counter, j, i, node[ce].prs, node[ym1].vyQ, node[ce].vyQ, node[xm1].vxQ, node[ce].vxQ);*/

		node[ce].vxQ = node[ce].vxQ - dt/dx*(pxp1 - node[ce].prs);
		node[ce].vyQ = node[ce].vyQ - dt/dy*(node[yp1].prs - node[ce].prs);
		node[ce].vxQ = node[ce].vx0 - dt/dx*(pxp1 - node[ce].prs);
		node[ce].vyQ = node[ce].vy0 - dt/dy*(node[yp1].prs - node[ce].prs);

	    /*printf(  "eeee %4d %16.8e %16.8e %16.8e %16.8e %16.8e\n", ce,
		2.0*node[xm2].prs - node[xm3].prs, 2.0*node[xm1].prs - node[xm2].prs,
		node[xm1].prs, node[xm2].prs, node[xm3].prs);*/

	}

	for( dum = 0; dum < bc.num_free2[0].yp; ++dum )
	{
/* Below, I'm applying the BC at the right. See page 435-436 in:

      Hoffman, Joe D., Numerical Methods for Engineers and Scientists, 4th Ed.,
         McGraw-Hill Inc., New York, 1992.

*/
		node_num = bc.free2.yp[dum];
		i = node[node_num].i;
		j = node[node_num].j;

		ce   = ny*i + j;
		ym1  = ny*i + j-1;
		ym2  = ny*i + j-2;
		ym3  = ny*i + j-3;
		xm1  = ny*(i-1) + j;
		xp1  = ny*(i+1) + j;
		node[ce].prs = 2.0*node[ym2].prs - node[ym3].prs;
		node[ce].prs = node[ym1].prs;
		node[ce].prs = (node[xp1].prs + 2.0*node[ym1].prs + node[xm1].prs)/4.0 -
		    dx/dt*(node[ce].vxQ - node[xm1].vxQ + node[ce].vyQ - node[ym1].vyQ)/4.0;
		/*node[ce].prs = 0.0;*/
		pxp1 = 2.0*node[ce].prs - node[ym1].prs;
		pyp1 = (18.0*node[ce].prs - 9.0*node[ym1].prs + 2.0*node[ym2].prs)/11.0;

	   /*if(debug_flag) printf(  "pppp  %4d %4d %4d %16.8e %16.8e %16.8e %16.8e %16.8e\n",
		    counter, j, i, node[ce].prs, node[ym1].vyQ, node[ce].vyQ, node[ym1].vxQ, node[ce].vxQ);*/

		node[ce].vxQ = node[ce].vxQ - dt/dx*(node[xp1].prs - node[ce].prs);
		node[ce].vyQ = node[ce].vyQ - dt/dy*(pyp1 - node[ce].prs);
		node[ce].vxQ = node[ce].vx0 - dt/dx*(node[xp1].prs - node[ce].prs);
		node[ce].vyQ = node[ce].vy0 - dt/dy*(pyp1 - node[ce].prs);

	    /*printf(  "eeee %4d %16.8e %16.8e %16.8e %16.8e %16.8e\n", ce,
		2.0*node[xm2].prs - node[xm3].prs, 2.0*node[xm1].prs - node[xm2].prs,
		node[xm1].prs, node[xm2].prs, node[xm3].prs);*/

	}

	return 1;
}

int apply_sorBC_inflow( BOUND bc, INC_NODE *node, SUBGRID subgrid)
{
	int i, j, k, m, dum, counter, node_num;
	int ce, xm1, xm2, xm3, xp1, xp2, xp3, ym1, ym2, ym3, yp1, yp2, yp3,
		xm1ym1, xm1yp1, xp1ym1, xp1ym2, xp1yp1;
	double vy_ym1, vx_xm1;
	double dx, dy, dz, dt;
	int  nx, ny;

	nx = subgrid.nx;
	ny = subgrid.ny;
	dx = subgrid.dx;
	dy = subgrid.dy;
	dt = subgrid.dt;

	for( dum = 0; dum < bc.num_inflow[0].xm; ++dum )
	{
/* I apply the BC at the left for pressure. */

		node_num = bc.inflow.xm[dum];
		i = node[node_num].i;
		j = node[node_num].j;

		ce  = ny*i + j;
		xp1 = ny*(i+1) + j;
		xp2 = ny*(i+2) + j;
		xp3 = ny*(i+3) + j;
		yp1 = ny*i + j+1;
		ym1 = ny*i + j-1;
		vx_xm1 = (18.0*node[ce].vxQ - 9.0*node[xp1].vxQ + 2.0*node[xp2].vxQ)/11.0;
		node[ce].prs = (node[yp1].prs + 2.0*node[xp1].prs + node[ym1].prs)/4.0 -
		    dx/dt*(node[ce].vxQ - vx_xm1 + node[ce].vyQ - node[ym1].vyQ)/4.0;

	    /*printf(  "eeee %4d %4d %4d %4d %16.8e %16.8e %16.8e\n", ce, counter, i, j, 
		node[ce].prs, node[ce].vxQ, node[ce].vyQ );*/
	}

	for( dum = 0; dum < bc.num_inflow[0].ym; ++dum )
	{
/* I apply the BC at the left for pressure. */

		node_num = bc.inflow.ym[dum];
		i = node[node_num].i;
		j = node[node_num].j;

		ce  = ny*i + j;
		xp1 = ny*(i+1) + j;
		xm1 = ny*(i-1) + j;
		yp1 = ny*i + j+1;
		yp2 = ny*i + j+2;
		yp3 = ny*i + j+3;
		vy_ym1 = (18.0*node[ce].vyQ - 9.0*node[yp1].vyQ + 2.0*node[yp2].vyQ)/11.0;
		node[ce].prs = (node[xp1].prs + 2.0*node[yp1].prs + node[xm1].prs)/4.0 -
		    dx/dt*(node[ce].vxQ - node[xm1].vxQ + node[ce].vyQ - vy_ym1)/4.0;

	    /*printf(  "eeee %4d %4d %4d %4d %16.8e %16.8e %16.8e\n", ce, counter, i, j, 
		node[ce].prs, node[ce].vxQ, node[ce].vyQ );*/
	}

	return 1;
}

int apply_sorBC_freestream( BOUND bc, INC_NODE *node, SUBGRID subgrid)
{
	int i, j, k, m, dum, counter, node_num;
	int ce, xm1, xm2, xm3, xp1, xp2, xp3, ym1, ym2, ym3, yp1, yp2, yp3,
		xm1ym1, xm1yp1, xp1ym1, xp1ym2, xp1yp1;
	double vy_ym1, vx_xm1, pxp1, pyp1;
	double dx, dy, dz, dt;
	int  nx, ny;

	nx = subgrid.nx;
	ny = subgrid.ny;
	dx = subgrid.dx;
	dy = subgrid.dy;
	dt = subgrid.dt;

	for( dum = 0; dum < bc.num_frestr[0].xp-1; ++dum )
	{
/* For pressure only at the right exit. */

		node_num = bc.frestr.xp[dum];
		i = node[node_num].i;
		j = node[node_num].j;

		ce  = ny*i + j;
		xm1 = ny*(i-1) + j;
		xm2 = ny*(i-2) + j;
		xm3 = ny*(i-3) + j;
		ym1 = ny*i + j-1;
		yp1 = ny*i + j+1;
		xm1yp1 = ny*(i-1) + j+1;  /* Note that j+1 must be <= ny-1 */
		node[ce].prs = node[xm1].prs;
		node[yp1].prs = node[xm1yp1].prs;

		pxp1 = node[ce].prs;

		node[ce].vxQ = node[ce].vx0 - dt/dx*(pxp1 - node[ce].prs);
		node[ce].vyQ = node[ce].vy0 - dt/dy*(node[yp1].prs - node[ce].prs);

		/*node[ce].prs = (node[yp1].prs + 2.0*node[xm1].prs + node[ym1].prs)/4.0 -
		    dx/dt*(node[ce].vxQ - node[xm1].vxQ + node[ce].vyQ - node[ym1].vyQ)/4.0;*/
		/*node[ce].prs = (18.0*node[xm1].prs - 9.0*node[xm2].prs + 2.0*node[xm3].prs)/11.0;*/
		/*node[ce].prs = 0.0;*/
	}

	for( dum = 0; dum < bc.num_frestr[0].xm-1; ++dum )
	{
/* For pressure only at the left exit. */

		node_num = bc.frestr.xm[dum];
		i = node[node_num].i;
		j = node[node_num].j;

		ce  = ny*i + j;
		xp1 = ny*(i+1) + j;
		xp2 = ny*(i+2) + j;
		xp3 = ny*(i+3) + j;
		ym1 = ny*i + j-1;
		yp1 = ny*i + j+1;
		xp1yp1 = ny*(i+1) + j+1;  /* Note that j+1 must be <= ny-1 */
		node[ce].prs = node[xp1].prs;
		node[yp1].prs = node[xp1yp1].prs;

		node[ce].vxQ = node[ce].vx0 - dt/dx*(node[xp1].prs - node[ce].prs);
		node[ce].vyQ = node[ce].vy0 - dt/dy*(node[yp1].prs - node[ce].prs);

		/*node[ce].prs = (node[yp1].prs + 2.0*node[xp1].prs + node[ym1].prs)/4.0 -
		    dx/dt*(node[ce].vxQ - node[xm1].vxQ + node[ce].vyQ - node[ym1].vyQ)/4.0;*/
		/*node[ce].prs = (18.0*node[xm1].prs - 9.0*node[xm2].prs + 2.0*node[xm3].prs)/11.0;*/
		/*node[ce].prs = 0.0;*/
	}


	for( dum = 0; dum < bc.num_frestr[0].yp-1; ++dum )
	{
/* For pressure only at the top exit. */

		node_num = bc.frestr.yp[dum];
		i = node[node_num].i;
		j = node[node_num].j;

		ce  = ny*i + j;
		xm1 = ny*(i-1) + j;
		xp1 = ny*(i+1) + j;
		ym1 = ny*i + j-1;
		ym2 = ny*i + j-2;
		ym3 = ny*i + j-3;
		xp1ym1 = ny*(i+1) + j-1;  /* Note that (i+1) must be <= nx-1 */
		node[ce].prs = node[ym1].prs;
		node[xp1].prs = node[xp1ym1].prs;

		pyp1 = node[ce].prs;

		node[ce].vxQ = node[ce].vx0 - dt/dx*(node[xp1].prs - node[ce].prs);
		node[ce].vyQ = node[ce].vy0 - dt/dy*(pyp1 - node[ce].prs);

		/*node[ce].prs = (node[xp1].prs + 2.0*node[ym1].prs + node[xm1].prs)/4.0 -
		    dx/dt*(node[ce].vxQ - node[xm1].vxQ + node[ce].vyQ - node[ym1].vyQ)/4.0;*/
		/*node[ce].prs = (18.0*node[ym1].prs - 9.0*node[ym2].prs + 2.0*node[ym3].prs)/11.0;*/
		/*node[ce].prs = 0.0;*/
	}

	for( dum = 0; dum < bc.num_frestr[0].ym-1; ++dum )
	{
/* For pressure only at the bottom exit. */

		node_num = bc.frestr.ym[dum];
		i = node[node_num].i;
		j = node[node_num].j;

		ce  = ny*i + j;
		xm1 = ny*(i-1) + j;
		xp1 = ny*(i+1) + j;
		yp1 = ny*i + j+1;
		yp2 = ny*i + j+2;
		yp3 = ny*i + j+3;
		xp1yp1 = ny*(i+1) + j+1;  /* Note that (i+1) must be <= nx-1 */
		node[ce].prs = node[yp1].prs;
		node[xp1].prs = node[xp1yp1].prs;

		node[ce].vxQ = node[ce].vx0 - dt/dx*(node[xp1].prs - node[ce].prs);
		node[ce].vyQ = node[ce].vy0 - dt/dy*(node[yp1].prs - node[ce].prs);

		/*vy_ym1 = (18.0*node[ce].vyQ - 9.0*node[yp1].vyQ + 2.0*node[yp2].vyQ)/11.0;
		vy_ym1 = node[ce].vyQ;
		node[ce].prs = (node[xp1].prs + 2.0*node[yp1].prs + node[xm1].prs)/4.0 -
		    dx/dt*(node[ce].vxQ - node[xm1].vxQ + node[ce].vyQ - vy_ym1)/4.0;*/
		/*node[ce].prs = (18.0*node[yp1].prs - 9.0*node[yp2].prs + 2.0*node[yp3].prs)/11.0;*/
		/*node[ce].prs = 0.0;*/
	}


	return 1;
}


int apply_sorBC_ends( BOUND bc, INC_NODE *node, SUBGRID subgrid)
{
	int i, j, k, m, dum, counter, node_num;
	int ce, xm1, xm2, xm3, xp1, xp2, xp3, ym1, ym2, ym3, yp1, yp2, yp3,
		xm1ym1, xm1yp1, xp1ym1, xp1ym2, xp1yp1;
	double vy_ym1, vx_xm1, pxp1, pyp1;
	double dx, dy, dz, dt;
	int  nx, ny;

	nx = subgrid.nx;
	ny = subgrid.ny;
	dx = subgrid.dx;
	dy = subgrid.dy;
	dt = subgrid.dt;

	if(bc.num_inflow[0].xm > 0)
	{
/* For pressure at top left. */

	    dum = bc.num_inflow[0].xm-1;
	    node_num = bc.inflow.xm[dum];
	    i = node[node_num].i;
	    j = node[node_num].j+1; /* Note I set this at 1 node more than the last node. */

	    ce  = ny*i + j;
	    xp1 = ny*(i+1) + j;
	    xp2 = ny*(i+2) + j;
	    xp3 = ny*(i+3) + j;

	    node[ce].prs = (18.0*node[xp1].prs - 9.0*node[xp2].prs + 2.0*node[xp3].prs)/11.0;

/* For pressure at bottom left. */

	    dum = 0;
	    node_num = bc.inflow.xm[dum];
	    i = node[node_num].i;
	    j = node[node_num].j-1; /* Note I set this at 1 node less than the first node. */

	    ce  = ny*i + j;
	    xp1 = ny*(i+1) + j;
	    xp2 = ny*(i+2) + j;
	    xp3 = ny*(i+3) + j;

	    node[ce].prs = (18.0*node[xp1].prs - 9.0*node[xp2].prs + 2.0*node[xp3].prs)/11.0;
	}

	if(bc.num_inflow[0].ym > 0)
	{
/* For pressure at bottom left. */

	    dum = bc.num_inflow[0].ym-1;
	    node_num = bc.inflow.ym[dum];
	    i = node[node_num].i+1; /* Note I set this at 1 node more than the last node. */
	    j = node[node_num].j;

	    ce  = ny*i + j;
	    yp1 = ny*i + j+1;
	    yp2 = ny*i + j+2;
	    yp3 = ny*i + j+3;

	    node[ce].prs = (18.0*node[yp1].prs - 9.0*node[yp2].prs + 2.0*node[yp3].prs)/11.0;

/* For pressure at bottom right. */

	    dum = 0;
	    node_num = bc.inflow.ym[dum];
	    i = node[node_num].i-1; /* Note I set this at 1 node less than the first node. */
	    j = node[node_num].j;

	    ce  = ny*i + j;
	    yp1 = ny*i + j+1;
	    yp2 = ny*i + j+2;
	    yp3 = ny*i + j+3;

	    node[ce].prs = (18.0*node[yp1].prs - 9.0*node[yp2].prs + 2.0*node[yp3].prs)/11.0;
	}


	if(bc.num_free2[0].xp > 0)
	{
/* For pressure at top right. */

	    dum = bc.num_free2[0].xp-1;
	    node_num = bc.free2.xp[dum];
	    i = node[node_num].i;
	    j = node[node_num].j+1; /* Note I set this at 1 node more than the last node. */

	    ce  = ny*i + j;
	    xm1 = ny*(i-1) + j;
	    xm2 = ny*(i-2) + j;
	    xm3 = ny*(i-3) + j;
	    ym1 = ny*i + j-1;

	    node[ce].prs = .5*(node[xm1].prs + node[ym1].prs);
	    node[ce].prs = (18.0*node[xm1].prs - 9.0*node[xm2].prs + 2.0*node[xm3].prs)/11.0;
	    node[ce].prs = node[ym1].prs;

/* For pressure at bottom right. */

	    dum = 0;
	    node_num = bc.free2.xp[dum];
	    i = node[node_num].i;
	    j = node[node_num].j-1; /* Note I set this at 1 node less than the first node. */

	    ce  = ny*i + j;
	    xm1 = ny*(i-1) + j;
	    xm2 = ny*(i-2) + j;
	    xm3 = ny*(i-3) + j;
	    yp1 = ny*i + j+1;

	    node[ce].prs = .5*(node[xm1].prs + node[yp1].prs);
	    node[ce].prs = (18.0*node[xm1].prs - 9.0*node[xm2].prs + 2.0*node[xm3].prs)/11.0;
	    node[ce].prs = node[yp1].prs;
	}

	if(bc.num_free2[0].yp > 0)
	{
/* For pressure at top right. */

	    dum = bc.num_free2[0].yp-1;
	    node_num = bc.free2.yp[dum];
	    i = node[node_num].i+1; /* Note I set this at 1 node more than the last node. */
	    j = node[node_num].j;

	    ce  = ny*i + j;
	    ym1 = ny*i + j-1;
	    ym2 = ny*i + j-2;
	    ym3 = ny*i + j-3;
	    xm1 = ny*(i-1) + j;

	    node[ce].prs = .5*(node[xm1].prs + node[ym1].prs);
	    node[ce].prs = (18.0*node[ym1].prs - 9.0*node[ym2].prs + 2.0*node[ym3].prs)/11.0;
	    node[ce].prs = node[xm1].prs;

/* For pressure at top left. */

	    dum = 0;
	    node_num = bc.free2.yp[dum];
	    i = node[node_num].i-1; /* Note I set this at 1 node less than the first node. */
	    j = node[node_num].j;

	    ce  = ny*i + j;
	    ym1 = ny*i + j-1;
	    ym2 = ny*i + j-2;
	    ym3 = ny*i + j-3;
	    xp1 = ny*(i+1) + j;

	    node[ce].prs = .5*(node[xm1].prs + node[ym1].prs);

	    node[ce].prs = .5*(node[xm1].prs + node[yp1].prs);
	    node[ce].prs = (18.0*node[ym1].prs - 9.0*node[ym2].prs + 2.0*node[ym3].prs)/11.0;
	    node[ce].prs = node[xp1].prs;
	}



	if(bc.num_frestr[0].xp > 0)
	{
/* For node at top right. */

	    dum = bc.num_frestr[0].xp - 1;
	    node_num = bc.frestr.xp[dum];
	    i = node[node_num].i;
	    j = node[node_num].j;

	    ce  = ny*i + j;
	    xm1 = ny*(i-1) + j;
	    ym1 = ny*i + j-1;
	    ym2 = ny*i + j-2;

	    pyp1 = (18.0*node[ce].prs - 9.0*node[ym1].prs + 2.0*node[ym2].prs)/11.0;
	    pxp1 = node[ce].prs;

	    node[ce].vxQ = node[ce].vx0 - dt/dx*(pxp1 - node[ce].prs);
	    node[ce].vyQ = node[ce].vy0 - dt/dy*(pyp1 - node[ce].prs);
	}

	if(bc.num_frestr[0].xm > 0)
	{
/* For node at bottom right. */

	    dum = bc.num_frestr[0].xm - 1;
	    node_num = bc.frestr.xm[dum];
	    i = node[node_num].i;
	    j = node[node_num].j;

	    ce  = ny*i + j;
	    ym1 = ny*i + j-1;
	    ym2 = ny*i + j-2;
	    xp1 = ny*(i+1) + j;

	    pyp1 = (18.0*node[ce].prs - 9.0*node[ym1].prs + 2.0*node[ym2].prs)/11.0;

	    node[ce].vxQ = node[ce].vx0 - dt/dx*(node[xp1].prs - node[ce].prs);
	    node[ce].vyQ = node[ce].vy0 - dt/dy*(pyp1 - node[ce].prs);
	}



	if(bc.num_frestr[0].yp > 0)
	{
/* For node at top right. */

	    dum = bc.num_frestr[0].yp - 1;
	    node_num = bc.frestr.yp[dum];
	    i = node[node_num].i;
	    j = node[node_num].j;

	    ce  = ny*i + j;
	    xm1 = ny*(i-1) + j;
	    xm2 = ny*(i-2) + j;
	    ym1 = ny*i + j-1;

	    pxp1 = (18.0*node[ce].prs - 9.0*node[xm1].prs + 2.0*node[xm2].prs)/11.0;
	    pyp1 = node[ce].prs;

	    node[ce].vxQ = node[ce].vx0 - dt/dx*(pxp1 - node[ce].prs);
	    node[ce].vyQ = node[ce].vy0 - dt/dy*(pyp1 - node[ce].prs);
	}


	if(bc.num_frestr[0].ym > 0)
	{
/* For node at bottom right. */

	    dum = bc.num_frestr[0].ym - 1;
	    node_num = bc.frestr.ym[dum];
	    i = node[node_num].i;
	    j = node[node_num].j;

	    ce  = ny*i + j;
	    xm1 = ny*(i-1) + j;
	    xm2 = ny*(i-2) + j;
	    yp1 = ny*i + j+1;

	    pxp1 = (18.0*node[ce].prs - 9.0*node[xm1].prs + 2.0*node[xm2].prs)/11.0;

	    node[ce].vxQ = node[ce].vx0 - dt/dx*(pxp1 - node[ce].prs);
	    node[ce].vyQ = node[ce].vy0 - dt/dy*(node[yp1].prs - node[ce].prs);
	}

/* For pressures that are set to 0.0. */

	for( dum = 0; dum < bc.num_prs0[0]; ++dum )
	{
		node_num = bc.prs0[dum];
		i = node[node_num].i;
		j = node[node_num].j;

		ce = ny*i + j;
		node[ce].prs = 0.0;
	}

	return 1;
}
